//  "BVdecomp.cc"
// December 2002
/**
  Application  for evaluating the bias variance decompostion of the error.\\
	 It evaluates the bias variance decomposition of the error using the 0/1 loss for two-class
	 classification problems.
*/

#include<stdlib.h>
#include"utils.h"
#include"bv2decomp.h"

static unsigned const SIZE = 256;

static bool b_save_bv_persample = false;
static bool b_save_margins = false;
static bool b_oob_eval = false;

static char learningset[SIZE];   
static char predictedbasename[SIZE];  
static char bsmatrix[SIZE];
int ndatasets;  
static char bvpersample[SIZE]; 
static char marginfile[SIZE];  
 
static void print_info_use(void);

int doarg(char* arg[], int i);

////////////////////////////////////////////////////////
int main(int argc, char* argv[])
{
  const int WIDTH = 16;
	AverageBV AvgBV;
  	
  if ((argc < 4) || ((argc%2) == 1))
	  print_info_use();
  

  strcpy(learningset, argv[1]);
  strcpy(predictedbasename, argv[2]);
  ndatasets = atoi(argv[3]);

	
	for (int i=4; i<argc; i+=2) 
     if (doarg(argv, i))
        print_info_use();
	
  
	// oob estimation of BV decomposition
	BV2decomp bv(learningset, predictedbasename, ndatasets);	
	if (b_oob_eval) {
    short **m = bv.get_bs_matrix(bsmatrix);
	  bv.calc_oob_bv_decomposition(AvgBV, m);
  }
  else
    bv.calc_bv_decomposition(AvgBV);
	
	// printing results on stdout
	cout.setf(ios::left, ios::adjustfield);
  cout.precision(8);  
  cout.width(WIDTH); cout << "# Loss";
  cout.width(WIDTH); cout << "bias";
  cout.width(WIDTH); cout << "net-variance";
  cout.width(WIDTH); cout << "unbias-var";
  cout.width(WIDTH); cout << "bias-var";
	cout.width(WIDTH); cout << "variance";
  cout << endl;
	bv.print();
	
	if (b_save_bv_persample)
	  bv.save_bv_per_sample(bvpersample);
	if (b_save_margins)  {
	  bv.calc_margins();
    bv.calc_cumulative_distr_margins();
		bv.save_cumulative_distr_margins(marginfile);
	}

  return 0;
}



void print_info_use(void)
{
  cerr << "
  BVdecomp. NEURObjects 2.0 application for evaluating bias
            variance decomposition of the error.
  Usage:    BVdecomp learningset pred n [options]
            learningset  data set file
            pred         base file name of the predicted classes (pred1.out ... predn.out)
            n            number of prediction files
  Options:
  -bvps string  file  storing bias-variance decomposition values per example
  -marg string  file  storing the cumulative distribution of the margins
  -bsm  string  file  storing the boostrap matrix for oob estimation
  " << endl;
  exit(1);
}


int doarg(char* arg[], int i)
{ 	
  if (!strcmp(arg[i], "-bvps")) {
    b_save_bv_persample = true;
		strcpy(bvpersample, arg[i+1]);}
	else if (!strcmp(arg[i], "-marg")) {
    b_save_margins = true;
		strcpy(marginfile, arg[i+1]);}
  else if (!strcmp(arg[i], "-bsm")) {
    b_oob_eval = true;
		strcpy(bsmatrix, arg[i+1]);}
  else 
    return 1;
  return 0;
} 
