////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////

/**
	 Trains the bagged ensemble using a specific data set.\\
	 It trains the base learners using bootstrap resampling.
	 The training set at each iteration is resampled using a uniform probability distribution.
	 At each iteration it stores  error training rate of the ensemble in the list train_errors.
	 @param trainfile  name of the training set
	 @param num_of_samples number of data to be resampled at each iteration. 
	                       If it is set to 0 (default) the number is equal to the cardinality
												 of the data set.
	 @param maxiterations maximun number of iterations(base learners).
	 @return the number of trained base learners
*/
unsigned WeightedBagging::train (char* trainfile, unsigned num_of_attributes = 2; unsigned num_of_samples = 0,
                unsigned maxiterations = MAX_ITERATIONS) {

  char* resampled_data = "resampled_data.temp";
  int it;
  float err;

  int num_train = wc (trainfile);  
	if (trainset !=NIL) {
	  delete trainset;
		trainset = NIL;
	}	  
  trainset = new TSetStrClass(n_attr, num_train, trainfile);

  if (num_of_samples == 0)
    num_of_samples = num_train;

  for (it = 0; it < maxiterations; it++) {
     trainset->Subsample_with_replacement (resampled_data, num_of_samples); 
	   BaseLearner* b = new BaseLearner(base);
	   b->set_number(++current_it); 
	   err = b->train(resampled_data);
     baselist.push_back(b);
	   train_errors.push_back(err);
  }
  system("rm -f  resampled_data.temp");
  return (current_it+1);
}



/*
	 Combines the bagged ensemble of trained base learners.\\
	 It performs a test of the bagged ensembled on an assignated test set (testfile). 
	 The trained base learners are stored in the baselist.
	 For each sample of the test file all the base learners give their prediction and the predicted class of
	 the ensemble is assigned through majority voting.\\
	 The base learner used in this method must have its model stored in baselist and must implement the method
	 test_pattern and must process data file in a format compatible with the class TrainingSet.
	 Output: error rate, confusion matrix and file of the classes predicted by the ensemble.
	 @param testfile  name of the testing set. 
	 @param predicted file storing the predicted classes.
	 @return error rate
*/
float WeightedBagging::combine(char* testfile, char* predicted) {

  int predicted_class;
	int trueclass;
	int num_errors = 0;
	vect vclasses(nclass);
	
	int num_test = wc (testfile);  
	if (testset !=NIL) {
	  delete testset;
		testset = NIL;
	}	  
  testset = new TrainingSet(n_attr, num_test, testfile);
  vect& target = testset.Read_target();
  matrix& data = testset.Read_data();
	vect vpredicted(num_test);
	
	for (int i=0; i<num_train; i++) {
	  vect& pat data.row(i);
	  for (int j=0; j<baselist.size(); j++) {
	    int cl = baselist[j]->test_pattern(pat);
			if (cl = -1) cl = 2; // for two-classes problems
			vclasses[cl-1]+=1;
		}
		vpredicted[i] = majority_voting(vclasses);
		if ((int)target[i] == -1)
		   trueclass = 2;
		else
		   trueclass = (int)target[i];
		confusion((int)vpredicted[i]-1, trueclass-1) += 1;
		if ((int)vpredicted[i] != (int)target[i])
		   num_errors++;
	}
	
	ofstream outp(predicted);
	for (int i=0; i<num_test); i++)
	  outp << (int)vpredicted[i] << endl;
		
	return err=(float)num_errors/(float)num_test;
		
}
