/* ******************************************************************************
NEURObjects 2.0
baselearner.cc January 2002
Giorgio Valentini
****************************************************************************** */

#include<strstream>
#include<cstdlib>
#include"baselearner.h"
#include"utils.h"


/* Computes the training error reading the number of amples and the number of misclassification from
  the output of the SVM-light application svm_learn */
static float SVMLight_compute_train_error(char* outfile) {

  unsigned num_samples = 0;
	unsigned num_errors;
	float error_rate;
	char c;
	const int szlong = 4096;
	char buf1[szlong];
	char buf2[szlong];
	
	
  ifstream in(outfile);
	if (!in) error_exit("Cannot open file  ", outfile);
	in.getline(buf1,szlong); //discard first line
	in.getline(buf1,szlong);
	istrstream is(buf1,szlong);
	for (int i=0; i < 4; i++)
	  is >> buf2;
	is >> c;
	is >> num_samples;
	
	in.getline(buf1,szlong); //discard first line
	in.getline(buf1,szlong);	
	is.seekg(0,ios::beg);
	for (int i=0; i < 2; i++)
	  is >> buf2;
	is >> c;
	is >> num_errors;
	
	return error_rate = (float)num_errors/(float)num_samples;
}


/* Computes the training error reading the number of amples and the number of misclassification from
  the output of the SVM-light application svm_classify */
static float SVMLight_compute_test_error(char* outfile) {

  unsigned num_samples = 0;
	unsigned num_errors;
	float error_rate;
	const int szlong = 4096;
	char buf1[szlong];
	char buf2[256];
	
	
  ifstream in(outfile);
	if (!in) error_exit("Cannot open file  ", outfile);
	in.getline(buf1,szlong); //discard first line
	in.getline(buf1,szlong);
	istrstream is(buf1,szlong);
	for (int i=0; i < 7; i++)
	  is >> buf2;
	is >> num_errors;	
	is >> buf2;
	is >> num_samples;
		
	return error_rate = (float)num_errors/(float)num_samples;
}


/* Saves the predicted classes readingthe output file of the   SVM-light application svm_classify  */
static void SVMLight_save_predicted(char* predicted) {

 
	const int sz = 256;
	char buf[sz];
	float res;
	unsigned cl;
	
	ostrstream s(buf,sz,ios::out);
	s << predicted << ".out" << ends;
	
  ifstream in(buf);
	if (!in) error_exit("Cannot open file  ", buf);
	ofstream out(predicted);
	if (!out) error_exit("Cannot open file  ", predicted);
	
	in >> res;
	while (!in.eof()) {	  
		if (res >= 0.0)
		  cl = 1;
		else
		  cl = 2;
		out << cl << endl;
		in >> res;
	}
	
	if (in.bad())
	  error_exit("Error reading file", buf);
  if (out.bad())
	  error_exit("Error writing file", predicted);	
}





///////////////////////////////////////////////////////////////////////////////////
// BaseSVM_light_linear


/**
	 Trains the base learner using a training set stored in a file.\\
	 The base learner is trained using the data set trainfile. The learned model is stored in the
	 file model. The file formats are in  SVM-light file format.
	 @param trainfile  name of the training set
	 @param model  model file name.
	 @return error rate
*/
float BaseSVM_light_linear::train (char* trainfile, char* model){
  
	const int sz = 256;
  char buf[sz];	
	float err;
	
	ostrstream s(buf,sz,ios::out);
	s << "svm_learn " << type << "-c " << C << " " << trainfile << " " << model << \
	     " > " << trainfile << ".out" << ends;
	system (buf);
	
	s.seekp(0,ios::beg);
	s << trainfile << ".out" << ends;
	err = SVMLight_compute_train_error(buf);	
	s.seekp(0,ios::beg);
	s << "rm -f " << trainfile << ".out" << ends;
	system (buf);
	
	return err;
}			


/**
	 Tests the base learner using a testing set stored in a file.\\
	 The learner is tested using a testing set stored in a file and a  previously learned model  stored in a
	 file, both in SVM-light format.
	 The output results (computed classes) are stored in a file.
	 Output file format: The output classes are stored one for each line. The classes must be labeled by
	 integers: The first class by 1, the second by 2, etc. The output files are used by hte ensemble classes to
	 combine the outputs of the base learners.	 
	 @param testfile  name of the testing set. 
	 @param model  model file name.
	 @param predicted computed class file.
	 @return error rate
*/
float BaseSVM_light_linear::test(char* testfile, char* model, char* predicted){
  
	const int sz = 256;
  char buf[sz];	
	float err;
	
	ostrstream s(buf,sz,ios::out);
	s << "svm_classify " << "-v 1 " << testfile << " " << model << " " << predicted << ".out" \
	     " > " << testfile <<  ".out" << ends;
	system (buf);
	
	s.seekp(0,ios::beg);
	s << testfile << ".out" << ends;
	err = SVMLight_compute_test_error(buf);	
	s.seekp(0,ios::beg);
	s << "rm -f " << testfile << ".out" << ends;
	system (buf);
	
	//save predicted classes into a file
	SVMLight_save_predicted(predicted);
	
	//s.seekp(0,ios::beg);
	//s << "rm -f " << predicted << ".out" << ends;
	//system (buf);
	
	return err;
}













