/* ******************************************************************************
NEURObjects 2.0
basemlp.cc January 2002
Completed in May 2002
Giorgio Valentini
****************************************************************************** */

#include<iostream>
#include<strstream>
#include<cstdlib>
#include<string.h>
#include"baselearner.h"
#include"basemlp.h"
#include"utils.h"
#include<algorithm>


// BaseMLP_OneLayer ---------------------------------------------------------

/// Dummy constructor
BaseMLP_OneLayer::BaseMLP_OneLayer() : OneLayer() {
   num = 0;
	 seed = 0;  
   max_iterations = MAXITS;
   RMS_threshold = ERR_TOL;
   learn = NULL;
	 //copy(predicted_classes.begin(), predicted_classes.end(), ostream_iterator<unsigned>(cout, " "));
}

/* Constructor. */
BaseMLP_OneLayer::BaseMLP_OneLayer (unsigned nout, unsigned nin, Learning* learn_alg, 
                  double errtol, unsigned maxits, unsigned n, char sep, int s)  
                 : OneLayer(nout, nin) {
	 seed = s;
	 num = n; 
	 separator = sep;  
   max_iterations = maxits;
   RMS_threshold = errtol;
   learn = learn_alg;
	 
}

/* Trains the base learner using a training set stored in a file. */
float BaseMLP_OneLayer::train_learner (char* trainfile, char* model){

  unsigned iter = 0;
	double error = 0;
	
	int num_train = wc (trainfile);   
	TrainingSet tset( Ninput(), num_train,  trainfile, separator, TrainingSet::last);
	
	tset.normalize();
	// saves the mean and stdev for the test
	mean = tset.Read_mean();
  stdev = tset.Read_stdev();
	init_weights_norm(seed);
	Learn_by_pattern(tset, *learn, iter, error, RMS_threshold, max_iterations, false);
	save_weights(model);
	confmatrix = test(tset);
	return (float)get_perc_errors();
}




/*Tests the base learner using a testing set stored in a file. */
float BaseMLP_OneLayer::test_learner(char* testfile, char* model, char* predicted){

  int num_train = wc (testfile);   
	TrainingSet tset( Ninput(), num_train,  testfile, separator, TrainingSet::last);
  tset.normalize(mean,stdev);
	read_weights(model);
	confmatrix  = test(tset,predicted_classes);
  save_predicted(predicted_classes, predicted);
	return (float)get_perc_errors();
}



/* Trains the base learner using a training set stored in memory. */
float BaseMLP_OneLayer::train_learner (TrainingSet& Tset, char* model = NULL){
  
	unsigned iter = 0;
	double error = 0;
	
	if (!Tset.Is_normalized())
	  Tset.normalize();
	// saves the mean and stdev for the test
	mean = Tset.Read_mean();
  stdev = Tset.Read_stdev();
	init_weights_norm(seed);
	Learn_by_pattern(Tset, *learn, iter, error, RMS_threshold, max_iterations, false);
	if (model != NULL)
	   save_weights(model);
	confmatrix = test(Tset);
	return (float)get_perc_errors();
}


/* Tests the base learner using a testing set stored in memory */
float BaseMLP_OneLayer::test_learner(TrainingSet& Tset, char* model = NULL, char* predicted = NULL) {

  if (!Tset.Is_normalized())
	  Tset.normalize(mean,stdev);
	if (model != NULL)
	  read_weights(model);
	confmatrix  = test(Tset,predicted_classes);
	if (predicted != NULL)
    save_predicted(predicted_classes, predicted);
	return (float)get_perc_errors();
}


/* Computes the predicted class on a specific input. */
inline unsigned BaseMLP_OneLayer::test_pattern (vect& v) {
  return predictclass(v);
}



// BaseMLP_TwoLayer ---------------------------------------------------------

/// Dummy constructor
BaseMLP_TwoLayer::BaseMLP_TwoLayer() : TwoLayer() {
   num = 0;
	 seed = 0;  
   max_iterations = MAXITS;
   RMS_threshold = ERR_TOL;
   learn = NULL;
	 //copy(predicted_classes.begin(), predicted_classes.end(), ostream_iterator<unsigned>(cout, " "));
}

/* Constructor. */
BaseMLP_TwoLayer::BaseMLP_TwoLayer (unsigned nout, unsigned nin, unsigned nhidden, Learning* learn_alg, 
                  double errtol, unsigned maxits, unsigned n, char sep, int s)  
                 : TwoLayer(nout, nhidden, nin) {
	 seed = s;
	 num = n; 
	 separator = sep;  
   max_iterations = maxits;
   RMS_threshold = errtol;
   learn = learn_alg;
	 
}

/* Trains the base learner using a training set stored in a file. */
float BaseMLP_TwoLayer::train_learner (char* trainfile, char* model){

  unsigned iter = 0;
	double error = 0;
	
	int num_train = wc (trainfile);   
	TrainingSet tset( Ninput(), num_train,  trainfile, separator, TrainingSet::last);
	
	tset.normalize();
	// saves the mean and stdev for the test
	mean = tset.Read_mean();
  stdev = tset.Read_stdev();
	init_weights_norm(seed);
	Learn_by_pattern(tset, *learn, iter, error, RMS_threshold, max_iterations, false);
	save_weights(model);
	confmatrix = test(tset);
	return (float)get_perc_errors();
}


/* Trains the base learner using a training set stored in memory. */
float BaseMLP_TwoLayer::train_learner (TrainingSet& Tset, char* model = NULL){
  
	unsigned iter = 0;
	double error = 0;
	
	if (!Tset.Is_normalized()) 
	  Tset.normalize();
	// saves the mean and stdev for the test
	mean = Tset.Read_mean();
  stdev = Tset.Read_stdev();
	init_weights_norm(seed);
	Learn_by_pattern(Tset, *learn, iter, error, RMS_threshold, max_iterations, false);
	if (model != NULL)
	   save_weights(model);
	confmatrix = test(Tset);
	return (float)get_perc_errors();
}

/*Tests the base learner using a testing set stored in a file. */
float BaseMLP_TwoLayer::test_learner(char* testfile, char* model, char* predicted){

  int num_train = wc (testfile);   
	TrainingSet tset( Ninput(), num_train,  testfile, separator, TrainingSet::last);
  tset.normalize(mean,stdev);
	read_weights(model);
	confmatrix  = test(tset,predicted_classes);
  save_predicted(predicted_classes, predicted);
	return (float)get_perc_errors();
}

/* Tests the base learner using a testing set stored in memory */
float BaseMLP_TwoLayer::test_learner(TrainingSet& Tset, char* model = NULL, char* predicted = NULL) {

  if (!Tset.Is_normalized())
	  Tset.normalize(mean,stdev);
	if (model != NULL)
	  read_weights(model);
	confmatrix  = test(Tset,predicted_classes);
	if (predicted != NULL)
    save_predicted(predicted_classes, predicted);
	return (float)get_perc_errors();
}


/* Computes the predicted class on a specific input. */
inline unsigned BaseMLP_TwoLayer::test_pattern (vect& v) {
  return predictclass(v);
}


// BaseMLP_ThreeLayer ---------------------------------------------------------

/// Dummy constructor
BaseMLP_ThreeLayer::BaseMLP_ThreeLayer() : ThreeLayer() {
   num = 0;
	 seed = 0;  
   max_iterations = MAXITS;
   RMS_threshold = ERR_TOL;
   learn = NULL;
	 //copy(predicted_classes.begin(), predicted_classes.end(), ostream_iterator<unsigned>(cout, " "));
}

/* Constructor. */
BaseMLP_ThreeLayer::BaseMLP_ThreeLayer (unsigned nout, unsigned nin, unsigned nhidden1, unsigned nhidden2, Learning* learn_alg, 
                  double errtol, unsigned maxits, unsigned n, char sep, int s)  
                 : ThreeLayer(nout, nhidden1, nhidden2, nin) {
	 seed = s;
	 num = n; 
	 separator = sep;  
   max_iterations = maxits;
   RMS_threshold = errtol;
   learn = learn_alg;
	 
}

/* Trains the base learner using a training set stored in a file. */
float BaseMLP_ThreeLayer::train_learner (char* trainfile, char* model){

  unsigned iter = 0;
	double error = 0;
	
	int num_train = wc (trainfile);   
	TrainingSet tset( Ninput(), num_train,  trainfile, separator, TrainingSet::last);
	
	tset.normalize();
	// saves the mean and stdev for the test
	mean = tset.Read_mean();
  stdev = tset.Read_stdev();
	init_weights_norm(seed);
	Learn_by_pattern(tset, *learn, iter, error, RMS_threshold, max_iterations, false);
	save_weights(model);
	confmatrix = test(tset);
	return (float)get_perc_errors();
}


/* Trains the base learner using a training set stored in memory. */
float BaseMLP_ThreeLayer::train_learner (TrainingSet& Tset, char* model = NULL){
  
	unsigned iter = 0;
	double error = 0;
	
	if (!Tset.Is_normalized()) 
	  Tset.normalize();
	// saves the mean and stdev for the test
	mean = Tset.Read_mean();
  stdev = Tset.Read_stdev();
	init_weights_norm(seed);
	Learn_by_pattern(Tset, *learn, iter, error, RMS_threshold, max_iterations, false);
	if (model != NULL)
	   save_weights(model);
	confmatrix = test(Tset);
	return (float)get_perc_errors();
}

/*Tests the base learner using a testing set stored in a file. */
float BaseMLP_ThreeLayer::test_learner(char* testfile, char* model, char* predicted){

  int num_train = wc (testfile);   
	TrainingSet tset( Ninput(), num_train,  testfile, separator, TrainingSet::last);
  tset.normalize(mean,stdev);
	read_weights(model);
	confmatrix  = test(tset,predicted_classes);
  save_predicted(predicted_classes, predicted);
	return (float)get_perc_errors();
}

/* Tests the base learner using a testing set stored in memory */
float BaseMLP_ThreeLayer::test_learner(TrainingSet& Tset, char* model = NULL, char* predicted = NULL) {

  if (!Tset.Is_normalized())
	  Tset.normalize(mean,stdev);
	if (model != NULL)
	  read_weights(model);
	confmatrix  = test(Tset,predicted_classes);
	if (predicted != NULL)
    save_predicted(predicted_classes, predicted);
	return (float)get_perc_errors();
}


/* Computes the predicted class on a specific input. */
inline unsigned BaseMLP_ThreeLayer::test_pattern (vect& v) {
  return predictclass(v);
}







