/* ******************************************************************************
NEURObjects 2.0
baselearner.cc January 2002
Giorgio Valentini
****************************************************************************** */

#include<iostream>
#include<strstream>
#include<cstdlib>
#include<string.h>
#include"baselearner.h"
#include"basesvml.h"
#include"utils.h"

///////////////////////////////////////////////////////////////////////////////////
// BaseSVM_light_common

/* Computes the training error reading the number of amples and the number of misclassification from
  the output of the SVM-light application svm_learn */
float BaseSVM_light_common::read_train_error(char* outfile) {

  unsigned num_samples = 0;
	unsigned num_errors;
	float error_rate;
	char c;
	const int szlong = 4096;
	char buf1[szlong];
	char buf2[szlong];
	
	
  ifstream in(outfile);
	if (!in) error_exit("Cannot open file  ", outfile);
	in.getline(buf1,szlong); //discard first line
	in.getline(buf1,szlong);
	istrstream is(buf1,szlong);
	for (int i=0; i < 4; i++)
	  is >> buf2;
	is >> c;
	is >> num_samples;
	
	//in.getline(buf1,szlong); //discard first line
	//in.getline(buf1,szlong);	
	//is.seekg(0,ios::beg);
	//for (int i=0; i < 2; i++)
	//  is >> buf2;
	
	// discard all the words before the number of errors: it stops on the word "finished"
	char vv[szlong];
	char* end_string = "finished";
	do {
	  in >> vv;
		//cout << "vv = " << vv << endl;
	} while (strcmp(vv,end_string));
	// In the input there is a string "(num_errors"
	in >> c;
	in >> num_errors;
	
	return error_rate = (float)num_errors/(float)num_samples;
}


/* Computes the training error reading the number of amples and the number of misclassification from
  the output of the SVM-light application svm_classify */
float BaseSVM_light_common::read_test_error(char* outfile) {

  unsigned num_samples = 0;
	unsigned num_errors;
	float error_rate;
	const int szlong = 4096;
	char buf1[szlong];
	char buf2[256];
	
	
  ifstream in(outfile);
	if (!in) error_exit("Cannot open file  ", outfile);
	in.getline(buf1,szlong); //discard first line
	in.getline(buf1,szlong);
	istrstream is(buf1,szlong);
	for (int i=0; i < 7; i++)
	  is >> buf2;
	is >> num_errors;	
	is >> buf2;
	is >> num_samples;
		
	return error_rate = (float)num_errors/(float)num_samples;
}


/* Saves the predicted classes readingthe output file of the   SVM-light application svm_classify  */
void BaseSVM_light_common::save_predicted(char* predicted) {

 
	const int sz = 256;
	char buf[sz];
	float res;
	unsigned cl;
	
	ostrstream s(buf,sz,ios::out);
	s << predicted << OUT << ends;
	
  ifstream in(buf);
	if (!in) error_exit("Cannot open file  ", buf);
	ofstream out(predicted);
	if (!out) error_exit("Cannot open file  ", predicted);
	
	in >> res;
	while (!in.eof()) {	  
		if (res >= 0.0)
		  cl = 1;
		else
		  cl = 2;
		out << cl << endl;
		in >> res;
	}
	
	if (in.bad())
	  error_exit("Error reading file", buf);
  if (out.bad())
	  error_exit("Error writing file", predicted);	
}


/*
	 Tests the base learner using a testing set stored in a file.	 
*/
float BaseSVM_light_common::test_learner(char* testfile, char* model, char* predicted){
  
	const int sz = 256;
  char buf[sz];	
	float err;
	
	ostrstream s(buf,sz,ios::out);
	s << "svm_classify " << "-v 1 " << testfile << " " << model << " " << predicted << OUT \
	  <<  " > " << testfile <<  OUT << ends;
	system (buf);
	s.seekp(0,ios::beg);
	s << testfile << OUT << ends;
	err = read_test_error(buf);	
	s.seekp(0,ios::beg);
	s << "rm -f " << testfile << OUT << ends;
	system (buf);
	
	//save predicted classes into a file
	save_predicted(predicted);
	
	//s.seekp(0,ios::beg);
	//s << "rm -f " << predicted << OUT << ends;
	//system (buf);
	
	return err;
}


///////////////////////////////////////////////////////////////////////////////////
// BaseSVM_light_linear


/**
	 Trains the base learner using a training set stored in a file.
*/
float BaseSVM_light_linear::train_learner (char* trainfile, char* model){
  
	const int sz = 256;
  char buf[sz];	
	float err;
	
	ostrstream s(buf,sz,ios::out);
	s << "svm_learn " << type << "-c " << C << " " << trainfile << " " << model << \
	     " > " << trainfile << OUT << ends;
	system (buf);
	
	s.seekp(0,ios::beg);
	s << trainfile << OUT << ends;
	err = read_train_error(buf);	
	s.seekp(0,ios::beg);
	s << "rm -f " << trainfile << OUT << ends;
	system (buf);
	
	return err;
}			


//////////////////////////////////////////////////////////////////////////////////////////////////

/**
   Wrapper classes for SVM-light polynomial SVM.\\	 
*/ 

// Constructor
BaseSVM_light_poly::BaseSVM_light_poly (int valdegree, float valC, float vals, float valr){
    num=0; 
		C=valC; 
		degree = valdegree;
		s = vals;
		r = valr;
		type = "-t 1 ";
}



/**
	 Trains the base learner using a training set stored in a file.
*/
float BaseSVM_light_poly::train_learner (char* trainfile, char* model){
  
	const int sz = 256;
  char buf[sz];	
	float err;
	
	ostrstream os(buf,sz,ios::out);
	os << "svm_learn " << type << "-c " << C << " " << "-d " << degree << " " <<\
	"-s " << s << " " << "-r " << r << " " << trainfile << " " << model << \
	" > " << trainfile << OUT << ends;
	//cout << buf << endl;
	system (buf);
	
	os.seekp(0,ios::beg);
	os << trainfile << OUT << ends;
	err = read_train_error(buf);	
	os.seekp(0,ios::beg);
	os << "rm -f " << trainfile << OUT << ends;
	system (buf);
	
	return err;
}			



//////////////////////////////////////////////////////////////////////////////////////////////////
/*
   Wrapper classes for SVM-light gaussian SVM.	 
*/ 

// Constructor
BaseSVM_light_gaussian::BaseSVM_light_gaussian (float valgamma, float valC){
    num=0; 
		C=valC; 
		gamma = valgamma;
		type = "-t 2 ";
}



/*
	 Trains the base learner using a training set stored in a file.
*/
float BaseSVM_light_gaussian::train_learner (char* trainfile, char* model){
  
	const int sz = 256;
  char buf[sz];	
	float err;
	
	ostrstream os(buf,sz,ios::out);
	os << "svm_learn " << type << "-c " << C << " " << "-g " << gamma << " " <<\
	trainfile << " " << model << " > " << trainfile << OUT << ends;
	//cout << buf << endl;
	system (buf);
	
	os.seekp(0,ios::beg);
	os << trainfile << OUT << ends;
	err = read_train_error(buf);	
	os.seekp(0,ios::beg);
	os << "rm -f " << trainfile << OUT << ends;
	system (buf);
	
	return err;
}			




//////////////////////////////////////////////////////////////////////////////////////////////////

/**
   Wrapper classes for SVM-light polynomial SVM.\\	 
*/ 

// Constructor
BaseSVM_light_sigmoid::BaseSVM_light_sigmoid (float valC, float vals, float valr){
    num=0; 
		C=valC; 
		s = vals;
		r = valr;
		type = "-t 3 ";
}



/**
	 Trains the base learner using a training set stored in a file.
*/
float BaseSVM_light_sigmoid::train_learner (char* trainfile, char* model){
  
	const int sz = 256;
  char buf[sz];	
	float err;
	
	ostrstream os(buf,sz,ios::out);
	os << "svm_learn " << type << "-c " << C << " "  <<\
	"-s " << s << " " << "-r " << r << " " << trainfile << " " << model << \
	" > " << trainfile << OUT << ends;
	//cout << buf << endl;
	system (buf);
	
	os.seekp(0,ios::beg);
	os << trainfile << OUT << ends;
	err = read_train_error(buf);	
	os.seekp(0,ios::beg);
	os << "rm -f " << trainfile << OUT << ends;
	system (buf);
	
	return err;
}			
