// "bootstraplight.cc"
// November 2002


#include<ctype.h>
#include<stdlib.h>
#include<math.h>
#include<time.h>
#include<iostream.h>
#include<fstream.h>
#include<deque.h>
#include"bootstrapLight.h"
#include"utils.h"

BootstrapLight::BootstrapLight ( char* data, char* base, int nf, int n, int s) {  
	dataset=data;	
	basename=base; 	
	nfiles=nf; 	
	ndata_sample=n;
	read_dataset(dataset);
	ndata = strings.size();
	seed = s;
	if (seed == 0)
	  seed = (int) time(NULL);
  srandom(seed);
	m = new (short*)[nfiles];
	for (int i=0; i<nfiles; i++) 
	  m[i] = new short[ndata];
	for (int i=0; i<nfiles; i++)
	  for (int j=0; j<ndata; j++)
		  m[i][j] = 1;
  write_bootstrap_samples(basename, nfiles);
}

/**
	 Constructor.
	 It reads from file the attributes and the class of the patterns.\\
	 It load only the data without generating any file;
	 @param data   data set to be read
	 @param seed random seed (def. 0). If seed=0 CPU time is used.
	 */
BootstrapLight::BootstrapLight ( char* data, int s) : m(NULL), nfiles(0), ndata_sample(0) {  
	dataset=data;	
	read_dataset(dataset);
	ndata = strings.size();
	cout << ndata << endl;
	for (int i =0; i<ndata; i++)
	  cout << strings[i] << endl; // *****************
	seed = s;
	if (seed == 0)
	  seed = (int) time(NULL);
  srandom(seed);
}

void BootstrapLight::read_dataset(char* dataset)  {
  
	ifstream ifs(dataset);
	if (!ifs)
	  error_exit("Cannot open file ", dataset);			
	// init random generator
	if (seed == 0)
	  seed = (int) time(NULL);
  srandom(seed);
	// read input file
	string line;
	getline(ifs,line); getline(ifs,line); 
	while (getline(ifs,line))
	  if (line != "") 
		  strings.push_back(line);
	if (ifs.bad())
	  error_exit("Error reading file ", dataset);
	ifs.close();
}



int BootstrapLight::unif_rand_generate(int min, int max)
{
  return (int)(min + ((float)random()/RAND_MAX) * max);
}



void BootstrapLight::write_bootstrap_samples(char* basename, int nf) {
  
	const int SIZE = 256;
	char bootstrap_sample[SIZE];
  ofstream ofs;
	ostrstream obasename(bootstrap_sample,SIZE,ios::out);
  for (int i = 1; i<=nf; i++) {
		  obasename.seekp(0,ios::beg);
			obasename << basename << i << ends;
			ofs.open(bootstrap_sample);
			if (!ofs)
	    	error_exit("Cannot open file ", bootstrap_sample); 
			// Heading
    	ofs << "# " << bootstrap_sample << " data file in SVM-light format." << endl;
    	ofs << "# Number of patterns: " << ndata_sample << endl;
			for (int j=0; j<ndata_sample; j++) {
		  	int k = unif_rand_generate(0, ndata);
				m[i-1][k] = 0;
				ofs << strings[k] << endl;
			}
			if (ofs.bad())
		  	error_exit("Error writing in ", bootstrap_sample);
			ofs.close(); 
	}
}


void BootstrapLight::write_splitted_samples(char* basename, int n, float ratio) {
  
	const int SIZE = 256;
	char train[SIZE];
	char test[SIZE];
	
	// container of the indices of the examples of class1 and 2
	deque<int> cl1; // container of the indices of the examples of class1 and 2
	deque<int> cl2;
	
	// Putting indices in the containers
	for (int i=0; i<ndata; i++) 
	  if ((strings[i][0] == '1') ||  (strings[i][0] == '+'))
		  cl1.push_back(i);
		else 
	    cl2.push_back(i);
	
	int n1 = cl1.size();
	int n2 = cl2.size();	
	if ((n1+n2)!=ndata)
	  error_exit("Error in BootstrapLight::write_splitted_samples");
	
	// calculating the number of examples for the training and test set separately for each class	
	int trainsize1 =(int)rint(n1*ratio);
	int trainsize2 =(int)rint(n2*ratio);
	int testsize1 = n1 - trainsize1;
	int testsize2 = n2 - trainsize2;
	// vector for stroing the indices of extracted samples
	short* d1 = new short [n1];
	short* d2 = new short [n2];
	cout << "Data : " << ndata << endl;
	cout << trainsize1 << endl;
	cout << trainsize2 << endl;
	cout << n1 << endl;
	cout << n2 << endl;
	
  ofstream otr;
	ofstream ots;
	ostrstream otrain(train,SIZE,ios::out);
	ostrstream otest(test,SIZE,ios::out);
  // cycling on all the splitted data
	for (int i = 1; i<=n; i++) {
		  // building file names
			otrain.seekp(0,ios::beg);
			otest.seekp(0,ios::beg);
			otrain << basename << i << TRAIN_B << ends;
			otest << basename << i << TEST_B << ends;
			otr.open(train);
			if (!otr)
	    	error_exit("Cannot open file ", train); 
			ots.open(test);
			if (!ots)
	    	error_exit("Cannot open file ", test); 
			// Heading
    	otr << "# " << train << " data file in SVM-light format." << endl;
    	otr << "# Number of patterns: " << trainsize1+trainsize2 << endl;
			// Heading
    	ots << "# " << test << " data file in SVM-light format." << endl;
    	ots << "# Number of patterns: " << testsize1+testsize2 << endl;
			
			// clearing d1 and d2
			for (int i=0; i<n1; i++)
	      d1[i] = 0;
	    for (int i=0; i<n2; i++)
	      d2[i] = 0;		
			// balanced uniform sampling between classes
			int j=0;
			while (j<trainsize1) {
			  int k = unif_rand_generate(0, n1);
				if (d1[k] == 0) {
				  d1[k] = 1;
					j++;
				}
			}
			j=0;
			while (j<trainsize2) {
			  int k = unif_rand_generate(0, n2);
				if (d2[k] == 0) {
				  d2[k] = 1;
					j++;
				}
			}
			
			// writing files
			for (int j = 0; j<n1; j++) 
			  if (d1[j] == 1)
				  otr << strings[cl1[j]] << endl;
				else
				  ots << strings[cl1[j]] << endl;
			for (int j = 0; j<n2; j++) 
			  if (d2[j] == 1)
				  otr << strings[cl2[j]] << endl;
				else
				  ots << strings[cl2[j]] << endl;
						
			// closing files
			if (otr.bad())
		  	error_exit("Error writing in ", train);
			otr.close(); 
			if (ots.bad())
		  	error_exit("Error writing in ", test);
			ots.close();
			// prepapre the labels file 
			extract_labels_light(test);
	}
	
	delete[] d1;
	delete[] d2;
}


void BootstrapLight::write_folded_samples(char* basename, int k) {
  
	const int SIZE = 256;
	char train[SIZE];
	char test[SIZE];
	
	// container of the indices of the examples of class1 and 2
	deque<int> cl1; // container of the indices of the examples of class1 and 2
	deque<int> cl2;
	
	// Putting indices in the containers
	for (int i=0; i<ndata; i++) 
	  if ((strings[i][0] == '1') ||  (strings[i][0] == '+'))
		  cl1.push_back(i);
		else 
	    cl2.push_back(i);
	
	int n1 = cl1.size();
	int n2 = cl2.size();	
	if ((n1+n2)!=ndata)
	  error_exit("Error in BootstrapLight::write_splitted_samples");
	
	// calculating the number of examples for the training and test set separately for each class	
	int testsize1 =(int)rint(n1/(float)k);
	int testsize2 =(int)rint(n2/(float)k);
	int trainsize1 = n1 - testsize1;
	int trainsize2 = n2 - testsize2;
	// vector for stroing the indices of extracted samples
	short* d1 = new short [n1];
	short* d2 = new short [n2];
	// clearing d1 and d2
	for (int i=0; i<n1; i++)
	  d1[i] = 0;
	for (int i=0; i<n2; i++)
	  d2[i] = 0;		
	cout << "Data : " << ndata << endl;
	cout << testsize1 << endl;
	cout << testsize2 << endl;
	cout << n1 << endl;
	cout << n2 << endl;
	
  ofstream otr;
	ofstream ots;
	ostrstream otrain(train,SIZE,ios::out);
	ostrstream otest(test,SIZE,ios::out);
  // cycling on all the splitted data
	for (int i = 1; i<k; i++) {
		  // building file names
			otrain.seekp(0,ios::beg);
			otest.seekp(0,ios::beg);
			otrain << basename << i << TRAIN_B << ends;
			otest << basename << i << TEST_B << ends;
			otr.open(train);
			if (!otr)
	    	error_exit("Cannot open file ", train); 
			ots.open(test);
			if (!ots)
	    	error_exit("Cannot open file ", test); 
			// Heading
    	otr << "# " << train << " data file in SVM-light format." << endl;
    	otr << "# Number of patterns: " << trainsize1+trainsize2 << endl;
			// Heading
    	ots << "# " << test << " data file in SVM-light format." << endl;
    	ots << "# Number of patterns: " << testsize1+testsize2 << endl;
						
			// balanced uniform sampling between classes
			int j=0;
			while (j<testsize1) {
			  int k = unif_rand_generate(0, n1);
				if (d1[k] == 0) {
				  d1[k] = i;
					j++;
				}
			}
			j=0;
			while (j<testsize1) {
			  int k = unif_rand_generate(0, n2);
				if (d2[k] == 0) {
				  d2[k] = i;
					j++;
				}
			}
			
			// writing files
			for (int j = 0; j<n1; j++) 
			  if (d1[j] < i)
				  otr << strings[cl1[j]] << endl;
				else
				  ots << strings[cl1[j]] << endl;
			for (int j = 0; j<n2; j++) 
			  if (d2[j] < i)
				  otr << strings[cl2[j]] << endl;
				else
				  ots << strings[cl2[j]] << endl;
						
			// closing files
			if (otr.bad())
		  	error_exit("Error writing in ", train);
			otr.close(); 
			if (ots.bad())
		  	error_exit("Error writing in ", test);
			ots.close(); 
			// prepare the labels file 
			extract_labels_light(test);
	}
	
	// last fold
	otrain.seekp(0,ios::beg);
	otest.seekp(0,ios::beg);
	otrain << basename << k << TRAIN_B << ends;
	otest << basename << k << TEST_B << ends;
	otr.open(train);
	if (!otr)
	  error_exit("Cannot open file ", train);
	ots.open(test);
	if (!ots)
	  error_exit("Cannot open file ", test);
	// Heading
  otr << "# " << train << " data file in SVM-light format." << endl;
  otr << "# Number of patterns: " << (testsize1+testsize2)*(k-1) << endl;
	// Heading
  ots << "# " << test << " data file in SVM-light format." << endl;
  ots << "# Number of patterns: " << ndata - (testsize1+testsize2)*(k-1) << endl;
	
	// writing files
	for (int j = 0; j<n1; j++) 
		if (d1[j] != 0)
			otr << strings[cl1[j]] << endl;
		else
			ots << strings[cl1[j]] << endl;
	for (int j = 0; j<n2; j++) 
		if (d2[j] != 0)
			otr << strings[cl2[j]] << endl;
		else
			ots << strings[cl2[j]] << endl;

	// closing files
	if (otr.bad())
		error_exit("Error writing in ", train);
	otr.close(); 
	if (ots.bad())
		error_exit("Error writing in ", test);
	ots.close(); 

	delete[] d1;
	delete[] d2;
}






BootstrapLight::~BootstrapLight() {
  for (int i=0; i<nfiles; i++)
		delete [] m[i];
	delete [] m;
}

void BootstrapLight::save_m(char* file) {
  
	ofstream ofs(file);
	if (!ofs)
	  error_exit("Cannot create ", file);  
	for (int i=0; i<nfiles; i++) {
	  for (int j=0; j<ndata; j++)
		  ofs << m[i][j] << " ";
		ofs << endl;
	}	
	if (ofs.bad())
	  error_exit("Error writing ", file);  		
}

/**
   Extract and convert class labels fron light data file.\\
	 It read from a light file the class labels, converts them in NEURObjects format and stores them
	 in a file lightdata.labels
	 @param lightdata light file from which the labels are extracted
*/
void BootstrapLight::extract_labels_light (char* lightdata){
  
	const int SIZE = 256;
  char dataset[SIZE];
	int cl; 
		
	ifstream ifs(lightdata);
	if (!ifs)
	  error_exit("Cannot open file ", lightdata);
	strcpy(dataset,lightdata);
	strcat(dataset,".labels");
	ofstream ofs(dataset);
	if (!ofs)
	  error_exit("Cannot open file ", dataset); 		
	string line;
	// getting away the first 2 lines
	getline(ifs,line);
	getline(ifs,line);
	ifs >> cl; 
	while (!ifs.eof()) {	  
		if (cl != 1)
		  cl = 2;
		ofs << cl << endl;	
	  getline(ifs,line);
		ifs >> cl; 
	}	 		
	if (ifs.bad())
	  error_exit("Error reading file ", lightdata);
	if (ofs.bad())
	  error_exit("Error writing file ", dataset);
}

