//cdt2light.cc
// Application for translating a .cdt file in a SVM-light readable file
#include<iostream.h>
#include<fstream.h>
#include<stdio.h>
#include<time.h>
#include<string.h>
#include<stdlib.h>
#include<math.h>

const char TAB = '\t';
const int DATA_SIZE = 8;
const int DIMBUF = 1024;
const float MISSING = -100000;

const int numexp = 96;
const int numdata = 4026;

// matrix of the DNA microarray values. Each line corresponds to a pattern (array exp),
// each column to a gene expression
float values[numexp][numdata];
// array of the label of the classes. classes[i] is the class of the pattern values[i][]
int classes [numexp];

void print_info_use(void)
{
  cerr << "
  cdt2light. Application for translating a .cdt file 
	in a SVM-light readable filea.
  Usage: cdt2light cdtfile pat-classfile lightfile
         cdtfile         : name of the .cdt file
         pat-classfile   : name of the file containing the label of the classes.
                           Each line corresponds to a column of the .cdt file.
         lightfile       : name of the SVM-light file" << endl;
  exit(1);
}
  

void save_pattern (ofstream& fdata, float * input, unsigned dim, int cl)
{
  fdata.setf(ios::left, ios::adjustfield);
  fdata.width(5); 
  fdata << cl;
  for (int j=0; j<dim; j++) {       
    fdata << j+1 << ":";
    fdata.width(DATA_SIZE);
    fdata << input[j];
  }
  fdata << endl;
}

// It computes p1 and M
void consume_header(ifstream& in)
{
  int i;
  char buf[DIMBUF];
  
  i=0;
	// read the first 4 tabs
	while (i < 4) {
	  in.getline(buf,DIMBUF,TAB);
	  i++;
	}
	if (in.bad()) {
	   cerr << "Error reading file" << endl;
		 exit(1);
	}
}
	  
  
// read a value from .cdt file
float read_value(ifstream& in)
{
  char buf[DIMBUF];
	float num;
	
	in.getline(buf,DIMBUF,TAB);
	if (in.bad()) {
	   cerr << "Error reading file" << endl;
		 exit(1);
	} 
	if (strlen(buf) == 0)
	  num = MISSING;
	else
	  num = atof(buf);
	return num;
}
	  
// read the jth line of the .cdt.file	
void read_line(ifstream& in, int j, int numvalue=numexp)
{
  char buf[DIMBUF];
	int i;
	
	for (i=0; i<numvalue-1; i++) 
	  values[i][j] = read_value(in);
	// the last value is not termined by a TAB
	in.getline(buf,DIMBUF);
	values[i][j]= atof(buf);
}

// read the classes corresponding to the patterns from a stream
void load_classes(ifstream& in, int numvalue=numexp)
{
  char buf[DIMBUF];
	int i;
	
	for (i=0; i<numvalue; i++) {
	  in >> classes[i];
		in.getline(buf,DIMBUF);
	}
  if (in.bad()) {
	   cerr << "Error reading file" << endl;
		 exit(1);
	} 
}


void load_values(ifstream& in)
{
  int i,j;
	
	for (j=0; j<numdata; j++) {
	  consume_header(in);
	  read_line(in, j, numexp);
  }
}


void write_values(ofstream& out)
{
  int i,j;
	float val;
	
	out.setf(ios::left, ios::adjustfield);
	for (i=0; i<numexp; i++) {
	  out.width(5);
		out << classes[i];
	  for (j=0; j<numdata; j++) {
		  out << j+1 << ":";
		  val=values[i][j];
			out.width(DATA_SIZE);
		  if (val==MISSING)
		    out << 0;
		  else
		    out << val;
		  }
	  out << endl;
	  if (out.bad()){
	     cerr << "Error writing file" << endl;
		   exit(1);}
	}
}


  
// **************************************
int main(int argc, char* argv[])
{
   
	char cdtfilename[DIMBUF];
	char patternclassfilename[DIMBUF];
  char lightfilename[DIMBUF]; 
  
	if (argc != 4) 
     print_info_use();
     
  // reading arguments
  strcpy(cdtfilename, argv[1]);
	strcpy(patternclassfilename, argv[2]);
  strcpy(lightfilename, argv[3]);
	
	// opening files
	ifstream in(cdtfilename);
	if (!in) {
	  cerr << "Cannot open file " << cdtfilename << endl;
		exit(1);
	}
	ifstream inclasses(patternclassfilename);
	if (!inclasses) {
	  cerr << "Cannot open file " << patternclassfilename << endl;
		exit(1);
	}
	ofstream out(lightfilename);
	if (!out) {
	  cerr << "Cannot open file " << lightfilename << endl;
		exit(1);
	}
	
	// reading data
	load_classes(inclasses);
	load_values(in);
	
	// writing header of data
  out << "# " << argv[3] << " data file in SVM-light format." << endl;
  out << "# Number of patterns: " << numexp << endl;
	
	// writing data
	write_values(out);
  
  return 0;
}
  
