// "dico.cc"


#include<time.h>
#include<math.h>
#include<fstream.h>
#include"dico.h"


static const unsigned DIMBUFNAMEFILE = 64;  
static const unsigned DIMBUF = 256;
Dico1::Dico1(unsigned nin, vect& vectclass, act_func fout)
:  OneLayer(1, nin, fout), vectorclass(vectclass.dim())
{
  Do_vectorclass(vectclass);
}
void Dico1::Do_vectorclass(vect& vectclass)
{
  unsigned i,type_fun;
  unsigned nclass = vectclass.dim();
  type_fun = Code_act_fun(layer[OUTPUT].Read_g());
  switch (type_fun)  {
    case SIGMOID:
           for (i=0; i<nclass; i++)
             switch((int)vectclass[i]) {
               case d_OPC::HIGH:   vectorclass[i] = SIG_ON; break;
               case d_OPC::LOW:    vectorclass[i] = SIG_OFF; break;
               case d_OPC::IGNORE: vectorclass[i] = IGNORE_CLASS; break;
             }              
           break;
    case TANHYP:
           for (i=0; i<nclass; i++)
             switch((int)vectclass[i]) {
               case d_OPC::HIGH:   vectorclass[i] = TANH_ON; break;
               case d_OPC::LOW:    vectorclass[i] = TANH_OFF; break;
               case d_OPC::IGNORE: vectorclass[i] = IGNORE_CLASS; break;
             }  
           break;
    case IDENTITY:
           for (i=0; i<nclass; i++)
             switch((int)vectclass[i]) {
               case d_OPC::HIGH:   vectorclass[i] = ID_ON; break;
               case d_OPC::LOW:    vectorclass[i] = ID_OFF; break;
               case d_OPC::IGNORE: vectorclass[i] = IGNORE_CLASS; break;
             }  
           break;
    default:
           error ("Not known activation function");
           break;
  }  
}
void Dico1::init_weights_norm(int seed)
{ 
  layer[OUTPUT].init_norm_weights(seed);
  Dico1::seed = seed;
}
stop_cond Dico1::Learn_by_pattern(TrainingSet& t_set, Learning& learn,
                                          unsigned& iter, double& error,
                                          double errtol, unsigned maxits, bool printout)
{
  unsigned i;
  unsigned num_train = t_set.Ntrain();
  vect& t = t_set.Read_target();  
  matrix& x = t_set.Read_data();    
  double denom = calc_denom_RMS(t, num_train);  
  vect target(1);  
  Zero(n_layers);  
  for (iter = 0; iter < maxits; iter++) {
    error = 0;
    for (i = 0; i < num_train; i++) {
      target[0] = vectorclass[((int)t[i])-1];
      vect& input = x.row(i);
      vect& out = compute(input);
      backprop(target, input);  
      learn.Update_delta_weights(n_layers, layer);
      learn.Update_weigths (n_layers, layer);      
      error += Calc_square_error(target, out);
    }
    error = RMS_error(error, denom);   
    Qerror.push((float)error);
    if (printout) cout << "." << flush;
    if (error <= errtol)
      {   if (printout) cout << endl << flush;
          return STOP_ERR_TOL;}
    learn.Update_param(error, iter);  
  }
  if (printout) cout << endl << flush;
  return STOP_NUM_ITERAZ;
}
stop_cond Dico1::Learn_by_epoch(TrainingSet& t_set, Learning& learn,
                                          unsigned& iter, double& error,
                                          double errtol, unsigned maxits,
																					bool printout)
{
  unsigned i;
  matrix*  Sum_Delta_W[MAXLAYERS];   
  vect*  Sum_delta_bias[MAXLAYERS];   
  unsigned num_train = t_set.Ntrain();
  vect& t = t_set.Read_target();  
  matrix& x = t_set.Read_data();    
  double denom = calc_denom_RMS(t, num_train);  
  vect target(1);  
  Build_Sum_Delta(Sum_Delta_W, Sum_delta_bias);
  Zero(n_layers);  
  do_Sum_deltas(n_layers,Sum_Delta_W,Sum_delta_bias);
  for (iter = 0; iter < maxits; iter++) {
    error = 0;
    Zero_Sum_Delta(n_layers, Sum_Delta_W, Sum_delta_bias);  
    for (i = 0; i < num_train; i++) {
      target[0] = vectorclass[((int)t[i])-1];
      vect& input = x.row(i);
      vect& out = compute(input);
      backprop(target, input);  
      learn.Update_delta_weights(n_layers, layer);
      Update_Sum_Delta(n_layers, Sum_Delta_W, Sum_delta_bias); 
      error += Calc_square_error(target, out);
    }
    Copy_Sum_Delta_To_Delta(n_layers, layer, Sum_Delta_W, Sum_delta_bias);
    learn.Update_weigths (n_layers, layer);   
    error = RMS_error(error, denom);   
    Qerror.push((float)error);
    if (printout) cout << "." << flush;
    if (error <= errtol) { 
      if (printout) cout << endl << flush; 
      Delete_Sum_Delta(Sum_Delta_W, Sum_delta_bias); return STOP_ERR_TOL;
    }
    learn.Update_param(error, iter);  
  }
  Delete_Sum_Delta(Sum_Delta_W, Sum_delta_bias);
  if (printout) cout << endl << flush;
  return STOP_NUM_ITERAZ;
}
double Dico1::Calc_square_error(vect& t, vect& out)
{
  double x = out[0] - t[0]; 
  return (x*x);
}
double Dico1::calc_denom_RMS(vect& t, unsigned n_train)
{
  int i,j;
  double x;
  double denom = 0;
  double t_mean;
  double target;
  for (i=0; i < n_train; i++)
    t_mean += vectorclass[((int)t[i])-1];
  t_mean /= n_train;
  for (i=0; i<n_train; i++)  {
    target = vectorclass[((int)t[i])-1];    
    x = t_mean - target;
    denom += (x*x);
  } 
  return denom;
} 
unsigned Dico1::test(TrainingSet& tset)
{
  unsigned i;
  unsigned num_train = tset.Ntrain();
  vect& t = tset.Read_target();  
  matrix& x = tset.Read_data();     
  double treshold;
  num_errors = 0;
  treshold = Calc_treshold();
  double on, off;
  analyze (layer[OUTPUT].Read_g(), on, off);
  for (i = 0; i < num_train; i++) {
    vect& input = x.row(i);
    vect& out = compute(input);
    if (bin_class(out, vectorclass[((int)t[i])-1], treshold) == NOTOK)
       { num_errors++;}    
  }
  perc_errors = (double)num_errors/(double)num_train;
  return num_errors;
}
double Dico1::Calc_treshold(void)
{
  double treshold;
  unsigned typefun = Code_act_fun(layer[OUTPUT].Read_g());
  if (typefun == SIGMOID) 
    { treshold = SIG_TRESHOLD;} 
  else if (typefun == TANHYP)
    {treshold = TANH_TRESHOLD;} 
  else if (typefun == IDENTITY)
    {treshold = ID_TRESHOLD;} 
  else
    error ("Not known activation function");
  return treshold;
}
int Dico1::bin_class(vect& out, double target, double treshold)
{
  binary_classification bc;
  if ((out[0]-treshold) >=0)  
    bc = (target >= treshold) ? OK : NOTOK;
  else    
    bc = (target < treshold) ? OK : NOTOK;
  return bc;
}
void Dico1::save_weights(char* namefile)
{
  char namebuf[DIMBUFNAMEFILE];
  strcpy(namebuf, namefile);
  strcat(namebuf, NET_EXT);
  ofstream f(namebuf);
  if (!f) error("Cannot open file ", namebuf);
  f.setf(ios::left, ios::adjustfield);
  f << "Version:  " << 1 << endl;
  f << namebuf << endl;
  save_weights(f);
  f << END << flush;
  if (f.bad())
    error("Error writing file ", namebuf);
}
void Dico1::save_weights(ofstream& f)
{
  unsigned fun_out;
  fun_out = Code_act_fun(layer[OUTPUT].Read_g());
  f.setf(ios::left, ios::adjustfield);
  f.width(INTEST_SIZE);
  f << "Number_layers: " << 1 << endl;
  f.width(INTEST_SIZE);
  f << "Dim_output: " << n_outputs << "    ";
  f << "Act.function: " << fun_out << endl;
  f.width(INTEST_SIZE);
  f << "Dim_input: " << n_inputs << endl << endl;
  f << "Weight matrix output layer :" << endl;
  layer[OUTPUT].save_weights(f);
  f << flush;
}
Dico2::Dico2(unsigned nhidden, unsigned nin, vect& vectclass, 
                                         act_func fout, act_func fhid1)
:  Dico1(nhidden, vectclass, fout)
{
  n_layers = 2;
  n_inputs = nin;
  n_hidden1 = nhidden;
  layer[HIDDEN1] = LayerNetTrain(nhidden, nin, fhid1);
}
void Dico2::init_weights_norm(int seed)
{ 
  layer[OUTPUT].init_norm_weights(seed);
  layer[HIDDEN1].init_norm_weights(seed);
  Dico1::seed = seed;
}
void Dico2::init_weights_const(double x)
{
  layer[OUTPUT].init_weights(x);
  layer[HIDDEN1].init_weights(x);
} 
void Dico2::print_weights(void)
{
  cout << "Weight matrix Two-Layer dichotomizer :" << endl;
  cout << "Output layer :" << endl;
  layer[OUTPUT].print_weights();
  cout << "Hidden layer :" << endl;
  layer[HIDDEN1].print_weights();
} 
vect& Dico2::compute (vect& vinput)
{
  vect& hidout = layer[HIDDEN1].compute(vinput);  
  return (layer[OUTPUT].compute (hidout));   
}
void Dico2::backprop(vect& t, vect& x)
{
   layer[OUTPUT].calc_grad(t, layer[HIDDEN1].Read_out());  
   layer[HIDDEN1].backprop(layer[OUTPUT].Read_delta(), layer[OUTPUT].Read_W(), x);
}
void Dico2::print_info_grad(void)
{
  int i;
  unsigned n_layers = Nlayers();
  cout << "Two-Layer dichotomizer : Delta and Gradient values. " << endl;
  for (i=0; i<n_layers; i++)  {
    if (i == 0)
      cout << "Output layer : " << endl;
    else
      cout << "Hidden layer : " << endl;
    layer[i].print_delta();
    layer[i].print_grad();
    layer[i].print_grad_bias();
  }
} 
void Dico2::save_weights(ofstream& f)
{
  unsigned fun_out, fun_hid1;
  fun_out = Code_act_fun(layer[OUTPUT].Read_g());
  fun_hid1 = Code_act_fun(layer[HIDDEN1].Read_g());
  f.setf(ios::left, ios::adjustfield);
  f.width(INTEST_SIZE);
  f << "Number_layers: " << 2 << endl;
  f.width(INTEST_SIZE);
  f << "Dim_output: " << n_outputs << "    ";
  f << "Act.function: " << fun_out << endl;
  f.width(INTEST_SIZE);
  f << "Dim_hidden1: " << n_hidden1 << "    ";
  f << "Act.function: " << fun_hid1 << endl;
  f.width(INTEST_SIZE);
  f << "Dim_input: " << n_inputs << endl << endl;
  f << "Weight matrix output layer :" << endl;
  layer[OUTPUT].save_weights(f);
  f << "Weight matrix hidden layer :" << endl;
  layer[HIDDEN1].save_weights(f); 
  f << flush;
}
Dico3::Dico3(unsigned nhidden1, unsigned nhidden2, unsigned nin, vect& vectclass, 
       act_func fout, act_func fhid1, act_func fhid2)
: Dico2(nhidden1, nhidden2, vectclass, fout, fhid1)
{
  n_layers = 3;
  n_hidden2 = nhidden2;
  n_inputs = nin;
  layer[HIDDEN2] = LayerNetTrain(nhidden2, nin, fhid2);
}   
void Dico3::init_weights_norm(int seed)
{ 
  Dico2::init_weights_norm(seed);
  layer[HIDDEN2].init_norm_weights(seed);
}
void Dico3::init_weights_const(double x)
{
  Dico2::init_weights_const(x);
  layer[HIDDEN2].init_weights(x);
} 
void Dico3::print_weights(void)
{
  cout << "Weight matrix of Three-Layer dichotomizer :" << endl;
  cout << "Output layer :" << endl;
  layer[OUTPUT].print_weights();
  cout << "I hidden layer :" << endl;
  layer[HIDDEN1].print_weights();
  cout << "II hidden layer :" << endl;
  layer[HIDDEN2].print_weights();
} 
vect& Dico3::compute (vect& vinput)
{
  vect& hidout2 = layer[HIDDEN2].compute(vinput);  
  vect& hidout1 = layer[HIDDEN1].compute(hidout2);  
  return (layer[OUTPUT].compute (hidout1));   
}
void Dico3::backprop(vect& t, vect& x)
{
   layer[OUTPUT].calc_grad(t, layer[HIDDEN1].Read_out());  
   layer[HIDDEN1].backprop(layer[OUTPUT].Read_delta(), layer[OUTPUT].Read_W(), layer[HIDDEN2].Read_out()); 
   layer[HIDDEN2].backprop(layer[HIDDEN1].Read_delta(), layer[HIDDEN1].Read_W(), x); 
}
void Dico3::print_info_grad(void)
{
  int i;
  unsigned n_layers = Nlayers();
  cout << "Three-Layer dichotomizer : Delta and gradient values. " << endl;
  for (i=0; i<n_layers; i++)  {
    switch(i) {
       case 0 : cout << "Output layer: " << endl; break;
       case 1 : cout << "I hidden layer : " << endl; break;
       case 2 : cout << "II hidden layer : " << endl; break;
       default : error ("ThreeLayer::print_info_grad : error"); break;
    }    
    layer[i].print_delta();
    layer[i].print_grad();
    layer[i].print_grad_bias();
  }
} 
void Dico3::save_weights(ofstream& f)
{ 
  unsigned fun_out, fun_hid1, fun_hid2;
  fun_out = Code_act_fun(layer[OUTPUT].Read_g());
  fun_hid1 = Code_act_fun(layer[HIDDEN1].Read_g());
  fun_hid2 = Code_act_fun(layer[HIDDEN2].Read_g());
  f.setf(ios::left, ios::adjustfield);
  f.width(INTEST_SIZE);
  f << "Number_layers: " << 3 << endl;
  f.width(INTEST_SIZE);
  f << "Dim_output: " << n_outputs << "    ";
  f << "Act.function: " << fun_out << endl;
  f.width(INTEST_SIZE);
  f << "Dim_hidden1: " << n_hidden1 << "    ";
  f << "Act.function: " << fun_hid1 << endl;
  f.width(INTEST_SIZE);
  f << "Dim_hidden2: " << n_hidden2 << "    ";
  f << "Act.function: " << fun_hid2 << endl;
  f.width(INTEST_SIZE);
  f << "Dim_input: " << n_inputs << endl << endl;
  f << "Weight matrix output layer :" << endl;
  layer[OUTPUT].save_weights(f);
  f << "Weight matrix first hidden layer :" << endl;
  layer[HIDDEN1].save_weights(f);
  f << "Weight matrix second hidden layer :" << endl;
  layer[HIDDEN2].save_weights(f);
  f << flush;
}
