// "dodata.cc"


#include<stdlib.h>
#include<iostream.h>
#include"utils.h"
#include"tset.h"
#include"patgen.h"


#define NUMTYPE 8


static char* typecode[] = {"ranspheric",
                           "rs",
                           "ranvariable",
                           "rv",
                           "readinfo",
                           "ri",
                           "readinfopatt",
                           "rip"};
static unsigned NCLASS = 2;  
static unsigned NPATTERN = 100;   
static unsigned DIM = 2;    
static double RANGEMIN = 1.0;  
static double RANGEMAX = 9.0;   
static double SIGMA = 1.0;      
static unsigned NPATMIN = 20;  
static unsigned NPATMAX = 100;  
static double SIGMIN = 0.5;    
static double SIGMAX = 2.0;    
static const int DIMBUF = 128;
static const int SMALLBUF = 32;
static char buf[DIMBUF];
static char type[SMALLBUF];  
static char namefile[DIMBUF];  
static char fileinfo[DIMBUF];  
static unsigned nclass = NCLASS;   
static unsigned npattern = NPATTERN;   
static unsigned dim = DIM;    
static int typenum;     
static double rangemin = RANGEMIN;  
static double rangemax = RANGEMAX;   
static double sigma = SIGMA;      
static unsigned npatmin = NPATMIN;  
static unsigned npatmax = NPATMAX;  
static double sigmin = SIGMIN;    
static double sigmax = SIGMAX;    
void print_info_use(void);
int Analyze_type(char* type);
int doarg_rs(char* arg[], int i);
int doarg_rv(char* arg[], int i);
int doarg_ri(char* arg[], int i);
int doarg_rip(char* arg[], int i);
int main(int argc, char* argv[])
{
  int i;
  if (argc < 4) print_info_use();
  strcpy(buf, argv[1]);
  if (strcmp(buf, "-t")) print_info_use();
  strcpy(type, argv[2]);
  if ((typenum=Analyze_type(type)) >= 8 ) print_info_use();
  strcpy(namefile, argv[3]);
  Patgen patgen;
  switch (typenum)  {
    case 0: case 1: 
       for (i=4; i<argc; i+=2)
         if (doarg_rs(argv, i))
            print_info_use();
       patgen.create_ran_spheric(namefile, nclass, npattern, dim, rangemin, rangemax, sigma);
       break;
    case 2: case 3: 
       for (i=4; i<argc; i+=2)
         if (doarg_rv(argv, i))
            print_info_use();
       patgen.create_ran_variable(namefile, nclass, npatmin, npatmax, dim, 
                                  rangemin, rangemax, sigmin, sigmax);
       break;
    case 4: case 5:
       if (argc!=6) print_info_use();
       if (doarg_ri(argv, 4)) print_info_use();
       patgen.create_reading_info(fileinfo, namefile);                              
       break;
    case 6: case 7:
       if (argc<6) print_info_use();
       for (i=4; i<argc; i+=2)
         if (doarg_rip(argv, i))
            print_info_use();
       patgen.create_reading_info_num_pattern(fileinfo, namefile, npatmin, npatmax);                               
       break;
    default:
       cout << "Tipo non riconosciuto" << endl;
  }
  return 0;
}


void print_info_use(void)
{
  cerr << "
  dodata. NEURObjects application for synthetic data generation.
  Usage: dodata -t type datafile [options]
  type = { ranspheric | rs | ranvariable | rv | readinfo | ri | readinfopatt | rip } 
  datafile (string) : name of the file generated by the application.
	 
  Options:
  The options vary according to the selected generation mode:  
 
 1.ranspheric (rs) option:  The clusters are randomly distributed, one for each 
                            class according to spheric normal distribution  
 -nc unsigned        number of the classes (default=2)
 -d  unsigned        number of attributes (dimension of generated patterns) (default=2)
 -np unsigned        number of patterns (default=100) 
 -rmin double        minimum value for the centers of clusters (default=1.0)
 -rmax double        maximum value for the centers of clusters (default=9.0)
 -sigma double       standard deviation (default=1.0) 
                                         
 2.ranvariable (rv) option:   The clusters are randomly distributed, one for each 
                              class according to a normal distribution with different
                              standard deviation and number of patterns  
 -nc unsigned        number of the classes (default=2)
 -d  unsigned        number of attributes (dimension of generated patterns) (default=2) 
 -pmin unsigned      minimum number of patterns for each class (default=20)
 -pmax unsigned      maximum number of patterns for each class (default=100) 
 -rmin double        minimum value for the centers of clusters (default=1.0)
 -rmax double        maximum value for the centers of clusters (default=9.0)
 -sigmin double      minimum allowed standard deviation (default=0.5)
 -sigmax double      maximum allowed standard deviation (default=2.0) 
                                          
 3.readinfo (ri) option:   The clusters are normal distributed, 
                           with parameters read from a file 
 -ninfo string       file storing the parameters
                                         
 4.readinfopatt (rip) option:   The clusters are normal distributed, with parameters 
                                read from a file and with random numbers of pattern per class 
 -ninfo string       file storing the parameters
 -pmin unsigned      minimum number of patterns for each class (default=20)
 -pmax unsigned      maximum number of patterns for each class (default=100)
  " << endl;
  exit(1);
}

int Analyze_type(char* type)
{
  unsigned i=0;
  while ((i<8) && (strcmp(type, typecode[i]))) i++;
  return i;
} 
int doarg_rs(char* arg[], int i)
{
  if (!strcmp(arg[i], "-nc"))
    nclass = atoi(arg[i+1]);
  else if (!strcmp(arg[i], "-np"))
    npattern = atoi(arg[i+1]);
  else if (!strcmp(arg[i], "-d"))
    dim = atoi(arg[i+1]);
  else if (!strcmp(arg[i], "-rmin"))
    rangemin = atof(arg[i+1]);
  else if (!strcmp(arg[i], "-rmax"))
    rangemax = atof(arg[i+1]);
  else if (!strcmp(arg[i], "-sigma"))
    sigma = atof(arg[i+1]);
  else 
    return 1;
  return 0;
}  
int doarg_rv(char* arg[], int i)
{
  if (!strcmp(arg[i], "-nc"))
    nclass = atoi(arg[i+1]);
  else if (!strcmp(arg[i], "-pmin"))
    npatmin = atoi(arg[i+1]);
  else if (!strcmp(arg[i], "-pmax"))
    npatmax = atoi(arg[i+1]);
  else if (!strcmp(arg[i], "-d"))
    dim = atoi(arg[i+1]);
  else if (!strcmp(arg[i], "-rmin"))
    rangemin = atof(arg[i+1]);
  else if (!strcmp(arg[i], "-rmax"))
    rangemax = atof(arg[i+1]);
  else if (!strcmp(arg[i], "-sigmin"))
    sigmin = atof(arg[i+1]);
  else if (!strcmp(arg[i], "-sigmax"))
    sigmax = atof(arg[i+1]);
  else 
    return 1;
  return 0;
}    
int doarg_ri(char* arg[], int i)
{
  if (!strcmp(arg[i], "-ninfo"))
    strcpy(fileinfo,arg[i+1]);
  else 
    return 1;
  return 0;
}    
int doarg_rip(char* arg[], int i)
{  
  if (!strcmp(arg[i], "-ninfo"))
    strcpy(fileinfo,arg[i+1]);
  else if (!strcmp(arg[i], "-pmin"))
    npatmin = atoi(arg[i+1]);
  else if (!strcmp(arg[i], "-pmax"))
    npatmax = atoi(arg[i+1]);
  else 
    return 1;
  return 0;
}    
