// "layer.cc"


#include<time.h>
#include<math.h>
#include<iostream.h>
#include<fstream.h>
#include"layer.h"
#include"utils.h"


#define DIMBUF 256

static char buf[DIMBUF];  
LayerNet::LayerNet (unsigned num_outputs, unsigned num_inputs, act_func f)
: W(num_outputs, num_inputs), out(num_outputs), a(num_outputs), bias(num_outputs)
{
  n_inputs = num_inputs;
  n_outputs = num_outputs;
  g = f;
  dg = do_deriv(f);
}
vect& LayerNet::compute (vect& vinput)
{
  if (vinput.dim() != W.dim2())
    error("Mismatch dimension matrix vector in  LayerNet::compute");
  a = W * vinput + bias;
  for (int i=0; i<n_outputs; i++)
    out[i] = g(a[i]);
  return out;
}
void LayerNet::print_weights()
{
  cout << "Weight matrix of the layer : ";
  W.print();
  cout << "Bias vector of the layer : " << endl;
  bias.print(); cout << endl;
}
void LayerNet::print_out()
{
  cout << "Vector computed by the layer : " << endl;
  out.print();  
}
void LayerNet::print_dim() const
{
  cout << "Layer dimensions : " << endl;  
  cout << "Input dimension = " << Read_ninputs() << endl;
  cout << "Output dimension = " << Read_noutputs() << endl;
}
void LayerNet::init_weights(double x)
{
  int rows = W.dim1();
  int cols = W.dim2();
  for( int i=0; i<rows; i++)
   {
     bias[i] = x;
     for( int j=0; j<cols; j++)
       W(i,j) = x;
   }
}
void LayerNet::init_norm_weights(long seed)
{
  if (seed == 0)
     seed = (long) -time(NULL);
  int i,j;
  double sigma = 1/sqrt((float)n_inputs);
  for(i=0; i<n_outputs; i++)
    {
      bias[i] = gasdev(&seed) * sigma;
      for(j=0; j<n_inputs; j++)
         W(i,j) = gasdev(&seed) * sigma;
    }
}
act_func LayerNet::do_deriv (act_func f)
{
  act_func df;
  if (f==sigmoid) 
    df = d_sigmoid; 
  else if (f==tanhyp) 
    df = d_tanhyp;
  else if (f==NO_identity)  
    df = NO_d_identity; 
  else
    error ("Not known activation function");
  return df;
}
void LayerNet::load_weights(istream& f)
{
  int i, j;
  eatwhite(f);   
  f.get(buf, 256 );
  for (i=0; i<n_outputs; i++)  {
    for (j=0; j<n_inputs; j++) 
      f >> W(i,j); 
  } 
  eatwhite(f);  
  f.get(buf, 256 );
  for (i=0; i<n_outputs; i++)
    f >> bias[i];
  if (f.bad())
    error("Erro reading weight file ");
}
LayerNetTrain::LayerNetTrain (unsigned num_outputs, unsigned num_inputs, act_func f)
: LayerNet (num_outputs, num_inputs, f), delta(num_outputs), grad(num_outputs, num_inputs),
  DeltaW(num_outputs, num_inputs), grad_bias(num_outputs), Delta_bias(num_outputs)
{}
vect& LayerNetTrain::calc_output_delta(vect& t) 
{
  int j;
  delta = out - t;  
  for (j=0; j<n_outputs; j++)
     delta[j] *= dg(out[j]); 
  return delta;
}
matrix& LayerNetTrain::calc_grad(vect& t, vect& vinput)
{
  int i, j;
  calc_output_delta(t);
  for (i=0; i<n_outputs; i++)
   {
     for (j=0; j<n_inputs; j++)    
      	grad(i,j) = delta[i] * vinput[j];
     grad_bias[i] = delta[i];
   }
  return grad;
}
vect& LayerNetTrain::calc_hidden_delta(vect& delta_up, matrix& W_up)
{
  int j;
  for (j=0; j<n_outputs; j++)     
      delta[j] = (W_up.col(j) * delta_up) * dg(out[j]);
  return delta;
}
matrix& LayerNetTrain::backprop(vect& delta_up, matrix& W_up, vect& vinput) 
{
  int i, j;
  calc_hidden_delta(delta_up, W_up);
  for (i=0; i<n_outputs; i++)
   {
     for (j=0; j<n_inputs; j++)    
      	grad(i,j) = delta[i] * vinput[j];
     grad_bias[i] = delta[i];
   }
  return grad;
}
void LayerNetTrain::print_delta(void)
{
  cout << "Layer deltas :" << endl;
  delta.print(); cout << endl;
}
void LayerNetTrain::print_grad(void)
{
  cout << "Gradient of the layer :" << endl;
  grad.print();
}
void LayerNetTrain::print_DeltaW(void)
{
  cout << "DeltaW of the layer :" << endl;
  DeltaW.print();
}
void LayerNetTrain::print_grad_bias(void)
{
  cout << "Bias gradient of the layer :" << endl;
  grad_bias.print(); cout << endl;
}
void LayerNetTrain::print_Delta_bias(void)
{
  cout << "Delta_bias of the layer :" << endl;
  Delta_bias.print(); cout << endl;
}
void LayerNetTrain::save_weights(ostream& f)
{
  int i, j;
  f.setf(ios::fixed, ios::floatfield);
  f.setf(ios::showpos);
  f.precision(WEIGHT_PRECISION); 
  for (i=0; i<n_outputs; i++)  {
    for (j=0; j<n_inputs; j++) 
      f << W(i,j) << " ";
    f << endl; 
  } 
  f << "Bias vector:" << endl;
  for (i=0; i<n_outputs; i++)
    f << bias[i] << " ";
  f << endl << endl;
}
