/**
  Application for evaluating joint entropy, mutual information and mutual error
	between 7 variables.  
*/



#include"mutualinfo7.h"



void print_usage(void) 
{
  cerr << "
  mi72. Application for evaluating joint entropy, mutual information and mutual error
	    between 7 variables.
  
	Correct usage " << endl;

  cerr << "
        mi76 filename numdata first last delta
	filename   :   name of the file to be read          
	numdata    :   number of the data                   
	first      :   first column of the file to be read  
	last       :   last column of the file to be read   
	delta      :   range of the first bin (the correct bin)
  " << endl;
  
  exit(1);
} 



int main(int argc, char* argv[])
{
  
  unsigned num_data, i;
  const unsigned num_bins = 6;
  unsigned first;
	unsigned last;
  float *binlimits;
	float min_v = 0.0;
	float max_v = 1.0;
	float delta; // limit of the first interval
	float x, ampl;
 
	
	if (argc < 6)
	  print_usage();
	num_data = atoi(argv[2]);
	first = atoi(argv[3]);
	last = atoi(argv[4]);	
	delta =  atof(argv[5]);
	
	binlimits = new float[num_bins+1];
	binlimits[0] = min_v;
	binlimits[1] = delta;
	x = max_v - delta;
	ampl = x /(num_bins-1);
	for (i=2; i < num_bins; i++) 
		  binlimits[i] = delta + ampl*(i-1);
	binlimits[num_bins] = max_v;
	

	
	// print to stdout file name
	cout << argv[1] << endl << endl;  
 
	  
	MIError7<num_bins> myMIerr7(num_data, argv[1], first, last, 
                              binlimits, min_v, max_v);
 
	myMIerr7.Calc_p();
	myMIerr7.Print_p();
	
	myMIerr7.Calc_Entropy();
	myMIerr7.Print_Entropy();
	
	myMIerr7.Calc_JointEntropy();
	myMIerr7.Print_JointEntropy();
	
	
	myMIerr7.Calc_Minfo();
	myMIerr7.Print_Minfo();
	
	myMIerr7.Calc_MICorrect2();
	myMIerr7.Print_MICorrect2();	
	
	myMIerr7.Calc_MIErr2(); 
	myMIerr7.Print_MIErr2();
	
	myMIerr7.Calc_MICorrect();
	myMIerr7.Print_MICorrect();	
	
	myMIerr7.Calc_MIErr(); 
	myMIerr7.Print_MIErr();
	
	
	// It prints to stdout the computed values to be collected by the scripts
	cout << endl;
	cout << myMIerr7.Read_JointEntropy() << endl;
	cout << myMIerr7.Read_Minfo() << endl;
	cout << myMIerr7.Read_MiErr2() << endl;
	cout << myMIerr7.Read_MiErr() << endl;
	
	delete[] binlimits;
	
		
}	
