/** 
   mm (mutual matrix)
   Application for evaluating mutual information among couples of variables
   Usage:
	 mm filename numdata first last numbins limitbin[1] limitbin[2] ... limitbin[numbins+1]   
*/


#include<stdlib.h>
#include"mutualinfo.h"
#include"utils.h"



void print_usage(void) 
{
  cerr << "
  mm. NEURObjects application for evaluating joint entropy, mutual information and mutual error
	    between pairs of variables.
  
	Correct usage: " << endl;

  cerr << "
  mm filename numdata first last numbins delta 
  filename   :   name of the file to be read          
  numdata    :   size of the data                     
  first      :   first column of the file to be read  
  last       :   last column of the file to be read   
  numbins    :   number of the bins by wich each variable is discretized
  delta      :   range of the first bin (the correct bin)
  min        :   minimum value of the variable (optional, def. = 0)
  max        :   maximum value of the variable (optional, def. = 1)
  " << endl;
  
  exit(1);
} 



int main(int argc, char* argv[])
{
 
  unsigned num_data, i;
  unsigned num_bins = 2;
  unsigned first;
	unsigned last;
  float *binlimits;
	float min_v = 0.0;
	float max_v = 1.0;
	float delta; // limit of the first interval
	float x, ampl;
 
	
	if (argc < 7)
	  print_usage();
	num_data = atoi(argv[2]);
	first = atoi(argv[3]);
	last = atoi(argv[4]);	
	
	
	num_bins = atoi(argv[5]);
	delta = atof(argv[6]);
  if (argc == 8)
    min_v = atof(argv[7]);
  if (argc == 9) {
    min_v = atof(argv[7]);
    max_v = atof(argv[8]);
  }
  	
  binlimits = new float[num_bins+1];
	binlimits[0] = min_v;
	binlimits[1] = delta;
	x = max_v - delta;
	ampl = x /(num_bins-1);
	for (i=2; i < num_bins; i++) 
		  binlimits[i] = delta + ampl*(i-1);
	binlimits[num_bins] = max_v;
	
	
	
	cout << argv[1] << endl << endl;  
	MutualMatrix mm(num_data, num_bins, argv[1], first, 
	                last, binlimits, min_v, max_v); 
	
	mm.Calc_JointEntropy_Matrix ();
	mm.Print_JointEntropy_Matrix ();
	cout << "Sum Joint Entr. Matrix = " << mm.SumElements_JointEntropy_Matrix() 
	                                    << endl << endl << endl;
	
	mm.Calc_MutualInfo_Matrix ();
	mm.Print_MutualInfo_Matrix ();	
	cout << "Sum Mutual Info  Matrix= " << mm.SumElements_MutualInfo_Matrix() 
	                                    << endl << endl << endl;
																			
	mm.Calc_MutualInfoError2_Matrix ();
	mm.Print_MutualInfoError2_Matrix ();	
	cout << "Sum Mutual Info Error 2 Matrix= " << mm.SumElements_MutualInfoError2_Matrix()
	                                           << endl << endl << endl;
																						 
	mm.Calc_MutualInfoError_Matrix ();
	mm.Print_MutualInfoError_Matrix ();	
	cout << "Sum Mutual Info Error Matrix= " << mm.SumElements_MutualInfoError_Matrix()
	                                         << endl << endl << endl;	
	
	
	// prints to stdout the values of the sum of the matrices (to be collected by the comparison script																		 	
	cout << mm.Read_SumEnM() << endl;
	cout << mm.Read_SumMiM() << endl;
	cout << mm.Read_SumMeM2() << endl;
	cout << mm.Read_SumMeM() << endl;

	
	
	
	delete[] binlimits;
}	






