// "mutualinfo.cc"

#pragma implementation
#include<stdlib.h>
#include<math.h>
#include<iostream.h>
#include<fstream.h>
#include"mutualinfo.h"
#include"utils.h"

static const float LOG_NAT_TO_2_FACT = 0.693147181;
void	MutualInfo2::builddata(unsigned num_data) 
{
  int i;
	data = new float*[num_data];
	 for (i=0; i<num_data; i++)
		 data[i] = new float[2];
}
	
void MutualInfo2::buildp(unsigned num_bins)
{
	int i;
	p = new float*[2];
	for (i=0; i<2; i++)
		 p[i] = new float[num_bins];	
	Init_p();
}	
void MutualInfo2::buildPjoint(unsigned num_bins)
{
	int i,j;
	Pjoint = new float*[num_bins];
	for (i=0; i<num_bins; i++)
		 Pjoint[i] = new float[num_bins];	
	Init_Pjoint();
}
	
MutualInfo2::MutualInfo2 (unsigned num_data, unsigned num_bins,  float **thedata,
	                        float min_v = MIN_VALUE, float max_v = MAX_VALUE)
{
	 
	nbins = num_bins;
  ndata = num_data;
	min = min_v;
	max = max_v;
	ampl = (max-min)/nbins;
	
	 
	builddata(num_data);
  for (int i=0; i<num_data; i++) {
	   data[i][0] = thedata[i][0];
		 data[i][1] = thedata[i][1];
	}
	
	 
	buildp(num_bins); 
	
	 
	buildPjoint(num_bins);
	
}
		  
int MutualInfo2::readdata(unsigned num_data, char *f, unsigned first, unsigned second) 
{
	char c;
	unsigned i,k;
	float buf;
	
	ifstream from(f); 
  if (!from) error("Can't open input file :", f); 
	
	
	for (k=0; k<num_data; k++) {
		i=0;
		 
		do from >> buf;
		while (++i != first);		
		data[k][0] = buf;
		 
		do from >> buf;
		while (++i != second);
	  data[k][1] = buf;
		toendofline(from);
	}
	
	 
  while (from.get(c));
  if (!from.eof()) {
     error("Error reading file ", f); 
		 return 1;
	}
	return 0;	
}
MutualInfo2::MutualInfo2 (unsigned num_data, unsigned num_bins,  char *f, 
	                        unsigned first, unsigned second, float min_v = MIN_VALUE, float max_v =
													MAX_VALUE)
{
  	
	nbins = num_bins;
  ndata = num_data;
	min = min_v;
	max = max_v;
	ampl = (max-min)/nbins;
	
	 
	builddata(num_data);
	 
	readdata(num_data, f, first, second); 
	 
	buildp(num_bins); 
	 
	buildPjoint(num_bins);
	
} 
MutualInfo2::~MutualInfo2 () 
{
   int i;
	 
	  
	 for (i=0; i<ndata; i++)
		 delete[] data[i];
	 delete[] data;
	  
	 for (i=0; i<2; i++)
	   delete[] p[i];
	 delete[] p;
	  
	 for (i=0; i<nbins; i++)
	   delete[] Pjoint[i];
	 delete[] Pjoint;	  	 
}
void MutualInfo2::printdata(void)
{
  int i;
	cout << "Print out data values of the two random variables:" << endl;
	for (i=0; i<ndata; i++)
	  cout << data[i][0] << "  " << data[i][1] << endl;
}
unsigned MutualInfo2::assignbin(unsigned num_var, unsigned num_data)
{ 
  int j;
	float bin_left;  
	
	for (j=nbins-1; j>=0; j--) {
	  bin_left = (ampl*j) + min;
		if (data[num_data][num_var] >= bin_left)
		  return j;
	}
	error_exit("Error in MutualInfo2::assignbin", "Overflow beyond the minimum value");
	return 0;
}
	
void MutualInfo2::Init_p (float init=0.0)
{
  unsigned i,j;
  for (i=0; i<2; i++)
	  for (j=0; j<nbins; j++)
		  p[i][j] = init;	  
}   
void  MutualInfo2::Calc_p (void)
{
  unsigned i,j,thebin;
	
	Init_p();
	for (i=0; i<2; i++)
	  for (j=0; j<ndata; j++) {
		  thebin = assignbin(i,j);
		  p[i][thebin]++;
		}
	for (i=0; i<2; i++)
	  for (j=0; j<nbins; j++)
		  p[i][j] = p[i][j] / (float)ndata;
}
void MutualInfo2::Print_p(void)
{
  int i,j;
	
	
	cout << "Matrix of the frequencies of the 2 random variables with ";
	cout << nbins << " bins:" << endl;
	
	cout.setf(ios::fixed, ios::floatfield);
  cout.setf(ios::showpos);
  cout.precision(5); 
	for (i=0; i<2; i++){
	  for(j=0; j<nbins; j++) 	  
		  cout << p[i][j] << "  ";
		cout << endl;
	}
	cout << endl;
}
	
void MutualInfo2::Init_Pjoint (float init=0.0)
{
  unsigned i,j;
  for (i=0; i<nbins; i++)
	  for (j=0; j<nbins; j++)
		  Pjoint[i][j] = init;	  
}   
	
void  MutualInfo2::Calc_Pjoint (void)
{
  unsigned i,j,thebin0, thebin1;
	
	Init_Pjoint();
	for (j=0; j<ndata; j++) {
		thebin0 = assignbin(0,j);
		thebin1 = assignbin(1,j);
		Pjoint[thebin0][thebin1]++;
	}
	for (i=0; i<nbins; i++)
	  for (j=0; j<nbins; j++)
		  Pjoint[i][j] = Pjoint[i][j] / (float)ndata;
}
void MutualInfo2::Print_Pjoint(void)
{
  int i,j;
	
	
	cout << "Matrix of the joint frequency of the 2 random variables with ";
	cout << nbins << " bins:" << endl;
	
	cout.setf(ios::fixed, ios::floatfield);
  cout.setf(ios::showpos);
  cout.precision(5); 
	for(i=0; i<nbins; i++) {
	  for (j=0; j<nbins; j++) 	  
		  cout << Pjoint[i][j] << "  ";
		cout << endl;
	}
	cout << endl;
}
float*  MutualInfo2::Calc_Entropy(void)
{
  int i,j;
	float c;
	
	Calc_p();
	H[0] = H[1] = 0;
	for (i=0; i<2; i++) {
	  for (j=0; j<nbins; j++) 
		  if ((c = p[i][j]) > 0)
	      H[i] += (c * log (c) / LOG_NAT_TO_2_FACT);
	  H[i] = - H[i];
	}
	return H;
}
	
void MutualInfo2::Print_Entropy(void)	
{
  int i;
	
	
	cout << "Entropies of the 2 random variables with ";
	cout << nbins << " bins:" << endl;
	
	cout.setf(ios::fixed, ios::floatfield);
  cout.setf(ios::showpos);
  cout.precision(5); 
	
  for (i=0; i<2; i++) 	  
	  cout << H[i] << "  ";
	cout << endl << endl;
}
float MutualInfo2::Calc_JointEntropy(void)
{
  int i,j;
	float c;
	
	Calc_Pjoint();
	Hjoint = 0;
	for (i=0; i<nbins; i++) {
	  for (j=0; j<nbins; j++) 
		  if ((c = Pjoint[i][j]) > 0)
	      Hjoint += (c * log (c) / LOG_NAT_TO_2_FACT);
	}
	Hjoint = -Hjoint;
	return Hjoint;
}
void MutualInfo2::Print_JointEntropy(void)	
{
  	
	cout << "Joint Entropy of the 2 random variables with ";
	cout << nbins << " bins: ";
	
	cout.setf(ios::fixed, ios::floatfield);
  cout.setf(ios::showpos);
  cout.precision(5); 
	 	  
	cout << Hjoint << endl << endl;
}
float MutualInfo2::Calc_Minfo(void)
{
  int i,j;
	float c0, c1, cjoint;
	float temp;  
	Calc_p();
	Calc_Pjoint();
  Minfo = 0;
	for (i=0; i<nbins; i++) {
	  for (j=0; j<nbins; j++) {
		  c0 = p[0][i];
			c1 = p[1][j];
			cjoint = Pjoint[i][j];
			if ((c0>0) && (c1>0) && (cjoint>0)) {
			  temp = (cjoint * log (cjoint/(c0*c1)) / LOG_NAT_TO_2_FACT);
				 
	      Minfo += temp;
			}
		}
	}	
	return Minfo;
}
void MutualInfo2::Print_Minfo(void)	
{
  	
	cout << "Mutual information of the 2 random variables with ";
	cout << nbins << " bins: ";
	
	cout.setf(ios::fixed, ios::floatfield);
  cout.setf(ios::showpos);
  cout.precision(5); 
	 	  
	cout << Minfo << endl << endl;
}
void MIError2::buildbin(float *binlimits = 0)
{  
	int j;
	if (binlimits == 0) {
	  for (j=0; j < nbins; j++)
		  bin[j] = min + j*ampl;
		 bin[nbins] = max;
	}
	else {
	  for (j=0; j < nbins; j++)
		  if (binlimits[j] > binlimits[j+1])
			  error_exit("MIError2: bad ordered bin limits");
	  for (j=0; j <= nbins; j++)
		  bin[j] = binlimits[j];
		min = bin[0];
		max = bin[nbins];
	}
}
MIError2::MIError2 (unsigned num_data, unsigned num_bins,  float **thedata, float *binlimits = 0,
	                  float min_v = 0.0, float max_v = 1.0)
: MutualInfo2 (num_data, num_bins, thedata, min_v, max_v)
{
		
	bin = new float[num_bins+1];
	 
	buildbin(binlimits);	
}
MIError2::MIError2 (unsigned num_data, unsigned num_bins,  char *f, unsigned first, 
                    unsigned second, float *binlimits = 0, float min_v = MIN_VALUE, 
										float max_v = MAX_VALUE)
: MutualInfo2 (num_data, num_bins, f, first, second, min_v, max_v)
{
  	
	bin = new float[num_bins+1];
	 
	buildbin(binlimits);		
} 
MIError2::~MIError2 () 
{
	 delete[] bin;
}
unsigned MIError2::assignbin(unsigned num_var, unsigned num_data)
{ 
  int j;
	float bin_left;  
	
	for (j=nbins-1; j>=0; j--) 
		if (data[num_data][num_var] >= bin[j])
		  return j;
	error_exit("Error in MIError2::assignbin", "Overflow beyond the minimum value");
	return 0;
}
float MIError2::Calc_MICorrect2(void)
{
  int j;
	float c0, c1, cjoint;
	Calc_p();
	Calc_Pjoint();
  MICorrect2 = 0;
	
	 
	c0 = p[0][0];  
	for (j=1; j<nbins; j++) {		  
			c1 = p[1][j];  
			cjoint = Pjoint[0][j];  
			if ((c0>0) && (c1>0) && (cjoint>0))
	      MICorrect2 += (cjoint * log (cjoint/(c0*c1)) / LOG_NAT_TO_2_FACT);
	}
	 
	c1 = p[1][0];  
	for (j=1; j<nbins; j++) {		  
			c0 = p[0][j];  
			cjoint = Pjoint[j][0];  
			if ((c0>0) && (c1>0) && (cjoint>0))
	      MICorrect2 += (cjoint * log (cjoint/(c0*c1)) / LOG_NAT_TO_2_FACT);
	}
	 
	c0 = p[0][0];  
	c1 = p[1][0];  
	cjoint = Pjoint[0][0];  
	if ((c0>0) && (c1>0) && (cjoint>0))
	   MICorrect2 += (cjoint * log (cjoint/(c0*c1)) / LOG_NAT_TO_2_FACT);
	
	return MICorrect2;
}
void MIError2::Print_MICorrect2(void)	
{
  	
	cout << "Mutual information generated by no errors or not correlated errors 
	         of the 2 random variables with ";
	cout << nbins << " bins: ";
	
	cout.setf(ios::fixed, ios::floatfield);
  cout.setf(ios::showpos);
  cout.precision(5); 
	 	  
	cout << MICorrect2 << endl << endl;
}
float MIError2::Calc_MIErr2(void) 
{
  MIErr2 = Calc_Minfo() - Calc_MICorrect2();
	return MIErr2;
}
void MIError2::Print_MIErr2(void)	
{
  	
	cout << "Mutual information generated by correlated errors 
	         of the 2 random variables with ";
	cout << nbins << " bins: ";
	
	cout.setf(ios::fixed, ios::floatfield);
  cout.setf(ios::showpos);
  cout.precision(5); 
	 	  
	cout << MIErr2 << endl << endl;
}
float MIError2::Calc_MICorrect(void)
{
	float c0, c1, cjoint;
	Calc_p();
	Calc_Pjoint();
	 
	c0 = p[0][0];  
	c1 = p[1][0];  
	cjoint = Pjoint[0][0];  
	if ((c0>0) && (c1>0) && (cjoint>0))
	   MICorrect = (cjoint * log (cjoint/(c0*c1)) / LOG_NAT_TO_2_FACT);
	
	return MICorrect;
}
void MIError2::Print_MICorrect(void)	
{
  	
	cout << "Mutual information generated by correct guess
	         of the 2 random variables with ";
	cout << nbins << " bins: ";
	
	cout.setf(ios::fixed, ios::floatfield);
  cout.setf(ios::showpos);
  cout.precision(5); 
	 	  
	cout << MICorrect << endl << endl;
}
float MIError2::Calc_MIErr(void) 
{
  MIErr = Calc_Minfo() - Calc_MICorrect();
	return MIErr;
}
void MIError2::Print_MIErr(void)	
{
  	
	cout << "Mutual information generated by  errors 
	         of the 2 random variables with ";
	cout << nbins << " bins: ";
	
	cout.setf(ios::fixed, ios::floatfield);
  cout.setf(ios::showpos);
  cout.precision(5); 
	 	  
	cout << MIErr << endl << endl;
}
MutualMatrix::MutualMatrix(unsigned num_data, unsigned num_bins,  char *f, unsigned first, 
	               unsigned last, float *binlimits = 0, float min_v = 0.0, float max_v = 1.0) 
{
  int i;
	
  nbins = num_bins;
  ndata = num_data;
	min = min_v;
	max = max_v;
	ampl = (max-min)/nbins;
	MutualMatrix::first = first;
	MutualMatrix::last = last;
	bin = binlimits;
	nvar = last - first + 1;
	strcpy(namefile, f);
	
	 
	EnM = new float*[nvar];
	MiM = new float*[nvar];
	MeM2 = new float*[nvar];
	MeM = new float*[nvar];
  for (i=0; i<nvar; i++) {
	  EnM[i] = new float[nvar];
		MiM[i] = new float[nvar];
		MeM2[i] = new float[nvar];
		MeM[i] = new float[nvar];
	}
		
}
MutualMatrix::~MutualMatrix () 
{
   int i;
	 
	  
	 for (i=0; i<nvar; i++) {
		 delete[] EnM[i];
		 delete[] MiM[i];
		 delete[] MeM[i];
		 delete[] MeM2[i];
	 }
	 delete[] EnM;
	 delete[] MiM;
	 delete[] MeM;
	 delete[] MeM2;	  
}
void MutualMatrix::Calc_JointEntropy_Matrix () 
{
	 
	int i,j; 
	
	for (i=0; i<nvar; i++)
	  for (j=0; j<i; j++) {
	  MIError2* ElemMIErr = new MIError2(ndata, nbins,  namefile, first+j, 
                                       first+i, bin, min, max);
		EnM[i][j] = ElemMIErr->Calc_JointEntropy();
		 
		delete ElemMIErr;
		}
}
float MutualMatrix::SumElements_JointEntropy_Matrix () 
{
  int i,j;
	float sum = 0.0;
	
	for (i=0; i<nvar; i++) 
	  for (j=0; j<i; j++)
			sum += EnM[i][j];
	SumEnM = sum;
	return sum;
}
	
void MutualMatrix::Print_JointEntropy_Matrix () 
{
  int i,j;
		
	cout << "Joint Entropy Matrix with " << nvar << " variables" << endl;
	
	cout.setf(ios::fixed, ios::floatfield);
  cout.setf(ios::showpos);
  cout.precision(5); 
	for (i=0; i<nvar; i++) {
	  for (j=0; j<i; j++) 
	    cout << EnM[i][j] << "  ";
		cout << endl;
	  }
	cout << endl;
}
void MutualMatrix::Calc_MutualInfo_Matrix () 
{
	 
	int i,j; 
	
	for (i=0; i<nvar; i++)
	  for (j=0; j<i; j++) {
	  MIError2* ElemMIErr = new MIError2(ndata, nbins,  namefile, first+j, 
                                       first+i, bin, min, max);
		MiM[i][j] = ElemMIErr->Calc_Minfo();
		 
		delete ElemMIErr;
		}
}
float MutualMatrix::SumElements_MutualInfo_Matrix () 
{
  int i,j;
	float sum = 0.0;
	
	for (i=0; i<nvar; i++) 
	  for (j=0; j<i; j++)
			sum += MiM[i][j];
	SumMiM = sum;
	return sum;
}
void MutualMatrix::Print_MutualInfo_Matrix ()
{
  int i,j;
		
	cout << "Mutual information Matrix with " << nvar << " variables" << endl;
	
	cout.setf(ios::fixed, ios::floatfield);
  cout.setf(ios::showpos);
  cout.precision(5); 
	for (i=0; i<nvar; i++) {
	  for (j=0; j<i; j++) 
	    cout << MiM[i][j] << "  ";
		cout << endl;
	  }
	cout << endl;
}
void MutualMatrix::Calc_MutualInfoError2_Matrix () 
{
	 
	int i,j; 
	
	for (i=0; i<nvar; i++)
	  for (j=0; j<i; j++) {
	  MIError2* ElemMIErr = new MIError2(ndata, nbins,  namefile, first+j, 
                                       first+i, bin, min, max);
		MeM2[i][j] = ElemMIErr->Calc_MIErr2();
		 
		delete ElemMIErr;
		}
}
float MutualMatrix::SumElements_MutualInfoError2_Matrix () 
{
  int i,j;
	float sum = 0.0;
	
	for (i=0; i<nvar; i++) 
	  for (j=0; j<i; j++)
			sum += MeM2[i][j];
	SumMeM2 = sum;
	return sum;
}
void MutualMatrix::Print_MutualInfoError2_Matrix ()
{
  int i,j;
		
	cout << "Mutual information error Matrix 
	         not considering single errors with " << nvar << " variables" << endl;
	
	cout.setf(ios::fixed, ios::floatfield);
  cout.setf(ios::showpos);
  cout.precision(5); 
	for (i=0; i<nvar; i++) {
	  for (j=0; j<i; j++) 
	    cout << MeM2[i][j] << "  ";
		cout << endl;
	  }
	cout << endl;
}
void MutualMatrix::Calc_MutualInfoError_Matrix () 
{
	 
	int i,j; 
	
	for (i=0; i<nvar; i++)
	  for (j=0; j<i; j++) {
	  MIError2* ElemMIErr = new MIError2(ndata, nbins,  namefile, first+j, 
                                       first+i, bin, min, max);
		MeM[i][j] = ElemMIErr->Calc_MIErr();
		 
		delete ElemMIErr;
		}
}
float MutualMatrix::SumElements_MutualInfoError_Matrix () 
{
  int i,j;
	float sum = 0.0;
	
	for (i=0; i<nvar; i++) 
	  for (j=0; j<i; j++)
			sum += MeM[i][j];
	SumMeM = sum;
	return sum;
}
void MutualMatrix::Print_MutualInfoError_Matrix ()
{
  int i,j;
		
	cout << "Mutual information error Matrix  with " << nvar << " variables" << endl;
	
	cout.setf(ios::fixed, ios::floatfield);
  cout.setf(ios::showpos);
  cout.precision(5); 
	for (i=0; i<nvar; i++) {
	  for (j=0; j<i; j++) 
	    cout << MeM[i][j] << "  ";
		cout << endl;
	  }
	cout << endl;
}
