// mutualinfo7.cc

#pragma implementation
#include<cstdlib>
#include<math.h>
#include<iostream.h>
#include<fstream.h>
#include"mutualinfo7.h"
#include"utils.h"



static const float LOG_NAT_TO_2_FACT = 0.693147181;



// ===================== Class MutualInfo7 ==========================================


// building of matrix data	 
template <int sz> void	MutualInfo7<sz>::builddata(unsigned num_data) 
{
  int i;
	data = new float*[num_data];
	 for (i=0; i<num_data; i++)
		 data[i] = new float[num_var];
}



template <int sz> int MutualInfo7<sz>::readdata(unsigned num_data, char *f, 
                                                unsigned first, unsigned last) 
{
	char c;
	unsigned i,k,currcol;
	float buf;
	
	ifstream from(f); 
  if (!from) error("Can't open input file :", f); 
		
	for (k=0; k<num_data; k++) {
		i=0;
		// read data from first column
		do from >> buf;
		while (++i != first);
		currcol = 0;		
		data[k][currcol] = buf;

		// read data from other columns
		do {
		  from >> buf;		 
		  data[k][++currcol] = buf;
		} while (++i != last);
	 

		toendofline(from);
	}
	
	// salta eventuali caratteri rimasti (che devono essere solo blank)
  while (from.get(c));
          
  if (!from.eof()) {
     error("Error reading file ", f); 
		 return 1;
	}
	return 0;	
}



	
/// building and initialization of matrix of single frequencies
template <int sz> void MutualInfo7<sz>::buildp(unsigned num_bins)
{
	int i;
	p = new float*[num_var];
	for (i=0; i<num_var; i++)
		 p[i] = new float[num_bins];	
	Init_p();
}	


template <int sz> void MutualInfo7<sz>::Init_p (float init)
{
  unsigned i,j;
  for (i=0; i<num_var; i++)
	  for (j=0; j<nbins; j++)
		  p[i][j] = init;	  
}  


template <int sz> void MutualInfo7<sz>::Init_Pjoint (float init)
{
  int i,j,k,l,m,n,o;
		for (i=0; i<sz; i++)
		  for (j=0; j<sz; j++)
			  for (k=0; k<sz; k++)
				  for (l=0; l<sz; l++)
					  for (m=0; m<sz; m++)
						  for (n=0; n<sz; n++)
							  for (o=0; o<sz; o++)
		               Pjoint[i][j][k][l][m][n][o] = init;
}   



template <int sz> void MutualInfo7<sz>::control_bins(void)
{ 
  if ((sz< MIN_BIN) || (sz > MAX_BIN)) 
	  error_exit("MutualInfo7::control_bins : ", "Number of bins out of range");
}
	  
  

// Constructor
template <int sz> MutualInfo7<sz>::MutualInfo7(unsigned num_data,  float **thedata,
	              float min_v, float max_v)
{
	 
	nbins = sz;
  ndata = num_data;
	min = min_v;
	max = max_v;
	ampl = (max-min)/nbins;
	
	// Control if bins are between MIN_BIN and MAX_BIN
	control_bins();
	
	// building and initialization of matrix data	
	builddata(num_data);
  for (int i=0; i<num_data; i++)
	  for (int j=0; j<num_var; j++)
	   data[i][j] = thedata[i][j];
	
	// building and initialization of matrix of marginal frequencies
	buildp(nbins); 
	
	// initialization of matrix of joint frequencies
	Init_Pjoint();
	
}



template <int sz> MutualInfo7<sz>::MutualInfo7 (unsigned num_data, char *f, 
	                        unsigned first, unsigned last, float min_v, float max_v)
{
  	
	nbins = sz;
  ndata = num_data;
	min = min_v;
	max = max_v;
	ampl = (max-min)/nbins;
	
	// building of matrix data	
	builddata(num_data);
	// load data from file f into array data
	readdata(num_data, f, first, last); 
	// building and initialization of matrix of single frequencies
	buildp(nbins); 
	// initialization of matrix of joint frequencies
	Init_Pjoint();
	
} 


template <int sz> MutualInfo7<sz>::~MutualInfo7 () 
{
   int i;
	 
	 // destroying matrix data
	 for (i=0; i<ndata; i++)
		 delete[] data[i];
	 delete[] data;
	 // destroying matrix p
	 for (i=0; i<num_var; i++)
	   delete[] p[i];
	 delete[] p;	  
}



/// printdata
template <int sz> void MutualInfo7<sz>::printdata(void)
{
  int i,j;
	cout << "Print out data values of the seven random variables:" << endl;
	for (i=0; i<ndata; i++) {
	  for (j=0; j<num_var; j++)
	    cout << data[i][j] << "  ";
		cout << endl;
	}
	cout << endl;
}




template <int sz> unsigned MutualInfo7<sz>::assignbin(unsigned num_var, unsigned num_data)
{ 
  int j;
	float bin_left; // left extremum of a bin
	
	for (j=nbins-1; j>=0; j--) {
	  bin_left = (ampl*j) + min;
		if (data[num_data][num_var] >= bin_left)
		  return j;
	}
	error_exit("Error in MutualInfo2::assignbin", "Overflow beyond the minimum value");
	return 0;
}


template <int sz> void  MutualInfo7<sz>::Calc_p (void)
{
  unsigned i,j,thebin;
	
	Init_p();
	for (i=0; i<num_var; i++)
	  for (j=0; j<ndata; j++) {
		  thebin = assignbin(i,j);
		  p[i][thebin]++;
		}
	for (i=0; i<num_var; i++)
	  for (j=0; j<nbins; j++)
		  p[i][j] = p[i][j] / (float)ndata;
}


template <int sz> void MutualInfo7<sz>::Calc_Pjoint (void)
{
  int i,j,k,l,m,n,o;
		
	Init_Pjoint();
	for (j=0; j<ndata; j++) 
		Pjoint[assignbin(0,j)][assignbin(1,j)][assignbin(2,j)][assignbin(3,j)][assignbin(4,j)][assignbin(5,j)][assignbin(6,j)]++;
	
	for (i=0; i<sz; i++)
		for (j=0; j<sz; j++)
			for (k=0; k<sz; k++)
				for (l=0; l<sz; l++)
					for (m=0; m<sz; m++)
						for (n=0; n<sz; n++)
							for (o=0; o<sz; o++)
		             Pjoint[i][j][k][l][m][n][o] /= (float)ndata;
}


/// print the matrix p
template <int sz> void MutualInfo7<sz>::Print_p(void)
{
  int i,j;
	
	
	cout << "Matrix of the frequencies of the 7 random variables with ";
	cout << nbins << " bins:" << endl;
	
	cout.setf(ios::fixed, ios::floatfield);
  cout.setf(ios::showpos);
  cout.precision(5); 
	for (i=0; i<num_var; i++) {
	  for(j=0; j<nbins; j++)	   	  
		  cout << p[i][j] << "  ";
		cout << endl;
	}
	cout << endl;
}


template <int sz> float MutualInfo7<sz>::Sum_Pjoint(void)
{
  int i,j,k,l,m,n,o;
	float sum = 0.0;
	
	for (i=0; i<sz; i++)
		for (j=0; j<sz; j++)
			for (k=0; k<sz; k++)
				for (l=0; l<sz; l++)
					for (m=0; m<sz; m++)
						for (n=0; n<sz; n++)
							for (o=0; o<sz; o++)
		             sum += Pjoint[i][j][k][l][m][n][o];
	return sum;
}


/// Calculate the entropy of the random variables
template <int sz> void  MutualInfo7<sz>::Init_Entropy(float init)
{
  int i;
	for (i=0; i<num_var; i++)
	  H[i] = init;
}


/// Calculate the entropy of the random variables
template <int sz> float*  MutualInfo7<sz>::Calc_Entropy(void)
{
  int i,j;
	float c;
	
	Calc_p();
	Init_Entropy();
	for (i=0; i<num_var; i++) {
	  for (j=0; j<nbins; j++) 
		  if ((c = p[i][j]) > 0)
	      H[i] += (c * log (c) / LOG_NAT_TO_2_FACT);
	  H[i] = - H[i];
	}
	return H;
}
	

/// Print to stdout the entropy of the random variables
template <int sz> void MutualInfo7<sz>::Print_Entropy(void)	
{
  int i;
	
	
	cout << "Entropies of the 7 random variables with ";
	cout << nbins << " bins:" << endl;
	
	cout.setf(ios::fixed, ios::floatfield);
  cout.setf(ios::showpos);
  cout.precision(5); 
	
  for (i=0; i<num_var; i++) 	  
	  cout << H[i] << "  ";
	cout << endl << endl;
}


/**
It calculates the joint entropy
@return joint entropy of the ra
*/
template <int sz> float MutualInfo7<sz>::Calc_JointEntropy(void)
{
  int i,j,k,l,m,n,o;
	float c;
	
	Calc_Pjoint();
	Hjoint = 0;
	for (i=0; i<sz; i++)
		for (j=0; j<sz; j++)
			for (k=0; k<sz; k++)
				for (l=0; l<sz; l++)
					for (m=0; m<sz; m++)
						for (n=0; n<sz; n++)
							for (o=0; o<sz; o++)
							  if ((c = Pjoint[i][j][k][l][m][n][o]) > 0)
	                Hjoint += (c * log (c) / LOG_NAT_TO_2_FACT);
	Hjoint = -Hjoint;
	return Hjoint;
}


/// Print to stdout the joint entropy of the random variables
template <int sz> void MutualInfo7<sz>::Print_JointEntropy(void)	
{
  	
	cout << "Joint Entropy of the 7 random variables with ";
	cout << nbins << " bins: ";
	
	cout.setf(ios::fixed, ios::floatfield);
  cout.setf(ios::showpos);
  cout.precision(5); 
	 	  
	cout << Hjoint << endl << endl;
}



template <int sz> float MutualInfo7<sz>::Calc_Minfo(void)
{
  int i,j,k,l,m,n,o;
	float prodp, cjoint;
  
	Calc_p();
	Calc_Pjoint();
  Minfo = 0.0;
	for (i=0; i<sz; i++)
		for (j=0; j<sz; j++)
			for (k=0; k<sz; k++)
				for (l=0; l<sz; l++)
					for (m=0; m<sz; m++)
						for (n=0; n<sz; n++)
							for (o=0; o<sz; o++) {
							  cjoint = Pjoint[i][j][k][l][m][n][o];
								if (cjoint>0) {
								  prodp = p[0][i]*p[1][j]*p[2][k]*p[3][l]*p[4][m]*p[5][n]*p[6][o];
								  Minfo += (cjoint * log (cjoint/prodp) / LOG_NAT_TO_2_FACT);
								}
							}													
	return Minfo;
}

/// Print to stdout the mutual information of the random variables
template <int sz> void MutualInfo7<sz>::Print_Minfo(void)	
{
  	
	cout << "Mutual information of the 7 random variables with ";
	cout << nbins << " bins: ";
	
	cout.setf(ios::fixed, ios::floatfield);
  cout.setf(ios::showpos);
  cout.precision(5); 
	 	  
	cout << Minfo << endl << endl;
}



/**
	 It calculates the Pearson's statistics. It evaluates the correlation between the 
	 7 random variables.
	 @return result of the Pearson's statistics.
*/
template <int sz> float MutualInfo7<sz>::Calc_Pearson(void) 
{
  int i,j,k,l,m,n,o;
	float prodp, cjoint, diff;
  
	Calc_p();
	Calc_Pjoint();
  Pearson = 0;
	for (i=0; i<sz; i++)
		for (j=0; j<sz; j++)
			for (k=0; k<sz; k++)
				for (l=0; l<sz; l++)
					for (m=0; m<sz; m++)
						for (n=0; n<sz; n++)
							for (o=0; o<sz; o++) {
							  cjoint = Pjoint[i][j][k][l][m][n][o];
								if (cjoint>0) {
								  prodp = p[0][i]*p[1][j]*p[2][k]*p[3][l]*p[4][m]*p[5][n]*p[6][o];
									diff = cjoint - prodp;
								  Pearson += ((diff*diff)/prodp);
								}
							}													
  Pearson = Pearson * ndata;
  return Pearson;
}

/// Print to stdout Pearson's statistics of the random variables
template <int sz> void MutualInfo7<sz>::Print_Pearson(void)	
{
  	
	cout << "Pearson's statistics of the 7 random variables with ";
	cout << nbins << " bins: ";
	
	cout.setf(ios::fixed, ios::floatfield);
  cout.setf(ios::showpos);
  cout.precision(5); 
	 	  
	cout << Pearson << endl << endl;
}



// ====================== Class MIError7 ==================================================

// Constructor 
template <int sz> MIError7<sz>::MIError7 (unsigned num_data,  float **thedata, 
	       float *binlimits, float min_v, float max_v)
: MutualInfo7<sz>(num_data,  thedata, min_v, max_v)
{
  bin = new float[nbins+1];
	// Build bin limits according to binlinmits
	buildbin(binlimits);	
}



// Constructor 
template <int sz> MIError7<sz>::MIError7 (unsigned num_data, char *f, unsigned first, 
        unsigned last, float *binlimits, float min_v, float max_v)
: MutualInfo7<sz>(num_data,  f, first, last, min_v, max_v)
{
  bin = new float[nbins+1];
	// Build bin limits according to binlinmits
	buildbin(binlimits);	
} 


template <int sz> void MIError7<sz>::buildbin(float *binlimits)
{  
	int j;
	if (binlimits == 0) {
	  for (j=0; j < nbins; j++)
		  bin[j] = min + j*ampl;
		 bin[nbins] = max;
	}
	else {
	  for (j=0; j < nbins; j++)
		  if (binlimits[j] > binlimits[j+1])
			  error_exit("MIError7: bad ordered bin limits");
	  for (j=0; j <= nbins; j++)
		  bin[j] = binlimits[j];
		min = bin[0];
		max = bin[nbins];
	}
}



template <int sz> unsigned MIError7<sz>::assignbin(unsigned num_var, unsigned num_data)
{ 
  int j;
	
	for (j=nbins-1; j>=0; j--) 
		if (data[num_data][num_var] >= bin[j])
		  return j;
	error_exit("Error in MIError7::assignbin", "Overflow beyond the minimum value");
	return 0;
}


// Destructor
template <int sz> MIError7<sz>::~MIError7 () 
{
	 delete[] bin;
}


// It calculates the component of mutual information generated by no errors or not correlated
// errors. 
template <int sz> float MIError7<sz>::Calc_MICorrect2(void)
{
  unsigned i,j;
	float prodp, cjoint;
	unsigned a[7]; //each a[i] maintain  the index of the bin of the i variable
	
	MICorrect2 = 0.0;
	// init the bin of all the variables to the correct bin
	for(i=0; i<7; i++)
	  a[i] = 0;
		
	// Add the mutual information generated by a single error
	for(i=0; i<7; i++) {
	  for(j=1; j<nbins; j++) {
		  a[i] = j;
		  cjoint = Pjoint[a[0]][a[1]][a[2]][a[3]][a[4]][a[5]][a[6]];
      if (cjoint>0) {
				prodp = p[0][a[0]]*p[1][a[1]]*p[2][a[2]]*p[3][a[3]]*p[4][a[4]]*p[5][a[5]]*p[6][a[6]];				
				MICorrect2 += (cjoint * log (cjoint/prodp) / LOG_NAT_TO_2_FACT);
			}
		}
		a[i] = 0;
	}

  // adding the element with correct value for all variables
	prodp = p[0][0]*p[1][0]*p[2][0]*p[3][0]*p[4][0]*p[5][0]*p[6][0];
	cjoint = Pjoint[0][0][0][0][0][0][0]; // correct values for all variables
	if (cjoint>0)
	   MICorrect2 += (cjoint * log (cjoint/prodp) / LOG_NAT_TO_2_FACT);
	return MICorrect2;

}


// It calculates the component of mutual information generated by correlated errors. 
template <int sz> float MIError7<sz>::Calc_MIErr2(void)
{
  MIErr2 = Calc_Minfo() - Calc_MICorrect2();
	return MIErr2;
}


/// Print to stdout the mutual information MICorrect generated by no errors or not correlated errors
template <int sz> void MIError7<sz>::Print_MICorrect2(void)	
{
  	
	cout << "Mutual information generated by no errors or not correlated errors of the 7 random variables with ";
	cout << nbins << " bins: ";
	
	cout.setf(ios::fixed, ios::floatfield);
  cout.setf(ios::showpos);
  cout.precision(5); 
	 	  
	cout << MICorrect2 << endl << endl;
}


/// Print to stdout the mutual information error generated by correlated errors
template <int sz> void MIError7<sz>::Print_MIErr2(void)	
{
  	
	cout << "Mutual information generated by correlated errors of the 7 random variables with ";
	cout << nbins << " bins: ";
	
	cout.setf(ios::fixed, ios::floatfield);
  cout.setf(ios::showpos);
  cout.precision(5); 
	 	  
	cout << MIErr2 << endl << endl;
}



// It calculates the component of mutual information generated by no errors 
template <int sz> float MIError7<sz>::Calc_MICorrect(void)
{ 
	float prodp, cjoint;
	prodp = p[0][0]*p[1][0]*p[2][0]*p[3][0]*p[4][0]*p[5][0]*p[6][0];
	cjoint = Pjoint[0][0][0][0][0][0][0]; // correct values for all variables
	if (cjoint>0)
	   MICorrect = (cjoint * log (cjoint/prodp) / LOG_NAT_TO_2_FACT);
	return MICorrect;
}


/// Print to stdout the mutual information MICorrect generated by no errors 
template <int sz> void MIError7<sz>::Print_MICorrect(void)	
{
  	
	cout << "Mutual information generated by no errors of the 7 random variables with ";
	cout << nbins << " bins: ";
	
	cout.setf(ios::fixed, ios::floatfield);
  cout.setf(ios::showpos);
  cout.precision(5); 
	 	  
	cout << MICorrect << endl << endl;
}


// It calculates the component of mutual information generated by  errors. 
template <int sz> float MIError7<sz>::Calc_MIErr(void)
{
  MIErr = Calc_Minfo() - Calc_MICorrect();
	return MIErr;
}


/// Print to stdout the mutual information error generated by  errors
template <int sz> void MIError7<sz>::Print_MIErr(void)	
{
  	
	cout << "Mutual information generated by  errors of the 7 random variables with ";
	cout << nbins << " bins: ";
	
	cout.setf(ios::fixed, ios::floatfield);
  cout.setf(ios::showpos);
  cout.precision(5); 
	 	  
	cout << MIErr << endl << endl;
}

// =========================================================================================
/// NB:
/// Per ogni tipo istanziato e' necessaria una istanziazione specifica


template class MutualInfo7<2>;
template class MutualInfo7<3>;
template class MutualInfo7<4>;
template class MutualInfo7<5>;
template class MutualInfo7<6>;
template class MutualInfo7<7>;
template class MutualInfo7<8>;
template class MutualInfo7<9>;
template class MutualInfo7<10>;

template class MIError7<2>;
template class MIError7<3>;
template class MIError7<4>;
template class MIError7<5>;
template class MIError7<6>;
template class MIError7<7>;
template class MIError7<8>;
template class MIError7<9>;
template class MIError7<10>;


/*
typedef MutualInfo7<2> MInfo_2;
typedef MutualInfo7<3> MInfo_3;
typedef MutualInfo7<4> MInfo_4;
typedef MutualInfo7<5> MInfo_5;
typedef MutualInfo7<6> MInfo_6;
typedef MutualInfo7<7> MInfo_7;
typedef MutualInfo7<8> MInfo_8;
typedef MutualInfo7<9> MInfo_9;
typedef MutualInfo7<10> MInfo_10;

typedef MIError7<2> MIErr_2;
typedef MIError7<3> MIErr_3;
typedef MIError7<4> MIErr_4;
typedef MIError7<5> MIErro_5;
typedef MIError7<6> MIErr_6;
typedef MIError7<7> MIErr_7;
typedef MIError7<8> MIErr_8;
typedef MIError7<9> MIErr_9;
typedef MIError7<10> MIErr_10;
*/
