// "pnd_cv.cc"


#include<stdlib.h>
#include"utils.h"
#include"tset.h"
#include"act_fun.h"
#include"polidico.h"
#include"appl.h"


static int num_attr = NATTR;    
static int nclass = NCLASS;  
static unsigned ndico = 0;  
static unsigned nfold = NFOLD;  
static unsigned nlayers = NLAYERS;  
static unsigned nhidden1 = NHIDDEN1;  
static unsigned nhidden2 = NHIDDEN2;  
static char namefile[128];  
static char resultfile[128];   
static char savefile[128];   
static char outfile[128];  
static char DMfile[128];  
static bool DMfile_is_present = false; 
static unsigned maxits = MAXITS;  
static double errtol = ERRTOL;     
static int learntype = LEARNTYPE;   
static int learnmode = LEARNMODE;  
static double eta = ETA;    
static double alpha = ALPHA;   
static double incr = INCR;   
static double decr = DECR;   
static double decrmom = DECRMOM;   
static unsigned savenet = 0;   
static unsigned saveout = 0;   
static unsigned decomptype = OPC; 
static int seed = 0;  
static bool b_printout = true;
static void print_info_use(void);
static int doarg(char* arg[], int i);
int main(int argc, char* argv[])
{
  int num_train, i;  
  Learning* gd;
  OneLayer* net;
  unsigned iter = 0;  
  double err = 0.0;   
  if (argc < 2) 
     print_info_use();
  if ((argc%2) == 1)
     print_info_use();
  strcpy(namefile, argv[1]);
  strcpy(resultfile, argv[1]);
  strcpy(savefile, argv[1]);
  strcpy(outfile, argv[1]);
  for (i=2; i<argc; i+=2)
         if (doarg(argv, i))
              print_info_use(); 
	if (ndico == 0)
	  ndico = DoDefaultcw(nclass); 
	if ((decomptype == DEC_FILE) && (!DMfile_is_present)) 
	  error_exit("Decomposition matrix file not specified");
			
  PoliDico<d_OPC>*       poliOPC;
  PoliDico<d_PWC>*       poliPWC;
  PoliDico<d_CC>*        poliCC;
  PoliDico<d_ECOC_ex>*   poliECOC_ex;
  PoliDico<d_ECOC_BCH>*  poliECOC_BCH;
	PoliDico<d_Rand>*      poliECOC_Rand;
  sigmoid_init(1.0);
  tanhyp_init(0.8);
	unsigned num_neuro[3] = {num_attr, nhidden2, nhidden1};
	act_func t_fun [3] = {sigmoid, sigmoid, sigmoid}; 
	DicoType dico(nlayers, num_neuro, t_fun);
	 
	switch (decomptype) {
	  case OPC: poliOPC  = new PoliDico<d_OPC>(nclass, dico, sigmoid);
	            cout << "OPC-PND" << endl;
	            cout << "Number of classes = " << poliOPC->Read_n_class() << endl;
	            cout << "Number of dichotomizers = " << poliOPC->Read_n_dico() << endl;
	            break;
	  case PWC: poliPWC  = new PoliDico<d_PWC>(nclass, dico, sigmoid);
	            cout << "PWC-PND" << endl;
	            cout << "Number of classes = " << poliPWC->Read_n_class() << endl;
	            cout << "Number of dichotomizers = " << poliPWC->Read_n_dico() << endl;
	            break;
	  case CC:  poliCC  = new PoliDico<d_CC>(nclass, dico, sigmoid);
	            cout << "CC-PND" << endl;
	            cout << "Number of classes = " << poliCC->Read_n_class() << endl;
	            cout << "Number of dichotomizers = " << poliCC->Read_n_dico() << endl;
	            break;
	  case ECOC_ex:  
	            poliECOC_ex  = new PoliDico<d_ECOC_ex>(nclass, dico, sigmoid);
	            cout << "ECOC-PND exhaustive" << endl;
	            cout << "Number of classes = " << poliECOC_ex->Read_n_class() << endl;
	            cout << "Number of dichotomizers = " << poliECOC_ex->Read_n_dico() << endl;
	            break;
	  case ECOC_BCH:  
	            poliECOC_BCH  = new PoliDico<d_ECOC_BCH>(nclass, dico, sigmoid, ndico);
	            cout << "ECOC-PND BCH" << endl;
	            cout << "Number of classes = " << poliECOC_BCH->Read_n_class() << endl;
	            cout << "Number of dichotomizers = " << poliECOC_BCH->Read_n_dico() << endl;
	            break;
		case DEC_FILE:  
	            poliECOC_Rand  = new PoliDico<d_Rand>(dico, sigmoid, DMfile);
	            cout << "ECOC-PND Random" << endl;
	            cout << "Number of classes = " << poliECOC_Rand->Read_n_class() << endl;
	            cout << "Number of dichotomizers = " << poliECOC_Rand->Read_n_dico() << endl;
	            break;
		default:  cout << "Not known PND" << endl;
	            exit(1);
	};
	
	 
	cout << "Number of folds: " << nfold << endl;
	
	 
	cout << "Dichotomizers features:" << endl;
	cout << "Number of layers: " << nlayers << endl;
	cout << "Input dimension: " << num_attr << endl;
	if (nlayers == 2)
	  cout << "Units of the hidden layer: " << nhidden2 << endl;
	else if (nlayers == 3)  {
	  cout << "Units of the I hidden layer: " << nhidden1 << endl;
		cout << "Units of the II hidden layer: " << nhidden2 << endl;
  }
	
	 
	cout << "Stop conditions:" << endl;
	cout << "Maximum number of iterations : " << maxits << endl;  
	cout << "RMS error threshold : " << errtol << endl; 
  cout << "Learning algorithm: ";
  switch (learntype) {
	  case GRADIENT_DESCENT: 
	            gd = new GradientDescent(eta);
	            cout << "Alg. Gradient descent" << endl;
	            cout << "Learning rate : " << eta << endl;
	            break;
	  case GRADIENT_DESCENT_DECR_LIN: 
	            gd = new GradDescDecrEtaLin(eta, decrmom);
	            cout << "Alg. Gradient descent linear decrement" << endl;
	            cout << "Learning rate : " << eta << endl;
	            cout << "Linear decrement rate : " << decrmom << endl;
	            break;
	  case GRADIENT_DESCENT_DECR_EXP: 
	            gd = new GradDescDecrEtaExp(eta, decrmom);
	            cout << "Alg. Gradient descent exponential decrement" << endl;
	            cout << "Learning rate : " << eta << endl;
	            cout << "Exponential decrement rate : " << decrmom << endl;
	            break;
	  case MOMENTUM_DESCENT:
	            switch(nlayers)  {
	              case 1: 
	                      gd = new MomentumDescent (1, num_attr, eta, alpha);
	                      break;
	              case 2: 
	                      gd = new MomentumDescent(1, nhidden2, num_attr, eta, alpha);
	                      break;
	              case 3: 
	                      gd = new MomentumDescent(1, nhidden1, nhidden2, num_attr, eta, alpha);
	                      break;
	            }
	            cout << "Alg. Gradient descent momentum" << endl;
	            cout << "Learning rate : " << eta << endl;
	            cout << "Momentum coefficient : " << alpha << endl;
	            break;
	  case MOMENTUM_DESCENT_DECR_LIN:
	            switch(nlayers)  {
	              case 1: 
	                      gd = new MomentumDescentDecrEtaLin(1, num_attr, eta, alpha, decrmom);
	                      break;
	              case 2: 
	                      gd = new MomentumDescentDecrEtaLin(1, nhidden2, num_attr, eta, alpha, decrmom);
	                      break;
	              case 3: 
	                      gd = new MomentumDescentDecrEtaLin(1, nhidden1, nhidden2, num_attr, eta, alpha, decrmom);
	                      break;
	            }
	            cout << "Alg. Momentum descent linear decrement" << endl;
	            cout << "Learning rate : " << eta << endl;
	            cout << "Momentum coefficient : " << alpha << endl;
	            cout << "Linear decrement rate : " << decrmom << endl;
	            break;
	  case MOMENTUM_DESCENT_DECR_EXP:
	            switch(nlayers)  {
	              case 1: 
	                      gd = new MomentumDescentDecrEtaExp(1, num_attr, eta, alpha, decrmom);
	                      break;
	              case 2: 
	                      gd = new MomentumDescentDecrEtaExp(1, nhidden2, num_attr, eta, alpha, decrmom);
	                      break;
	              case 3: 
	                      gd = new MomentumDescentDecrEtaExp(1, nhidden1, nhidden2, num_attr, eta, alpha, decrmom);
	                      break;
	            }
	            cout << "Alg. Momentum descent exponential decrement" << endl;
	            cout << "Learning rate : " << eta << endl;
	            cout << "Momentum coefficient : " << alpha << endl;
	            cout << "Exponential decrement rate : " << decrmom << endl;
	            break;
	  case BOLD_DRIVER: 
	            gd = new BoldDriver(eta, incr, decr);
	            cout << "Alg. Bold Driver" << endl;
	            cout << "Initial learning rate : " << eta << endl;
	            cout << "Increment rate : " << incr << endl;
	            cout << "Decrement rate : " << decr << endl;
	            break;
	  case BOLD_DRIVER_MOMENTUM:
	            switch(nlayers)  {
	              case 1: 
	                      gd = new BoldMomentum(1, num_attr, eta, alpha, incr, decr);
	                      break;
	              case 2: 
	                      gd = new BoldMomentum(1, nhidden2, num_attr, eta, alpha, incr, decr);
	                      break;
	              case 3: 
	                      gd = new BoldMomentum(1, nhidden1, nhidden2, num_attr, eta, alpha, incr, decr);
	                      break;
	            }
	            cout << "Alg. Bold Driver with momentum" << endl;
	            cout << "Initial learning rate : " << eta << endl;
	            cout << "Momentum coefficient : " << alpha << endl;
	            cout << "Increment rate : " << incr << endl;
	            cout << "Decrement rate : " << decr << endl;
	            break;
	  default:  cout << "Not known learning algorithm" << endl;
	            exit(1);
	};
  LearnType::learn_mode mode;
  if (learnmode == BYPATT) {   
     mode = LearnType::BYPATTERN;
     cout << "By pattern learning." << endl;
  }
  else {
     mode = LearnType::BYEPOCH; 
     cout << "By epoch learning." << endl;
  }    
  LearnType learninfo(gd, maxits, errtol, mode);
  switch (decomptype) {
	  case OPC: poliOPC->CrossValidate(namefile, nfold, num_attr, resultfile, 
	                                   learninfo, outfile, saveout, seed, b_printout);
	            delete poliOPC;
	            break;
	  case PWC: poliPWC->CrossValidate(namefile, nfold, num_attr, resultfile, 
	                                   learninfo, outfile, saveout, seed, b_printout);
	            delete poliPWC;
	            break;
	  case CC:  poliCC->CrossValidate(namefile, nfold, num_attr, resultfile, 
	                                  learninfo, outfile, saveout, seed, b_printout);  
	            delete poliCC;
							break;
	  case ECOC_ex:  
	            poliECOC_ex->CrossValidate(namefile, nfold, num_attr, resultfile, 
	                                       learninfo, outfile, saveout, seed, b_printout);  
	            delete poliECOC_ex;
							break;
		case ECOC_BCH:  
	            poliECOC_BCH->CrossValidate(namefile, nfold, num_attr, resultfile, 
	                                        learninfo, outfile, saveout, seed, b_printout);  
	            delete poliECOC_BCH;
							break;
		case DEC_FILE:  
	            poliECOC_Rand->CrossValidate(namefile, nfold, num_attr, resultfile, 
	                                        learninfo, outfile, saveout, seed, b_printout);  
	            delete poliECOC_Rand;
							break;
	  default:  cout << "Not known PND" << endl;
	            exit(1);
	}  
  delete gd;
  return 0;
}
void print_info_use(void)
{
  cerr << "
  pnd_cv. NEURObjects application for training and testing
          PND, using cross-validation techniques.			
  Usage: pnd_cv datafile [options]
  datafile           base name of the files used for cross validation
	
  Options:
 -res string         file summarizing cross-validation results    
 -decomp string      Decomposition schemes:  
                     OPC        (one per class)  
                     PWC        (pairwise coupling)  
                     CC         (correcting classifiers)
                     ECOC_ex    (Error Correcting Ouput Code - exaustive algorithm)
                     ECOC_BCH   (Error Correcting Ouput Code - BCH algorithm)
                     DEC_FILE   (Decomposition matrix read from a file)
 -DMf string       file of the decompositon matrix (DEC_FILE option required)
 -ndico unsigned     codeword length  
 -nc unsigned        number of the classes 
 -nf unsigned        number of the folds 
 -nl unsigned        number of the layers (considering also the input layer):                         
                     1: one layer (no hidden units) 
                     2: two layers (1 hidden layer) 
                     3: three layers (2 hidden layers)                      
 -h1 unsigned        number of hidden units of the first hidden layer of the three layers MLP 
 -h2 unsigned        number of hidden units of the second hidden layer of the three layers MLP or
                     number of hidden units of the two layers MLP (default 4)
 -d  unsigned        sample dimension/number of input units 
 -maxit unsigned     maximal number of iterations of the backpro algorithm  (default 500) 
 -maxerr double      maximal normalized root mean square error (RMS) (default 0.1) 
 -alg string         Backpropagation learning algorithm type:                      
                     gd         (gradient descent)  
                     gd_dl      (gradient descent linear decrement) 
                     gd_de      (gradient descent exponential decrement) 
                     md         (gradient descent with momentum) 
                     md_dl      (gradient descent with momentum and linear decrement) 
                     md_de      (gradient descent with momentum and exponential decrement) 
                     bold       (bold driver) 
                     boldmom    (bold driver with momentum)
 -rate double        learning rate (default 0.07) 
 -mom double         momentum rate (default 0.5) 
 -incbold double     bold driver increment rate 
 -decbold double     bold driver decrement rate 
 -decr double        gradient decrement  (for gradient descent with momentum and linear or exponential
                     decrement) 
 -seed               seed for the random initialization of the weights. 
                     If 0 the initialization is performed using current computer time.  
 -out string         save pnd outputs onto a file 
 -p integer          1 = printing a point at each iteration; 0 = noprint
  " << endl;
  exit(1);
}  
int doarg(char* arg[], int i)
{
  int j;
	
  if (!strcmp(arg[i], "-res"))
    strcpy(resultfile, arg[i+1]);
  else if (!strcmp(arg[i], "-decomp")) { 
    if ((decomptype = Analyze_dec(arg[i+1])) >= NUMDEC)
       print_info_use();
  }
  else if (!strcmp(arg[i], "-nc")) {
    nclass = atoi(arg[i+1]);
		if (nclass < 2) error_exit("Classes must be 2 or more");}
  else if (!strcmp(arg[i], "-ndico"))
    ndico = atoi(arg[i+1]);
	else if (!strcmp(arg[i], "-nf")) {
    nfold = atoi(arg[i+1]);
		if (nfold < 2) error_exit("Folds must be 2 or more");}
  else if (!strcmp(arg[i], "-nl")) {
    nlayers = atoi(arg[i+1]);
		if ((nlayers < 1) || (nlayers > 3)) error_exit("Number of layers must be between 1 and 3");}
  else if (!strcmp(arg[i], "-h1"))
    nhidden1 = atoi(arg[i+1]);
  else if (!strcmp(arg[i], "-h2"))
    nhidden2 = atoi(arg[i+1]);
  else if (!strcmp(arg[i], "-d")) {
    num_attr = atoi(arg[i+1]);
		if (num_attr < 1) error_exit("Number of attributes must be 1 or more");}
  else if (!strcmp(arg[i], "-maxit"))
    maxits = atoi(arg[i+1]);
  else if (!strcmp(arg[i], "-maxerr"))
    errtol = atof(arg[i+1]);
  else if (!strcmp(arg[i], "-alg")) { 
    if ((learntype = Analyze_alg(arg[i+1])) >= NUMALG)
       print_info_use();
  }
  else if (!strcmp(arg[i], "-mode")) {
    if ((learnmode = Analyze_mode(arg[i+1])) >= NUMMODE)
       print_info_use();
  }
  else if (!strcmp(arg[i], "-rate"))
    eta = atof(arg[i+1]);
  else if (!strcmp(arg[i], "-mom"))
    alpha = atof(arg[i+1]);
  else if (!strcmp(arg[i], "-incbold"))
    incr = atof(arg[i+1]);
  else if (!strcmp(arg[i], "-decbold"))
    decr = atof(arg[i+1]);
  else if (!strcmp(arg[i], "-decr"))
    decrmom = atof(arg[i+1]);
  else if (!strcmp(arg[i], "-out")) {
    saveout = 1; 
    strcpy(outfile, arg[i+1]);
  }
	else if (!strcmp(arg[i], "-p")) {
    j = atoi(arg[i+1]);
		if (j == 0)
		   b_printout = false;
	}
	else if (!strcmp(arg[i], "-DMf")) {
    strcpy(DMfile,arg[i+1]);
		DMfile_is_present = true;
	}
	else if (!strcmp(arg[i], "-seed"))
    seed = atoi(arg[i+1]);
  else 
    return 1;
  return 0;
} 
