// polidico.h
  

#ifndef POLIDICO_H
#define POLIDICO_H

#pragma interface
#include<iostream>
#include<time.h>
#include<iomanip>
#include<vector>
#include"act_fun.h"
#include"utils.h"
#include"tset.h"
#include"decomp.h"
#include"matrix.h"
#include"dico.h"

static const char * const POUT_EXT = ".pout";


/**
  Class for storing the structure of a dichotomizer.\\
	@memo Class for storing the structure of a dichotomizer -   HTML documentation under construction.
  @see Dichotomizers
  @see PoliDico
*/
class DicoType  {
  
public:
  
  unsigned n_layers;
  
  unsigned n_neuro[MAXLAYERS];
  
  act_func type_fun[MAXLAYERS];
  
  
 
  DicoType( unsigned nlayers, unsigned num_neuro[], act_func t_fun[])
  { 
    int i;
    n_layers = nlayers;
    for (i=0; i<nlayers; i++)  {
      n_neuro[i] = num_neuro[i];
      type_fun[i] = t_fun[i];
    }
  }
  
 
  DicoType(const DicoType& d)
  { 
    int i;
    n_layers = d.n_layers;
    for (i=0; i<n_layers; i++)  {
      n_neuro[i] = d.n_neuro[i];
      type_fun[i] = d.type_fun[i];
    }
  }
  
  
  DicoType& operator=(const DicoType& d)
  { 
    int i;
    n_layers = d.n_layers;
    for (i=0; i<n_layers; i++)  {
      n_neuro[i] = d.n_neuro[i];
      type_fun[i] = d.type_fun[i];
    }
    return *this;
  }
};

    


/**
  Class for implememting  Parallel Non linear Dichotomizers (PND).\\
	It implements different PND using different decomposition schemes.\\
	@memo PND ensemble -  HTML documentation under construction.
*/

template <class SCOMP>
class PoliDico {
  
private:
  
 
  unsigned n_class;
  
 
  unsigned n_dico;
  
  
  SCOMP scomp;
 
  Dico1 ** vDico;
  
  // Number of errors on the test set using L1 norm
  unsigned num_errors;
	
	// Number of errors on the test set using L2 norm
  unsigned num_errorsL2;
	
	// Number of errors on the test set using Hamming distance
  unsigned num_errorsHD;
  
 
  double perc_errors;
  
 
  unsigned iter_tot;
  
  
  act_func funout;
  
  
  vect* err;
  
  
  vector<unsigned>* iter;
  
  static const char SEP = ' ';

  
  void Build_dico(DicoType& dico);
  
  
  void print_dico_err_iteraz_info(unsigned n, unsigned iter_dico);
  
  
  void TransformMatrix(matrix& m);
  
  
  unsigned Calc_class(vect& f, matrix& DM, vect& input, double& dist);
	
	unsigned Calc_class_Hamming(vect& f, matrix& DM, vect& input, double& dist);
	
	unsigned Calc_class_L2Norm(vect& f, matrix& DM, vect& input, double& dist);                                                         
	
	// computes the errors of the dichotomizers on a single example
	void Calc_dico_errors(vect& f, matrix& DM, unsigned targetclass, double threshold);
	
	// Initializes the array sDicoerr to zero.
	void Init_vDicoerr(void);
	
	// pointer to a vector of the PND errors (considering L1 norm) on different folds
	unsigned * abserr;
	
	// pointer to a vector of the PND errors (considering L2 norm) on different folds
	unsigned * abserrL2;
	
	// pointer to a vector of the PND errors (considering Hamming distance) on different folds
	unsigned * abserrHD;
	
	// pointer to a vector of the percentual PND errors (considering L1 norm) on different folds
	float * percerr;
	
	// Vector of the dichotomizers errors. It stores the number of errors of each dichotomizers on all the folds 
	unsigned * vDicoerr;
  
  // Print results of each fold of CV onto a stream
  void PrintFoldResults (ostream& out, unsigned nfold, vector<unsigned>& itercv);
  
  // Computes the total errors of CV using L1, L2 norm and Hamming distance
  void Calc_total_cv_errors (unsigned nfold, vector<unsigned>& itercv, unsigned& itertot, float& errtot, float& errtotL2, float& errtotHD);
  
  // Computes the standard deviation of error of CV 
  float Calc_cv_stdev (int nfold, int ntrain, float err_rate, unsigned* fold_tot_err);
  
  // Computes accuracy of each dichotomizer with cross validation
  void Calc_cv_dico_accuracy (int ntrain, float& minerrordico, float& maxerrordico, float& avgerrordico, float& stdverrordico);
  
  // Copmutes the minimun Hamming distance etc. of the decomposition
  void Calc_HD (float& MHD, float& AHD, float& ERP, float& MDD, float& ADD);
  
public:  

	PoliDico(){}

	
	
	PoliDico (unsigned num_class, DicoType& dico, act_func funact,  unsigned num_dico = 0);
	
	
	PoliDico (DicoType& dico, act_func funact,  char* DMfile);
	
	
	~PoliDico();  
	
	unsigned Read_n_dico(void) const {return n_dico;}
	
	
  unsigned Read_n_class(void) const {return n_class;}
  
  
  void print_info(void);
  
  
  void init_weights_norm_poli (int seed = 0)
  {
    unsigned i;
    for (i=0; i<n_dico; i++)
      vDico[i]->init_weights_norm(seed);
  }
  
  
  
  void init_weights_const_poli (double x)
  {
    unsigned i;
    for (i=0; i<n_dico; i++)
      vDico[i]->init_weights_const(x);
  }
  
                  
 
  int LearnPoli (TrainingSet& tset, LearnType& learninfo, bool printout=true);
  
  
  void save_decomp_matrix(ofstream& f);
  
 
  void save_out(TrainingSet& tset, ofstream& f);
  
 
  void save_out(TrainingSet& tset, char* outfile);
  
  
  void SavePoli (char *namefile);
  
  
  
  void save_errors(char* name);
  
 
  matrix test(TrainingSet& tset);
  
  
  int CrossValidate(char *name, unsigned nfold, unsigned num_attr, 
                     char *filerr, LearnType& learninfo, 
                     char* outfile = "", int saveout = 0, int seed=0, bool printout=true);
  
  
                       
  void SaveCV (char *name, vector<unsigned>& itercv, unsigned ntrain, matrix& conf, unsigned nfold, clock_t start, clock_t end);
  
                            
  void print_resultCV (vector<unsigned>& itercv, unsigned ntrain, unsigned nfold, clock_t start, clock_t end);  
  
  
  void print_errors(void)
  {  
    cout << "PND errors:" << endl;
    cout.precision(4);
    cout << setw (20) << "Number of errors : " << num_errors << " : Error rate % : " << perc_errors*100 << " : ";
  }
  
  
  double Read_err_dico(int i)
  { return (*err)[i]; }
	
	
	SCOMP& Read_scomp(void) {return scomp;}
   
};

#endif
