// randomsubspace.cc

#include<strstream.h>
#include"randomsubspace.h"
#include"utils.h"

RandomSubspace::RandomSubspace() : Bagging()
{ }
										 
/**
	 Constructor.
	 @param baselearner base learner to be used by the bagging algorithm
	 @param nclass number of classes;
*/
RandomSubspace::RandomSubspace(BaseLearner* baselearner, unsigned num_class) :
  Bagging(baselearner, num_class)
{ }

/**
	 Copy Constructor.
*/
RandomSubspace::RandomSubspace(const RandomSubspace& Rs) {
 // n_attr = Bag.n_attr;
 nclass = Rs.nclass;
 confusion = Rs.confusion;
 
 // copy of the base learners of Rs into this
 deque<BaseLearner*>::const_iterator it;
 for (it = Rs.baselist.begin(); (it != Rs.baselist.end()); it++) {
	 baselist.push_back((*it)->clone());
 }  
 
 modellist = Rs.modellist;
 base_learner_train_errors = Rs.base_learner_train_errors;
 base_learner_test_errors = Rs.base_learner_test_errors;
 ensemble_errors = Rs.ensemble_errors;
 maxit = Rs.maxit;
 current_it = Rs.current_it;
 base = Rs.base;
 trainset = Rs.trainset;
 testset = Rs.testset; 
}



/* Destructor.*/
RandomSubspace::~RandomSubspace() 
{ }


/*
	 Test the trained base learners stored in files.
*/
void RandomSubspace::test_base_learners(char* testfile, unsigned startiteration, 
                                unsigned maxiterations, char* dir_res) {
	
	unsigned const SIZE = 256;
	char model[SIZE];
	char results[SIZE];
	char testrs[SIZE];
	float err;
	unsigned k;		
	
	ostrstream omodel(model,SIZE,ios::out);
	ostrstream oresults(results,SIZE,ios::out);
	ostrstream otest(testrs,SIZE,ios::out);
	// clear the  error list if startieration is not consecutive to the previous
	if ((base_learner_test_errors.size() + 1) != startiteration)
	  base_learner_test_errors.clear();
	// It permits testing using models and results starting at an arbitrary number
	if (baselist.size() <= (maxiterations-startiteration))
	  error_exit("Bagging::test_base_learners : number of base learners not sufficient");
	if (baselist.size() == (maxiterations-startiteration+1))
	  k=0;
	else
	  k=startiteration-1;
	// testing of the base learners  
	for (int i=startiteration; i<=maxiterations; i++) {	   
		 omodel.seekp(0,ios::beg);	
		 oresults.seekp(0,ios::beg);
		 otest.seekp(0,ios::beg);
		 //if ((i>modellist.size()) || (i<1))
		 //  error_exit("Bagging::test_base_learners : modellist out of range");	 		 
		 omodel << modellist[k] << ends;
		 otest << testfile << i << ends;
     if (!strcmp(dir_res,""))
       oresults  << BASRES << i << ends;
     else
		   oresults << dir_res << "/" << BASRES << i << ends;
		 //if ((i>baselist.size()) || (i<1))
		 //  error_exit("Bagging::test_base_learners : baselist out of range");	
	   err = (baselist[k++])->test_learner(testrs, model, results);
		 base_learner_test_errors.push_back(err);
	}
}


