// roc.cc


#include<cstdlib>
#include"roc.h"

void ROCvalue::print_ROC() const {
  cout.setf(ios::fixed, ios::floatfield);
  cout.setf(ios::showpoint);
  cout.precision(4);
	cout.width(8);
	cout << comp_specificity;
	cout.precision(4);
	cout.width(8);
	cout << sensitivity << endl;
}


void ROCvalue::print_all() const {
  cout.setf(ios::fixed, ios::floatfield);
  cout.setf(ios::showpoint);
  cout.precision(4);
	cout.width(8);
	
	cout << sensitivity;
	cout.precision(4);
	cout.width(8);
	cout << specificity;
	cout.precision(4);
	cout.width(8);
	cout << comp_specificity;
	cout.precision(4);
	cout.width(8);
	cout << precision << endl;
}


void ROC::init_conf_matrix(void) {
  for (int i=0; i<size_matrix; i++)
	  for (int j=0; j<size_matrix; j++)
		  conf_matrix [i][j] = 0;
}
  

float ROC::calc_min(void) {
 float min = predicted[0];
 for (int i=1; i < predicted.size(); i++) {
   if (predicted[i]<min)
	   min = predicted[i];
 }
 return min;
}


float ROC::calc_max(void) {
 float max = predicted[0];
 for (int i=1; i < predicted.size(); i++) {
   if (predicted[i]>max)
	   max = predicted[i];
 }
 return max;
}


void ROC::read_data(char * filename) {
  float x,y;
	
	ifstream f(filename); 
  if (!f) {
	   cerr << "Can't open data file :" << filename << endl;
		 exit(1);
		 }
		 
	while (f >> x) {
	  f >> y;
		predicted.push_back(x);
		target.push_back(y);
		numdata++;
	}
	
	if (f.bad()) {
	   cerr << "Error reading file" << filename << endl;
		 exit(1);
		 }
}


void ROC::read_data_2(char * pred_file, char * target_file) {
  float x,y;
  ifstream f;
  
	// read the prediction file
  f.open(pred_file); 
  if (!f) {
	   cerr << "Can't open data file :" << pred_file << endl;
		 exit(1);
		 }
		 
	while (f >> x) 
		predicted.push_back(x);
	
	if (f.bad()) {
	   cerr << "Error reading file" << pred_file << endl;
		 exit(1);
		 }
	f.close();
	
	// read the target file
	f.open(target_file); 
  if (!f) {
	   cerr << "Can't open data file :" << target_file << endl;
		 exit(1);
		 }
		 
	while (f >> x) 
		target.push_back(x);
	
	if (f.bad()) {
	   cerr << "Error reading file" << target_file << endl;
		 exit(1);
		 }
	f.close();	 
	if (predicted.size() != target.size()) {
	   cerr << "Prediction and target file size does not match" << endl;
		 exit(1);
		 }
	else
	   numdata = predicted.size();
}

void ROC::do_conf_matrix(float thresh) {
  threshold = thresh;
	int response = 0;
	int expected = 0;
	
	init_conf_matrix();
	for (int i=0; i<numdata; i++) {
	  if (predicted[i] >= threshold)
		  response = 0;
		else
		  response = 1;
		if (target[i] == +1)
		  expected = 0;
		else
		  expected = 1;
	  conf_matrix[response][expected]++;
	}
}
		
  
ROCvalue * ROC::compute_ROC_value (float thresh) {
  ROCvalue * v = new ROCvalue();
	do_conf_matrix(thresh);
	float TP,TN,FP,FN;
	TP = conf_matrix[0][0];
	FP = conf_matrix[0][1];
	FN = conf_matrix[1][0];
	TN = conf_matrix[1][1];
	v->sensitivity = TP/(TP+FN);
	v->specificity = TN/(TN+FP);
	v->comp_specificity = 1 - v->specificity;
	v->precision = TP/(TP+FP);
	return v;
}


void ROC::compute_ROC_curve (int num_interval) {
  float min, max, intwidth, t;
	ROCvalue * v;
	
	min = calc_min();
	max = calc_max();
	intwidth = (max-min)/num_interval;
	
	t=max+1;
	v = compute_ROC_value (t);
	rocvalues.push_back(*v);
	t=max;
	for (int i=0; i < num_interval-1; i++) {
	  t -= intwidth;
		v = compute_ROC_value (t);
	  rocvalues.push_back(*v);
	}
	t=min;
  v = compute_ROC_value (t);
	rocvalues.push_back(*v);
}


void ROC::print_ROC_curve (void) const {
  cout << "ROC curve values (1-specificity, sensitivity  pairs): " << endl << endl;
  for (int i=0; i<rocvalues.size(); i++)
	  rocvalues[i].print_ROC();
}


void ROC::print_ROC_curve_all (void) const {
  cout << "ROC curve values (sensitivity, specificity, 1-specificity, precision): " << endl << endl;
  for (int i=0; i<rocvalues.size(); i++)
	  rocvalues[i].print_all();
}


void ROC::print_predicted (void) const {
   cout << "Predicted values: " << endl << endl;
	 for (int i=0; i<predicted.size(); i++)
	   cout << predicted[i] << endl;
}

void ROC::print_target (void) const {
   cout << "Target values: " << endl << endl;
	 for (int i=0; i<target.size(); i++)
	   cout << target[i] << endl;
}
