#ifndef SNR_H
#define SNR_H

#include<iostream.h>
#include <fstream.h>
#include <strstream.h>
#include <string.h>
#include <ctype.h>
#include <vector.h>
#include <deque.h>
#include "matrix.h"
#include "tset.h"


/**
   Class implemeting the signal to noise ratio test for feature ranking and selection.\\
*/ 
class SNRTest {

public:
	/// vector of the classes random generated
	vect class_rand;
	/// vector of the SNR values of the features
	vect SNR;
	/// vector of the SNR values of the features after random permutation of the classes
	vect SNR_rand;
	/// vector of the ranked indices of the features
	vect ranked_indices;
	/// vector of the ranked SNR values of the features
	vect ranked_values;
	/// counter of the random SNR exceeding the computed SNR
	vect s_counter;
	/// number of the features
	unsigned n_attr;
	/// number of the patterns
	unsigned n_patt;
	/// queue list of the selected features
	deque<int> selected_features;
	/// data
	TrainingSet tset;
	

   /// Dummy constructor
   SNRTest (void) { n_attr = n_patt = 0;}
	 
	 /**
	 Constructor.
	 @param num_attr attribute cardinality
	 @param num_patt  cardinality of the patterns
	 @param file name of the input data file
	 */
   SNRTest ( unsigned num_attr, unsigned num_patt,  char *file);
	 
	 
	 /// Destructor
   virtual ~SNRTest (){};
	 
	 /**
	 Computes the SNR for a single feature
	 @param index index of the feature
	 @param cl class for which the SNR has to be computed
	 @param label vector of the classes
	 @return value of the computed SNR 
	 */
	 double single_feature_SNR_compute(int index, int cl, vect& label);
	 
	 
	  /**
	 Computes the SNR for all the features
	 @param cl class for which the SNR has to be computed
	 */
	 void SNR_compute(int cl);
	 
	 /**
	 Print the SNR for all the features
	 @param o output stream
	 */
	 void SNR_print(ostream& o);
	 
	  /**
	 Computes the SNR for all the features for a random permutation
	 @param cl class for which the SNR has to be computed
	 */
	 void SNR_rand_compute(int cl);
	 
	 /**
	 Print the SNR computed at random for all the features
	 @param o output stream
	 */
	 void SNR_rand_print(ostream& o);
	 
	  /**
	 Random permutation of the label of the classes
	 */
	 void permute_class_labels(void);
	 
	  /**
	 Performs random permutation test for SNR
	 @param n number of permutations
	 @param cl class for which the test is performed
	 */
	 void permutation_test(int n, int cl);
	 
	  /**
	 Selects a set of features
	 @param significance significance of the selected features
	 */
	 void select_features(float significance);
	 
	  /**
	 Saves the set of selected feature indices and their SNR values
	 @param indexfile file of the indices of the selected features
	 */
	 void save_feature_indices(char* indexfile);
	 
	  /**
	 Loads the set of selected feature indices from a file
	 @param indexfile file of the indices of the  features
	 */
	 void load_feature_indices(char* indexfile);
	 
	 /**
	 Save the currently selected set of features
	 @param patternfile file of the selected features
	 */
	 void save_selected_features(char* patternfile);
	 
	  /**
	 Save the currently selected set of features and their indices
	 @param indexfile file of the indices of the selected features
	 @param patternfile file of the selected features
	 */
	 void save_features(char* indexfile, char* patternfile);
	 
	  /**
	 Ranks the features according to their SNR 
	 */
	 void rank_features(void);
	 
	  /**
	 Prints the first n ranked features 
	 param n number of the features to be printed 
	 @param o output stream
	 */
	 void print_ranked_feature_indices(int n, ostream& o);
	 
	 /**
	 Save the first n ranked feature indices 
	 param n number of the features to be saved
	 @param ranked_indices_file file of the ranked indices
	 */
	 void save_ranked_feature_indices(int n, char* ranked_indices_file);
	 
	 /**
	 Save the first n ranked features 
	 param n number of the features to be saved 
	 @param ranked_feature_file file of the ranked features
	 */
	 void save_ranked_features(int n, char* ranked_feature_file);
	 
protected:	 
	 
	  
};













#endif
