/* ******************************************************************************
NEURObjects 2.02
stattest.h December 2002
Giorgio Valentini
****************************************************************************** */
#ifndef STATTEST_H
#define STATTEST_H

#include <iostream>
#include <fstream>
#include <string>
#include <deque>
#include <cstdlib>
#include <cstring>


using namespace std;

/**
  Cumulative distribution of the t Student distribution.\\
	Row 1 to 30 refers to the degree of freedom (that is number of trials minus one),  
	column 0 : 0.90 probability\\
	column 1 : 0.95 probability\\
	column 2 : 0.975 probability\\
	column 3 : 0.99 probability\\
*/
static const float t [][4] = { {3.078, 6.314, 12.706, 31.821},
                            {1.886, 2.920, 4.303, 6.965},
														{1.638, 2.353, 3.182, 4.541},
														{1.533, 2.132, 2.776, 3.747},
														{1.476, 2.015, 2.571, 3.365},
														
														{1.440, 1.943, 2.447, 3.143},
														{1.415, 1.895, 2.365, 2.998},
														{1.397, 1.860, 2.306, 2.896},
														{1.383, 1.833, 2.262, 2.821},
														{1.372, 1.812, 2.228, 2.764},
														
														{1.363, 1.796, 2.201, 2.718},
														{1.356, 1.782, 2.179, 2.681},
														{1.350, 1.771, 2.160, 2.650},
														{1.345, 1.761, 2.145, 2.624},
														{1.341, 1.753, 2.131, 2.602},
														
														{1.337, 1.746, 2.120, 2.583},
														{1.333, 1.740, 2.110, 2.567},
														{1.330, 1.734, 2.101, 2.552},
														{1.328, 1.729, 2.093, 2.539},
														{1.325, 1.725, 2.086, 2.528},
														
														{1.323, 1.721, 2.080, 2.518},
														{1.321, 1.717, 2.074, 2.508},
														{1.319, 1.714, 2.069, 2.500},
														{1.318, 1.711, 2.064, 2.492},
														{1.316, 1.708, 2.060, 2.485},
														
														{1.315, 1.706, 2.056, 2.479},
														{1.314, 1.703, 2.052, 2.473},
														{1.313, 1.701, 2.048, 2.467},
														{1.311, 1.699, 2.045, 2.462},
														{1.310, 1.697, 2.042, 2.457}
													};

/// P values for the McNemar's test
static const float pValues [] = {.001, .005, .010, .025, .050, 0.10};

/// Values of  chi square for the corresponding P values
static const float chiValues[] = {10.82757, 7.879439, 6.634897, 5.023886, 3.841459, 2.705543};

static const int  nValues = 6;

/**
   Class for performing the Mc Nemar's test.\\
*/ 
class McNemar {

protected:

/// value of the chi2 function for beta = 0.95 with 1 degree of freedom
static const float chi95 = 3.841459;


/// Contingency table
int cont_tab[2][2];

/// Number of incorrect predictions of the first classifier and correct of the second
int n01;

/// Number of correct predictions of the first classifier and incorrect of the second
int n10;

/// Container of the predictions of the first classifier
deque<int> pred1;

/// Container of the predictions of the second classifier
deque<int> pred2;

/// Container of the correct predictions
deque<int> target;

/// Result of the Pearson statistics
float gamma;

/**
	 Load the labels from a file.\\
   It loads the labels from a file, storing them in a container.
	 @param file file storing the labels
	 @param pred container where the labels will be loaded
*/
void load_labels(char* file, deque<int>& pred);

/// It builds the contingency table
void build_cont_table(void);

/**
   Computes the  Pearson statistic of the Mc Nemar's test.\\
   @return value of the computed Pearson statistic
*/
float compute_pearson(void);

/////////////////////////////////////////////////////
/////////////////////////////////////////////////////
public:


/// a beta value for chi2
static const float beta95 = 0.95;

/// Default constructor
McNemar(){}

/**
	 Constructor.
	 @param p1 file storing the prediction of the first classifier
	 @param p2 file storing the prediction of the second classifier
	 @param t file storing the correct predictions
*/
McNemar (char* p1, char* p2, char* t);


/// Destructor
~McNemar(){}

/**
   Test of the null hypothesis.\\
   The null hypothesis (that is the two classifiers are equivalent) is rejected 
   if the result of the Pearson statistic is larger than value of the chi square function with 1 degree
   of freedom and level of confidence 0.1.
   The actual level of confidence is returned in level. Level is set to 1 if the null hypothesis 
   cannot be rejected.
   @param level level of confidence
   @return true if the null hypothesis is rejected, otherwise false
*/
bool McNemar::Reject_null(float& level);

/// Returns the computed value of the Pearson statistics
float get_gamma(void){return gamma;}
};



/**
   Class for performing the resampled paired t  test.\\
*/ 
class ResPairT {

protected:


/// Container of the misclassifiaction rates of the first classifier
deque<float> p1;

/// Container of  the misclassifiaction rates of the second classifier
deque<float> p2;

/// Container of the differences between p1 and p2
deque<float> p;

/// Value of the t statistic
float t_stat;

/// Average value of the values stored in p
float p_avg;

/// Standard deviation of the values stored in p
float p_sd;

/// Degree of freedom of the t distribution (number of trials minus 1)
int n;
													
/**
	 Load the misclassification rates from a file.\\
   It loads the misclassification rates from a file, storing them in a container.
	 @param file file storing the misclassification rates
	 @param misrate container where the misclassification rates will be loaded
*/
void load_errors(char* file, deque<float>& misrate);

/**
	 Computes the differences betwemm misclasification rates.\\
	 It computes p = p1 - p2, for each element stored in the containers.
*/
void calc_p(void);

/**
	 Computes the average of the values stored in p.\\
	 @return average value of p
*/
float calc_avg(void);

/**
	 Computes the standard deviation of the values stored in p.
	 @return standard deviation of  p
*/
float calc_sd(void);





/////////////////////////
public:

/// SUpported degree of confidence
static const int c90 = 0;
static const int c95 = 1;
static const int c975 = 2;
static const int c99 = 3;



/**
	 Constructor.
	 @param f1 file storing the misclassification rates of the first classifier
	 @param f2 file storing the misclassification rates of the second classifier
	 @param num number of trials
*/
ResPairT (char* f1, char* f2);


/// Destructor
~ResPairT(){}

/**
	 Computes the t satistic.
	 @return value of the computed t statistic
*/
float calc_t_stat(void);

/**
   Test of the null hypothesis.\\
   The null hypothesis (that is the two classifiers are equivalent) is rejected with confidence degree
   if the result of the t statistic is larger than value of the cumulative t Student distribution with
	 confidence conf. Conf can assume the following values:\\
	 c90 = confidence 90% \\
	 c95 = confidence 95% \\
	 c975 = confidence 97.5% \\
	 c99 = confidence 99% \\ 
   @param conf confidence
   @return true if the null hypothesis is rejected, otherwise false
*/
bool Reject_null(int conf);


/**
   Test of the null hypothesis.\\
   It performs the resampled paired t test. If the null hypothesis can be rejected with confidence at
	 least 90 %, the corresponding value of the confidence is returned. Otherwise the value 0 is returned.
   @return degree of confidence at which the null hypothesis is rejected. If 0 the null hypothesis
	 cannot be rejected.
*/
float test(void);

/// Returns the computed value of the t statistic
float get_t(void){return t_stat;}

/// Returns the number of trials
int get_n(void){return n;}

};

#endif
