// tset.cpp
//Implementazione della classe TrainingSet e derivate

#include<ctype.h>
#include<stdlib.h>
#include<math.h>
#include<time.h>
#include<iostream.h>
#include<fstream.h>

#include<LEDA/matrix.h>
#include<LEDA/dictionary.h>

#include"tset.h"
#include"utils.h"


/* 
--------------------------------------------------------
   Variabili e costanti locali
--------------------------------------------------------
*/
static const unsigned BUFSIZE = 256;
static  const char SEPARATOR = ',';

/*
--------------------------------------------------------------------------------

   Costruttori

--------------------------------------------------------------------------------
*/



TrainingSet::TrainingSet ( unsigned num_attr, unsigned num_train, char separator = ',')
: data(num_train, num_attr), target(num_train), mean(num_attr), stdev(num_attr)
{
   n_attr = num_attr;
   n_train = num_train;
	 sep = separator;
   n_class = 0;
   is_normalized = 0;
}


TrainingSet::TrainingSet ( unsigned num_attr, unsigned num_train, char *file,  
                           char separator = ',', class_pos pos = last)
: data(num_train, num_attr), target(num_train), mean(num_attr), stdev(num_attr)
{
   n_attr = num_attr;
   n_train = num_train;
	 sep = separator;
   n_class = read(file, pos);
   is_normalized = 0;
}


/*
--------------------------------------------------------------------------------

   Distruttore

--------------------------------------------------------------------------------
*/

TrainingSet::~TrainingSet ()
{
   
}




/*
--------------------------------------------------------------------------------

   Legge il training set file e inizializza la matrice data con gli attributi
   del training set ed il vettore target con le classi target.
   Ritorna il numero delle classi lette.

--------------------------------------------------------------------------------
*/

unsigned TrainingSet::read(char *file, class_pos pos = last)
{
  int i,j;
  char c;
  unsigned n = 0;
  
  ifstream from(file); 
  if (!from) error("Cannot open input file : ", file);
  
  if (pos == last)
    {
  		for (i=0; i<n_train; i++)
    		{
    			for (j=0; j<n_attr; j++)
    			 {
      			from >> data(i,j) >> c;
      			if (c != sep)
        			error_exit("Error  reading separator");
    			 }
      		from >> target[i];
      		if (target[i] > n) n = (unsigned) target[i];					
    		}
    }
  else // pos == first
    {
  		for (i=0; i<n_train; i++)
    		{
    		  from >> target[i];
      		if (target[i] > n) n = (unsigned) target[i];
    			for (j=0; j<n_attr; j++)
    			 {
      			from >> c >> data(i,j);
      			if (c != sep)
        			error_exit("Error  reading separator");
    			 }      		
    		}
    }
     
  // salta eventuali caratteri rimasti (che devono essere solo blank)
  while (from.get(c));
          
  if (!from.eof()) 
     error("Error reading file ", file); 
  return n;
}




/*
--------------------------------------------------------------------------------
  Stampa la matrice degli attributi del training set
--------------------------------------------------------------------------------
*/
inline void TrainingSet::print_data()
{
  cout << "Matrix of attributes :" << endl;
   data.print();
}
  
/*
--------------------------------------------------------------------------------
  Stampa il vettore dei target del training set
--------------------------------------------------------------------------------
*/
inline void TrainingSet::print_target() 
{
  cout << "Target vector :" << endl; 
  cout << target << endl;
}


/*
--------------------------------------------------------------------------------
  Stampa il numero dei campioni, il numero egli attributi e delle classi del training set
--------------------------------------------------------------------------------
*/
inline void TrainingSet::print_info()
{
  cout << "General info on data set : "<< endl;
  cout << "Number of patterns : " << Ntrain() << endl;
  cout << "Number of attributes : " << Nattr() << endl;
  cout << "Number of classes : " << Nclass() << endl;
  mean_calc();
  stdev_calc();
  if (is_normalized)
      cout << "Normalized attributes " << endl;     
  else
      cout << "Not normalized attributes." << endl;      
  cout << "Mean value of the attributes: " << endl;
  mean.print(); cout << endl;
  cout << "Standard deviation of attributes: " << endl;
  stdev.print(); cout << endl;
    
}
  
/*
--------------------------------------------------------------------------------
  Calcola il valor medio degli attributi  del training set
--------------------------------------------------------------------------------
*/
void TrainingSet::mean_calc(void)
{
  vector v(n_attr);
  
  for (int i=0; i<n_train; i++)
     v = v + data.row(i);
  v = v * (1.0/n_train);
  mean = v;  
}


/*
--------------------------------------------------------------------------------
  Calcola la deviazione standard degli attributi  del training set
--------------------------------------------------------------------------------
*/
void TrainingSet::stdev_calc(void)
{
  vector v(n_attr);
  double x;
  int j;
  
  for (int i=0; i<n_train; i++)
    for (j=0; j<n_attr; j++)
      {
        x = data(i,j) - mean[j];
        v[j] += (x*x);
      }
  
  v = v * (1.0/(n_train-1));
  for (j=0; j<n_attr; j++)
    stdev[j] = sqrt(v[j]);
  
}



/*
--------------------------------------------------------------------------------
  Normalizza i dati di input:
  normalized_x = (x - mean(x)) / dev.standard
--------------------------------------------------------------------------------
*/
void TrainingSet::normalize(void) 
{
  vector v(n_attr);
	mean_calc();
  stdev_calc();
  int i, j;
  
	// avoid division by zero in case stdev = 0
	for (j=0; j<n_attr; j++)
	  if (stdev[j] == 0)
		  v[j] = 1.0;
		else
		  v[j] = stdev[j];
	
  for (i=0; i<n_train; i++)
     for (j=0; j<n_attr; j++)
       data(i,j) = (data(i,j) - mean[j]) / v[j];
         
  is_normalized = 1;
} 


/*
--------------------------------------------------------------------------------
  Normalizza i dati di input con media m e deviazione standard sv.\\
  normalized_x = (x - mean(x)) / dev.standard.
  @param m vettore della media degli attributi
  @sv vettore della deviazione standard.\\  
--------------------------------------------------------------------------------
*/
void TrainingSet::normalize(vector& m, vector& sv) 
{
  int i, j;
  vector v(n_attr);
	
	// avoid division by zero
	for (j=0; j<n_attr; j++)
	  if (sv[j] == 0)
		  v[j] = 1.0;
		else
		  v[j] = sv[j];
	
  for (i=0; i<n_train; i++)
     for (j=0; j<n_attr; j++)
       data(i,j) = (data(i,j) - m[j]) / v[j];
         
  is_normalized = 1;
} 




/**
     Suddivide il data set in train e validation set(n-fold cross - validation).\\
     Il data set corrente viene suddiviso in numfold parti (fold). Vengono creati numfold coppie di file 
     ognuna costituita da un fold (test set) e da numfold-1 fold (training set). Ogni fold a rotazione
     costituisce il test set. Le coppie di file vengono salvate in file secondo il formato standard.\\
     I campioni vengono assegnati in modo casuale ai diversi fold in modo da formare una partizione del data
     set di partenza.
     I nomi dei file vengono assegnati secondo la seguente sintassi: \\
     name.fi.train (training file formato da numfold -1 fold)\\
     name.fi.test (test file formato da 1 fold)\\
     i e' un numero che varia da 1 a numfold.
     @param numfold numero dei fold
     @param name nome base dellle coppie di file
     @sv vettore della deviazione standard.\\  
*/
void TrainingSet::FoldRand(unsigned numfold, char* name)
{
  unsigned i,n,j;
  int seed = (int) -time(NULL);
  
  unsigned numtest = (unsigned) floor(n_train/numfold);
  unsigned numtrain = n_train - numtest;
  
  // memorizza gli indici dei campioni estratti dal data set: 0 estratto - 1 non estratto
  array<unsigned> vmark(n_train); 
  // memorizza gli indici dei campioni inseriti nel fold corrente: 1 inserito - 0 non inserito
  array<unsigned> vfold(n_train);
  
  for (i=0; i<n_train; i++)  {
    vmark[i] = 1;
    vfold[i] = 0;
  }
  
  // costrisce il test ed il traimning file secondo la tecnica di folding
  for (n=1; n<numfold; n++)  {
    j=0;
    while (j<numtest)  {
      i = (unsigned)(ran1(&seed)*(n_train-1));
      if (vmark[i] == 1)  {
        vmark[i] = 0;
        vfold[i] = 1;
        j++;
      }
    }
    DoTestTrainFile(name, n, vfold);
    for (i=0; i<n_train; i++) vfold[i] = 0;
  } 
  DoTestTrainFile(name, numfold, vmark); 
} 
  
  
/**
     Memorizza i train e  i validation set (n-fold cross - validation).\\
     Nei file name vengono memorizzati i train ed i validation set sevondo le informazioni memorizzate nell'
     array v. Se v[i] = 1 il dato data(i,:) ed il target t[i] vanno memorizzati nel file name.fn.test,
     altrimenti vengono memeorizzati nel file name.fn.train.     
     @param name nome base della coppie di file
     @param n nesimo fold
     @param v memorizza l' assegnamento del dato al training o al test set 
*/
void TrainingSet::DoTestTrainFile(char* name, unsigned n, array<unsigned>& v) 
{
  unsigned i,k;
  const unsigned DIMBUF = 128;  
  char namebuf[DIMBUF];
  int len = strlen(name);
  ostrstream ost(namebuf,DIMBUF);
  
  ost << name << ".f" << n << FOLD_TRAIN_SUFFIX << '\0';  
  // creazione stream di output per il file di training 
  ofstream ftrain(namebuf);
  if (!ftrain) error("Cannot open train file ", namebuf);
  
  ost.seekp(0);
  ost << name << ".f" << n << FOLD_TEST_SUFFIX << '\0'; 
  // creazione stream di output per il file di testing 
  ofstream ftest(namebuf);
  if (!ftest) error("Cannot open test file ", namebuf);
  
  // scrittura dei pattern nel train e test set
  for(i=0; i < n_train; i++) {
    vector& patt = data.row(i);
    k = (unsigned)target[i];
    if (v[i] == 1)  
      save_pattern(ftest, patt, k);
    else 
      save_pattern(ftrain, patt, k);
  }

  
  // fine scrittura file  
  if (ftrain.bad())
    error("Error writing training set file ", name);
  ftrain << flush;
  if (ftest.bad())
    error("Error writing test set file ", name);
  ftest << flush;
    
}


/*
--------------------------------------------------------------------------------
  Salva il singolo pattern patt e la relativa classe target k nello stream fdata
--------------------------------------------------------------------------------
*/  
void TrainingSet::save_pattern(ofstream& fdata, vector& patt, unsigned k)
{
  for (int j=0; j<n_attr; j++)  
    fdata << patt[j] << sep;
  fdata << k << endl;
}



/**
     Salva il data set in un file.\\
		 Nel file name vengono salvati i pattern: in ogni riga sono elencati gli attributi ed in fondo alla
		 riga il target.   
     @param name nome del file      
*/
void TrainingSet::save(char* name) 
{
  unsigned i,k;
    
  ofstream f(name);
  if (!f) error("Cannot open data set file ", name);
   
  // scrittura dei pattern nel data set
  for(i=0; i < n_train; i++) {
    vector& patt = data.row(i);
    k = (unsigned)target[i];
    save_pattern(f, patt, k);
  }
  
  // fine scrittura file  
  if (f.bad())
    error("Error writing data set file ", name);
  f << flush;
}



/*
--------------------------------------------------------------------------------

   Classe TSetStrClass

--------------------------------------------------------------------------------
*/

/*
--------------------------------------------------------------------------------

   Costruttori

--------------------------------------------------------------------------------
*/

TSetStrClass ::TSetStrClass ( unsigned num_attr, unsigned num_train, char separator = ',')
: TrainingSet ( num_attr, num_train, separator), D()
{}

TSetStrClass::TSetStrClass ( unsigned num_attr, unsigned num_train, char *file, 
                             char separator = ',', class_pos pos = last)
: TrainingSet ( num_attr, num_train, separator), D()
{   
   n_class = read(file, 1, pos);
}

/* Costruisce il test set sulla base del dizionario dict */
TSetStrClass::TSetStrClass ( unsigned num_attr, unsigned num_train, char *file,
                  dictionary <string,unsigned>& dict, char separator = ',', class_pos pos = last) 
: TrainingSet ( num_attr, num_train, separator)
{ 
   D = dict;
   n_class = read(file, 0, pos);
}
   


/*
--------------------------------------------------------------------------------

   Legge il training set file e inizializza la matrice data con gli attributi
   del training set ed il vettore target con le classi target. La classe target nel file di input e' in
   formato stringa e viene automaticamente tradotta in formato numerico. 
   Se build_dict = 1 viene costruito il dizionario D assegnando alle classi-stringa 
   numeri-classe progressivi. 
   Se build_dict = 0 vengono assegnati alle classi-stringa numeri-classe sulla base 
   del dizionario esistente D.
   Ritorna il numero delle classi lette. 

--------------------------------------------------------------------------------
*/

unsigned TSetStrClass::read(char *file, unsigned build_dict, class_pos pos = last)
{
  int i,j;
  char c=0;
  char buf[BUFSIZE];
  unsigned n = 0;  // numero delle classi
  
  ifstream from(file); 
  if (!from) error("Cannot open input file :", file);
  
  if (pos == last)
    {
  		for (i=0; i<n_train; i++)
    		{
    			for (j=0; j<n_attr; j++)
    			 {
      			from >> data(i,j) >> c;
      			if (c != sep)
        			error_exit("Error reading separator");
    			 }
      		from >> buf;
      		if (build_dict)
      		  TranslateStringClass(buf, n, i);
      		else
      		  AssignStringClass(buf, n, i);
    		}
    }
  else // pos == first
    {      
  		for (i=0; i<n_train; i++)
    		{
    		  load_string (from, buf);
          if (build_dict)
      		  TranslateStringClass(buf, n, i);
      		else
      		  AssignStringClass(buf, n, i);
    			for (j=0; j<n_attr; j++)
    			 {
      			from  >> c >> data(i,j);
      			if (c != sep)
        			error_exit("Error reading separator");
    			 }      		
    		}
    }
     
     
  // salta eventuali caratteri rimasti (che devono essere solo blank)
  while (from.get(c));
          
  if (!from.eof()) 
     error("Error reading file ", file); 
  return n;
}





/*
--------------------------------------------------------------------------------

   Traduce la classe s in formato numerico utilizzando il dizionario D. Ad ogni nuova classe
   s viene associato un numero n progresivamente crescente 

--------------------------------------------------------------------------------
*/
void TSetStrClass::TranslateStringClass (string s, unsigned& n, int i)
    {
      dic_item it;
      
      if ((it = D.lookup(s)) == nil)
         { D.insert(s, ++n); target[i] = n;}
      else
           target[i] = D.inf(it);
    }
    

/*
--------------------------------------------------------------------------------

   Traduce la classe s in formato numerico utilizzando il dizionario D. Ad ogni nuova classe
   s viene associato il numero corrispondente alla chiave s. 
   n ritorna la classe maggiore; i e' l' indice dell' iesimo target 

--------------------------------------------------------------------------------
*/
void TSetStrClass::AssignStringClass(string s, unsigned& n, int i)
{
  dic_item it;
  unsigned cl;
  
  if ((it = D.lookup(s)) != nil) {        
        target[i] = cl = D.inf(it);
        if (cl > n) n = cl;
  }
  else
     error("TSetStrClass::AssignStringClass :", "class not present in the dictionary.");  
}  
 
    
   
/*
--------------------------------------------------------------------------------
   Stampa il dizionario delle classi su cout
--------------------------------------------------------------------------------
*/    
void TSetStrClass::print_dict()
{
  dic_item it;
  cout << "Stampa dizionario :" << endl;
  forall_items(it,D) cout << D.key(it) << " : " << D.inf(it) << endl;
  cout << "Dimensione di D : " << D.size() << endl;
}

/*
--------------------------------------------------------------------------------
   Stampa il vettore dei target in formato numerico e stringa
--------------------------------------------------------------------------------
*/ 
void TSetStrClass::print_target_dict()
{
  dic_item it;
  cout << "Target vector legend :" << endl;
  for (int i=0; i<n_train; i++)
    forall_items(it,D)
      if (D.inf(it) == target[i])
        {
          cout << i << " : " << target[i] << " : " << D.key(it) << endl;
          break;
        }
  cout << endl;
}
       
    
    
    
    
    
    
    
    
    
    
    
    
