// "tset.cc"


#include<ctype.h>
#include<stdlib.h>
#include<math.h>
#include<time.h>
#include<iostream.h>
#include<fstream.h>
#include"tset.h"
#include"utils.h"

static const unsigned BUFSIZE = 256;
static  const char SEPARATOR = ',';
static const int DATA_SIZE_LIGHT = 15;

TrainingSet::TrainingSet ( unsigned num_attr, unsigned num_train, char separator)
: data(num_train, num_attr), target(num_train), mean(num_attr), stdev(num_attr)
{
   n_attr = num_attr;
   n_train = num_train;
	 sep = separator;
   n_class = 0;
   is_normalized = 0;
}
TrainingSet::TrainingSet ( unsigned num_attr, unsigned num_train, char *file,  
                           char separator, class_pos pos)
: data(num_train, num_attr), target(num_train), mean(num_attr), stdev(num_attr)
{
   n_attr = num_attr;
   n_train = num_train;
	 sep = separator;
   n_class = read(file, pos);
   is_normalized = 0;
}
TrainingSet::~TrainingSet ()
{
}
unsigned TrainingSet::read(char *file, class_pos pos)
{
  int i,j;
  char c;
  unsigned n = 0;
  ifstream from(file); 
  if (!from) error("Cannot open input file : ", file);
  if (pos == last)
    {
  		for (i=0; i<n_train; i++)
    		{
    			for (j=0; j<n_attr; j++)
    			 {
      			if (sep==BLANK) 
						  from >> data(i,j);
						else {
						  from >> data(i,j) >> c;
      			  if (c != sep)
        			  error_exit("Error  reading separator in ", file);
						  }
    			 }
      		from >> target[i];
					from.get(c);
					if (isspace(c))
					  from.putback(c);
          else
					  error_exit("Error reading input file: ", "Dimension of input doesn' t agree with parameters");
      		if (target[i] > n) n = (unsigned) target[i];
					
    		}
    }
  else  
    {
  		for (i=0; i<n_train; i++)
    		{
    		  if (pos != noclass) {
          from >> target[i];
      		if (target[i] > n) n = (unsigned) target[i]; }
    			for (j=0; j<n_attr; j++) {
      			if (sep==BLANK) 
						  from >> data(i,j);
						else {						
						  from >> c >> data(i,j);
      			  if (c != sep)
        			  error_exit("Error  reading separator in ", file);
    			  }
					}
					from.get(c);
					if (isspace(c))
					  from.putback(c);
          else
					  error_exit("Error reading input file: ", "Dimension of input doesn' t agree with parameters");      		
    		}
    }
  while (from.get(c));
  if (!from.eof()) 
     error_exit("Error reading file ", file); 
  return n;
}
inline void TrainingSet::print_data()
{
  cout << "Matrix of attributes :" << endl;
   data.print();
}
inline void TrainingSet::print_target() 
{
  cout << "Target vector :" << endl; 
  cout << target << endl;
}
inline void TrainingSet::print_info()
{
  cout << "General info on data set : "<< endl;
  cout << "Number of patterns : " << Ntrain() << endl;
  cout << "Number of attributes : " << Nattr() << endl;
  cout << "Number of classes : " << Nclass() << endl;
  mean_calc();
  stdev_calc();
  if (is_normalized)
      cout << "Normalized attributes " << endl;     
  else
      cout << "Not normalized attributes." << endl;      
  cout << "Mean value of the attributes: " << endl;
  mean.print(); cout << endl;
  cout << "Standard deviation of attributes: " << endl;
  stdev.print(); cout << endl;
}
void TrainingSet::mean_calc(void)
{
  vect v(n_attr);
  for (int i=0; i<n_train; i++)
     v = v + data.row(i);
  v = v * (1.0/n_train);
  mean = v;  
}
void TrainingSet::stdev_calc(void)
{
  vect v(n_attr);
  double x;
  int j;
  for (int i=0; i<n_train; i++)
    for (j=0; j<n_attr; j++)
      {
        x = data(i,j) - mean[j];
        v[j] += (x*x);
      }
  v = v * (1.0/(n_train-1));
  for (j=0; j<n_attr; j++)
    stdev[j] = sqrt(v[j]);
}
void TrainingSet::normalize(void) 
{
  vect v(n_attr);
	mean_calc();
  stdev_calc();
  int i, j;
	 
	for (j=0; j<n_attr; j++)
	  if (stdev[j] == 0)
		  v[j] = 1.0;
		else
		  v[j] = stdev[j];
	
  for (i=0; i<n_train; i++)
     for (j=0; j<n_attr; j++)
       data(i,j) = (data(i,j) - mean[j]) / v[j];
  is_normalized = 1;
} 
void TrainingSet::normalize_var(void) 
{
  vect v(n_attr);
	mean_calc();
  stdev_calc();
  int i, j;
	 
	for (j=0; j<n_attr; j++)
	  if (stdev[j] == 0)
		  v[j] = 1.0;
		else
		  v[j] = stdev[j];
	
  for (i=0; i<n_train; i++)
     for (j=0; j<n_attr; j++)
       data(i,j) = (data(i,j) - mean[j]) / (v[j]*v[j]);
  is_normalized = 1;
} 
void TrainingSet::normalize(vect& m, vect& sv) 
{
  int i, j;
  vect v(n_attr);
	
	 
	for (j=0; j<n_attr; j++)
	  if (sv[j] == 0)
		  v[j] = 1.0;
		else
		  v[j] = sv[j];
	
  for (i=0; i<n_train; i++)
     for (j=0; j<n_attr; j++)
       data(i,j) = (data(i,j) - m[j]) / v[j];
  is_normalized = 1;
} 
void TrainingSet::FoldRand(unsigned numfold, char* name)
{
  unsigned i,n,j;
  long seed = (long) -time(NULL);
  unsigned numtest = (unsigned) floor((float)n_train/numfold);
  unsigned numtrain = n_train - numtest;
  vector<unsigned> vmark(n_train); 
  vector<unsigned> vfold(n_train);
  for (i=0; i<n_train; i++)  {
    vmark[i] = 1;
    vfold[i] = 0;
  }
  for (n=1; n<numfold; n++)  {
    j=0;
    while (j<numtest)  {
      i = (unsigned)(ran1(&seed)*(n_train-1));
      if (vmark[i] == 1)  {
        vmark[i] = 0;
        vfold[i] = 1;
        j++;
      }
    }
    DoTestTrainFile(name, n, vfold);
    for (i=0; i<n_train; i++) vfold[i] = 0;
  } 
  DoTestTrainFile(name, numfold, vmark); 
} 


void TrainingSet::save_subset(char* name, unsigned numsample, vect & subset)
{
  unsigned i,k;
  ofstream f(name);
  if (!f) error("Cannot open subset data file ", name);
  for(i=0; i < numsample; i++) {
    vect& patt = data.row((unsigned)subset[i]);
    k = (unsigned)target[(unsigned)subset[i]];
    save_pattern(f, patt, k);
  }
  if (f.bad())
    error("Error writing data set file ", name);
  f << flush;
}

void TrainingSet::Subsample_with_replacement (char* name, unsigned numsample, long initseed)
{
  static int tt = 0;
	int i,j;
	static long seed;
	
	// it initializes seed only on the firt call
	if ((tt==0)  && (initseed == 0))
	  seed = (long) -time(NULL);
	else seed = initseed;
	tt=1;
	// Vector recording the position in data and target of the selected subset of data
  vect subset(numsample);
  for (i=0; i<numsample; i++) {
	  j = (unsigned)(ran1(&seed)*(n_train-1));
		subset[i] = j;
	}
  save_subset(name, numsample, subset);
} 

TrainingSet* TrainingSet::generate_subsample(unsigned numsample, vect & subset) {
  
	unsigned i,k;
	TrainingSet* tset = new TrainingSet(n_attr, numsample);
  for(i=0; i < numsample; i++) {
    vect& patt = data.row((unsigned)subset[i]);
    k = (unsigned)target[(unsigned)subset[i]];
    tset->target[i] = k;
		tset->data.row(i) = patt;
  }
  return tset;
}

TrainingSet* TrainingSet::Subsample_with_replacement (unsigned numsample, long initseed)
{
  static int tt = 0;
	int i,j;
	static long seed;
	TrainingSet* tset;
	
	// seed initialization only on the first call
	if (tt==0)
	  if (initseed == 0)
	    seed = (long) -time(NULL);
	  else seed = initseed;
	tt=1;
	// Vector recording the position in data and target of the selected subset of data
  vect subset(numsample);
  for (i=0; i<numsample; i++) {
	  j = (unsigned)(ran1(&seed)*(n_train-1));
		subset[i] = j;
	}
  tset = generate_subsample(numsample, subset);
	return tset;
} 


void TrainingSet::DoTestTrainFile(char* name, unsigned n, vector<unsigned>& v) 
{
  unsigned i,k;
  const unsigned DIMBUF = 128;  
  char namebuf[DIMBUF];
  int len = strlen(name);
  ostrstream ost(namebuf,DIMBUF);
  ost << name << ".f" << n << FOLD_TRAIN_SUFFIX << '\0';  
  ofstream ftrain(namebuf);
  if (!ftrain) error("Cannot open train file ", namebuf);
  ost.seekp(0);
  ost << name << ".f" << n << FOLD_TEST_SUFFIX << '\0'; 
  ofstream ftest(namebuf);
  if (!ftest) error("Cannot open test file ", namebuf);
  for(i=0; i < n_train; i++) {
    vect& patt = data.row(i);
    k = (unsigned)target[i];
    if (v[i] == 1)  
      save_pattern(ftest, patt, k);
    else 
      save_pattern(ftrain, patt, k);
  }
  if (ftrain.bad())
    error("Error writing training set file ", name);
  ftrain << flush;
  if (ftest.bad())
    error("Error writing test set file ", name);
  ftest << flush;
}
void TrainingSet::save_pattern(ofstream& fdata, vect& patt, unsigned k)
{
  for (int j=0; j<n_attr; j++)  
    fdata << patt[j] << sep;
  if (k!=0)
    fdata << k << endl;
  else
    fdata << endl;  
}
void TrainingSet::save(char* name) 
{
  unsigned i,k;
  ofstream f(name);
  if (!f) error("Cannot open data set file ", name);
  for(i=0; i < n_train; i++) {
    vect& patt = data.row(i);
    k = (unsigned)target[i];
    save_pattern(f, patt, k);
  }
  if (f.bad())
    error("Error writing data set file ", name);
  f << flush;
}
void TrainingSet::save_pattern_light_format(ofstream& fdata, vect& patt, int k)
{
  //fdata.width(5);
	fdata << k << " ";
	for (int j=0; j<n_attr; j++) {
	  fdata << j+1 << ":";
		//fdata.width(DATA_SIZE_LIGHT); 
    fdata << patt[j] << " ";
	}
	fdata << endl;
}

void TrainingSet::save_light_format(char* name) 
{
  int i,k;
  ofstream f(name);
  if (!f) error("Cannot open light data set file ", name);
	
	f << "# " << name << " data file in SVM-light format." << endl;
  f << "# Number of patterns: " << n_train << endl;
	f.setf(ios::left, ios::adjustfield);
	// writing data
  for(i=0; i < n_train; i++) {
    vect& patt = data.row(i);
    k = (int)target[i];
		if (k!=1) k = -1;
    save_pattern_light_format(f, patt, k);
  }
  if (f.bad())
    error("Error writing data set file ", name);
  f << flush;
}

void TrainingSet::set_thresholds(float low_thresh, float hi_tresh){
  int i,j;
  for (i=0; i<n_train; i++)
    for (j=0; j<n_attr; j++) {
      if (data[i][j] < low_thresh)
        data[i][j] = low_thresh;
      else if (data[i][j] > hi_tresh)
        data[i][j] = hi_tresh;
    }
}

void TrainingSet::Transpose_data(void) { 
  matrix m = data.transpose();
  n_train = m.dim1(); 
  n_attr = m.dim2();
  vect v1(n_attr);
  mean = v1;
  stdev = v1;
  data = m; 
}

void TrainingSet::load_target(char* t){
  vect v(n_train);
  ifstream f(t);
  if (!f) error("Cannot open target file ", t);
  for (int i=0; i<n_train; i++) {
    f >> v[i];
    if (v[i] <=0)
      error_exit("Error rading target file ", t);
  }
  if (f.bad())
    error_exit("Error rading target file ", t);
  target = v;
}

/**
	  Generate a k-dimensional training set, with k<nattr, randomly keeping values from the original training set.\\
		@param k number of the dimensions to be randomly drawn
    @param v vector storing the indices of the randomly selected dimensions
		@return k-dimensional training set 		
  */
TrainingSet* TrainingSet::generate_random_subspace(int k, unsigned v[])
{
   static int tt=0;
	 static long seed;	   
	 bool dim_selected[n_attr];
	 
	 if (k >= n_attr)
	   error_exit("TrainingSet::generate_random_subspace : ","k must be less or equal than the number of attributes");
	 
	 for (int i=0; i<n_attr; i++)
	    dim_selected[i] = false;
	 
	 // it initializes the seed only on the first call of the method
	 if (tt==0)
	    seed = (long) -time(NULL);
	 tt=1;
	 
	 // selecting k random dimensions
	 unsigned n=0;
	 unsigned dim_index;
	 while (n<k) {
		 dim_index = (unsigned)(ran1(&seed)*(n_attr-1));
		 if (dim_selected[dim_index] == false) {
		    dim_selected[dim_index] = true;
				v[n++] = dim_index;		 		 
		 }	 
	 }
	 
	 // building the random subspace
	 TrainingSet* rstset = new TrainingSet(k, n_train);
	 rstset->target = target;
	 for (int j=0; j<k; j++) 
	   for (int i=0; i<n_train; i++) 
	     rstset->data.el(i,j) = data.el(i,v[j]);
	 
	 return rstset;
}
	 

////////////////////////////////////////////////////////////////////////////////////

TSetStrClass ::TSetStrClass ( unsigned num_attr, unsigned num_train, char separator)
: TrainingSet ( num_attr, num_train, separator), D()
{}
TSetStrClass::TSetStrClass ( unsigned num_attr, unsigned num_train, char *file, 
                             char separator, class_pos pos)
: TrainingSet ( num_attr, num_train, separator), D()
{   
   n_class = read(file, 1, pos);
}
TSetStrClass::TSetStrClass ( unsigned num_attr, unsigned num_train, char *file,
                  map <string,unsigned>& dict, char separator, class_pos pos) 
: TrainingSet ( num_attr, num_train, separator)
{ 
   D = dict;
   n_class = read(file, 0, pos);
}
unsigned TSetStrClass::read(char *file, unsigned build_dict, class_pos pos)
{
  int i,j;
  char c=0;
  char buf[BUFSIZE];
  unsigned n = 0;   
  ifstream from(file); 
  if (!from) error("Cannot open input file :", file);
  if (pos == last)
    {
  		for (i=0; i<n_train; i++)
    		{
    			for (j=0; j<n_attr; j++)
    			 {
      			if (sep==BLANK) 
						  from >> data(i,j);
						else {						
						  from >> data(i,j) >> c;
      			  if (c != sep)
        			  error_exit("Error reading separator in ", file);
						}
    			 }
      		from >> buf;
					if (strchr(buf, sep) != NULL)
					  error_exit("Error: Dimension of input doesn' t agree with parameters");
      		if (build_dict) 
      		  TranslateStringClass(buf, n, i); 
      		else
      		  AssignStringClass(buf, n, i);
    		}
    }
  else  
    {      
      for (i=0; i<n_train; i++)
    		{
    		  //load_string (from, buf, sep); ATTENZIONE load_string non funziona con i BLANK. 
          // Questa versione funziona solo con blank come separatore !!!!  BACO DA CORREGGERE!!!
          if (pos != noclass) {
            from >> buf;
					  if (buf[0] == sep)
					    error_exit("Error: Dimension of input doesn' t agree with parameters");
            if (build_dict)
      		    TranslateStringClass(buf, n, i);
      		  else
      		    AssignStringClass(buf, n, i);
          }
    			for (j=0; j<n_attr; j++)
    			 {
      			if (sep==BLANK) 
						  from >> data(i,j); 
						else {
						  from  >> c >> data(i,j);
      			  if (c != sep)
        			  error_exit("Error reading separator in ", file);
						}
    			 }      		
    		}
    }
  while (from.get(c));
  if (!from.eof()) 
     error_exit("Error reading file ", file); 
  return n;
}
void TSetStrClass::TranslateStringClass (string s, unsigned& n, int i)
    {
      map<string,unsigned>::iterator it = D.find(s);
			if (it == D.end()) { // if it is not present
			  D.insert(make_pair(s,++n));
				target[i] = n;
			}
			else
			  target[i] = (*it).second;
		}
				
void TSetStrClass::AssignStringClass(string s, unsigned& n, int i)
{
  unsigned cl;
	map<string,unsigned>::iterator it = D.find(s);
	if (it != D.end()) { // if it is  present
	  target[i] = cl = (*it).second;
		if (cl > n) n=cl; 
  }
  else
     error_exit("TSetStrClass::AssignStringClass :", "class not present in the dictionary.");  
}  
void TSetStrClass::print_dict()
{
  map<string,unsigned>::iterator it;
	cout << "Dictionary :" << endl;
	for (it = D.begin(); it!=D.end(); it++)
	  cout << (*it).first << " : " << (*it).second << endl;
	cout << "Dictionary dimension : " << D.size() << endl << endl;
}

