// "utils.cc"
#define IA 16807
#define IM 2147483647
#define AM (1.0/IM)
#define IQ 127773
#define IR 2836
#define NTAB 32
#define NDIV (1+(IM-1)/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)

#include<stdlib.h>
#include<ctype.h>
#include<math.h>
#include<iostream.h>
#include<iomanip.h>
#include<fstream.h>
#include<string>
#include <unistd.h>
#include"utils.h"

#define DIMBUF  128000

using namespace std;

void error(char* s, char* s2)
{
  cerr << s << s2 << endl;
}
void error_exit(char* s, char* s2)
{
  cerr << s << s2 << endl;
	exit(1);
}
float calc_max(float *v, unsigned dim)
{
  unsigned i;
	float max = v[0];
	
	for (i=1; i<dim; i++)
	  if (v[i] > max)
		  max = v[i];
	return max;
}
float calc_min(float *v, unsigned dim)
{
  unsigned i;
	float min = v[0];
	
	for (i=1; i<dim; i++)
	  if (v[i] < min)
		  min = v[i];
	return min;
}
int wc_word (char *file)
{
  const int IN = 1;
  const int OUT = 0;
  char c;
  int nw = 0;
  int state = OUT;
  ifstream from(file);
  if (!from) error_exit("Cannot open input file ", file);
  while (from.get(c))
    {
      if (isspace(c)) state = OUT;
      else if (state == OUT)
        {
          state = IN;
          nw++;
        }
    }
  if (!from.eof())
    cerr << "Error reading file in wc " << endl;
  from.close(); 
  return nw;
}
int wc_word (string& s)
{
  const int IN = 1;
  const int OUT = 0;
  char c;
  int nw = 0;
  int state = OUT;
  
	for (int i=0; i<s.size(); i++)
    {
			if (isspace((char)s[i])) state = OUT;
      else if (state == OUT)
        {
          state = IN;
          nw++;
        }
    }
  return nw;
}
int wc (char *file)
{
  string line;
	const int IN = 1;
  const int OUT = 0;
  char c;
  int nw = 0;
  int state = OUT;
  ifstream from(file);
  if (!from) error_exit("Cannot open input file ", file);
  while (getline(from,line))
	    if (line.length()!=0)
         nw++;
  if (!from.eof())
    cerr << "Error reading file in wc " << endl;
  from.close(); 
  return nw;
}
void load_string (ifstream& from, char *buf, char sep)
{
  char c;
  int k = -1;
  for(;;)
   {
     from >> c;
     if ((c == sep) || (isspace(c)))
    	 {
    		 from.putback(c);
    		 buf[++k] = '\0';
    		 break;
    	 }
    	else
    		 buf[++k] = c;
   }
}  
void extract(char from[], char to[], char c)
{
  int i;
  int len = strlen(from);
  for (i=0; i<len; i++)  {
    if (from[i] == c) break;
    to[i]=from[i];
  }
  to[i]='\0';
}
istream& eatwhite(istream& is)
{
  char c;
  while(is.get(c)) {
    if (!isspace(c))  {
      is.putback(c);
      break;
    }
  }
  return is;
}
double Elapsed_sec(clock_t start, clock_t end)
{
  return (((double) (end - start)) / sysconf(_SC_CLK_TCK));
} 
void print_elapsed_cpu_time(clock_t start, clock_t end, ostream& f)
{
  char sec[8];
	double elapsed_sec = Elapsed_sec(start, end);
  unsigned minutes = (unsigned)floor(elapsed_sec/60);
  unsigned seconds = (unsigned)(elapsed_sec - minutes*60);
	unsigned msec = (unsigned)((elapsed_sec - (minutes*60 + seconds))*1000);
  f << "CPU time (min.sec.msec) : " << minutes <<   "."  << sec << "." << msec << endl;
}
void print_elapsed_time_from_start(ostream& f)
{
  struct tms buf;
	times (&buf);
	double el = (double)(buf.tms_utime + buf.tms_stime + buf.tms_cutime + buf.tms_cstime)/sysconf(_SC_CLK_TCK); 
	f << "Elapsed CPU time (sec.): " << el << endl;
}
void print_elapsed_time(struct tms *buf, ostream& f)
{
	double el = (double)((*buf).tms_utime + (*buf).tms_stime + (*buf).tms_cutime + (*buf).tms_cstime)/sysconf(_SC_CLK_TCK); 
	f << "Elapsed CPU time (sec.): " << el << endl;
}
double NormL1(const vect& v1, const vect& v2) 
{
  int j;
  double dist = 0;
  int len = v1.dim();
  for (j=0; j<len; j++) 
    dist += fabs(v1[j] - v2[j]);
  return dist;
}
double NormL2(const vect& v1, const vect& v2) 
{
  int j;
  double x, dist = 0;
  int len = v1.dim();
  for (j=0; j<len; j++) {
    x = v1[j] - v2[j];
    dist += x*x;
  }
  return dist;
}
double Hamming(const vect& v1, const vect& v2)
{
  int j;
  double dist = 0;
  int len = v1.dim();
  unsigned val1, val2;
  for (j=0; j<len; j++) {
     if (v1[j] < 0.5) val1 = 0; else val1 = 1;
     if (v2[j] < 0.5) val2 = 0; else val2 = 1;
     if (val1!=val2) dist+=1;
  }        
  return dist;
}
void print_conf_matrix(matrix& conf, ostream& f)
{
  int dim = conf.dim1();
  f.width(5);
	f.setf(ios::fixed, ios::floatfield);
  //f.precision(0); 
  f << "Confusion matrix :" << endl;
    for (int i=0; i<dim; i++)  {
      for (int j=0; j<dim; j++)
	      f << setw(5) << (unsigned)conf(i,j) << " ";
      f << endl;
    }
  f << endl << flush;
}
void SumConf(matrix& sumconf, matrix& conf) 
{
  int dim1 = conf.dim1();
  int dim2 = sumconf.dim1();
  if (dim1!=dim2)
    error_exit ("SumConf : dimensions do not agree");
  for (int i=0; i<dim1; i++)  
    for (int j=0; j<dim1; j++) 
      sumconf(i,j) = (sumconf(i,j) + conf(i,j));
} 
void CopyConf(matrix& conf_to, matrix& conf_from)
{
  int dim1 = conf_to.dim1();
  int dim2 = conf_from.dim1();
  if (dim1!=dim2)
    error_exit ("CopyConf : dimensions do not agree");
  for (int i=0; i<dim1; i++)  
    for (int j=0; j<dim1; j++) 
      conf_to(i,j) = conf_from(i,j);
}

istream& toendofline(istream& is)
{
  char c;
	while (is.get(c)) 
	  if (c == '\n') break;
	return is;
}


void delete_dir(char file[]) {
  
	const int SIZE = 256;
	char buf [SIZE];
	
	strcpy (buf, file);
	int len = strlen(buf);
	int i = len+1;
	bool found = false;
	
	while ((!found) && (i>0)) 
	  if (buf[--i] == '/')
		  found = true;
	if (buf[i] == '/')
	  strncpy(file, buf+i+1, len - i);
}

/*  ******************************************
float gasdev(int *idum)
{
	static int iset=0;
	static float gset;
	float fac,r,v1,v2;
	
	if  (iset == 0) {
		do {
			v1=2.0*ran1(idum)-1.0;
			v2=2.0*ran1(idum)-1.0;
			r=v1*v1+v2*v2;
		} while (r >= 1.0);
		fac=sqrt(-2.0*log(r)/r);
		gset=v1*fac;
		iset=1;
		return v2*fac;
	} else {
		iset=0;
		return gset;
	}
}

******************************************** */

/* ******************************************
float ran1(int *idum)
{
	static long ix1,ix2,ix3;
	static float r[98];
	float temp;
	static int iff=0;
	int j;
	
	if (*idum < 0 || iff == 0) {
		iff=1;
		ix1=(54773 -(*idum)) % 259200 ;
		ix1=(7141 *ix1+ 54773 ) % 259200 ;
		ix2=ix1 % 134456 ;
		ix1=(7141 *ix1+ 54773 ) % 259200 ;
		ix3=ix1 % 243000 ;
		for (j=1;j<=97;j++) {
			ix1=(7141 *ix1+ 54773 ) % 259200 ;
			ix2=(8121 *ix2+ 28411 ) % 134456 ;
			r[j]=(ix1+ix2* (1.0/ 134456 ) )* (1.0/ 259200 ) ;
		}
		*idum=1;
	}
	ix1=(7141 *ix1+ 54773 ) % 259200 ;
	ix2=(8121 *ix2+ 28411 ) % 134456 ;
	ix3=(4561 *ix3+ 51349 ) % 243000 ;
	j=1 + ((97*ix3)/ 243000 );
	if ((j > 97) || (j < 1)) error("Error in ran1");
	temp=r[j];
	r[j]=(ix1+ix2* (1.0/ 134456 ) )* (1.0/ 259200 ) ;
	return temp;
}

***************************************************** */


float gasdev(long *idum)
{
	float ran1(long *idum);
	static int iset=0;
	static float gset;
	float fac,rsq,v1,v2;

	if  (iset == 0) {
		do {
			v1=2.0*ran1(idum)-1.0;
			v2=2.0*ran1(idum)-1.0;
			rsq=v1*v1+v2*v2;
		} while (rsq >= 1.0 || rsq == 0.0);
		fac=sqrt(-2.0*log(rsq)/rsq);
		gset=v1*fac;
		iset=1;
		return v2*fac;
	} else {
		iset=0;
		return gset;
	}
}

float ran1(long *idum)
{
	int j;
	long k;
	static long iy=0;
	static long iv[NTAB];
	float temp;

	if (*idum <= 0 || !iy) {
		if (-(*idum) < 1) *idum=1;
		else *idum = -(*idum);
		for (j=NTAB+7;j>=0;j--) {
			k=(*idum)/IQ;
			*idum=IA*(*idum-k*IQ)-IR*k;
			if (*idum < 0) *idum += IM;
			if (j < NTAB) iv[j] = *idum;
		}
		iy=iv[0];
	}
	k=(*idum)/IQ;
	*idum=IA*(*idum-k*IQ)-IR*k;
	if (*idum < 0) *idum += IM;
	j=iy/NDIV;
	iy=iv[j];
	iv[j] = *idum;
	if ((temp=AM*iy) > RNMX) return RNMX;
	else return temp;
}

#undef IA
#undef IM
#undef AM
#undef IQ
#undef IR
#undef NTAB
#undef NDIV
#undef EPS
#undef RNMX
