/*******************************************************************************
vect.h 
*******************************************************************************/
#ifndef VECT_H
#define VECT_H

#include<iostream>
#include <fstream>
#include <strstream>
#include <string>
#include <ctype.h>

//
using namespace std;


/**
   Class for manipulating vector of real numbers.\\
*/ 
class vect {

  friend class matrix;
  
	// a vector is allocated as an array of double
  double* v;
	// dimension of the vector
  int d;
  
	// Checks if dimension of the vector agrees
  void control_dimensions(const vect&) const;
 
public:

/**
	 Constructor.
	 Dummy constructor initialized to the zero-dimensional vector
*/
vect(); 

 /**
	 Constructor.
	 The vector is initialized to the dimension n. All its elements are initialized to zero
	 @param n vector dimension
	 */
vect(int n, double val = 0.0); 

 /**
	 Constructor.
	 The vector is initialized using the array of double array.
	 @param array array of double
	 @param n vector dimension
	 */
vect(double* array, int n); 

/**
	 Copy Constructor.
*/
vect(const vect&);

/**
	 Destructor.
*/
~vect(); 

/**
	 Overloading assignment operator.
*/
vect& operator=(const vect&);

/**
	 Returns the dimension of the vector. \\
	 @return dimension of the vector
	*/ 
int    dim()    const { return d; }


/**
	 Index overloading. \\
	 It returns the i-th componenet of the vector.
	 @param i i-th component of the vector 
	 @return  value of i-th component of the vector
	*/   
double& operator[](int i);

/**
	 Index overloading. \\
	 It returns the i-th componenet of the vector.
	 @param i i-th component of the vector 
	 @return  value of i-th component of the vector
*/     
double  operator[](int) const;

/**
	 Returns the square of the Euclidean length of the vector. \\
	 @return  square of the Euclidean length of the vector
*/     
double sqr_length() const;

/**
	 Returns the  Euclidean length of the vector. \\
	 @return  Euclidean length of the vector
*/  
double length() const;


///	 Returns the vector normalized. 
vect norm() const { return *this/length(); }

	  
/**
	 Overloading += operator.
*/
vect& operator+=(const vect&);

/**
	 Overloading -= operator.
*/
vect& operator-=(const vect&);

/**
	 Overloading addition.
*/  
vect  operator+(const vect& v1) const;

/**
	 Overloading subtraction.
*/  
vect  operator-(const vect& v1) const;

/**
	 Overloading Scalar multiplication.
*/  
double  operator*(const vect& v1) const;

/**
	 Componentwise multiplication.
*/
vect  operator*(double r) const;

/**
	 Unary minus.
*/
vect  operator-() const;

/**
	Componentwise division.
*/  
vect  operator/(double) const;
  
/**
	Boolean equality test
*/  
bool     operator==(const vect& w) const;

/**
	Boolean inequality test
*/ 
bool     operator!=(const vect& w)  const { return !(*this == w); }

/**
	Multiplication scalat * vector
*/ 
friend vect operator*(double f, const vect& v) { return v *f; }

/**
	Prints the vector componentwise to ostream.
*/ 
void print(ostream& O);

/**
	Prints the vector componentwise to cout.
*/ 
void print() { print(cout); }

/**
	Reads numbers from an istream.\\
	Reads d = v.dim() numbers from input stream I and writes 
         them into v[0] ... v[d-1].
*/ 
void read(istream& I);

/**
	Reads numbers from cin.\\
*/ 
void read() { read(cin); }

/**
	Writes v componentwise to the output stream O.
*/   
friend ostream& operator<<(ostream& O, const vect& v);

/**
	Read v componentwise from the input stream I.
*/ 
friend istream& operator>>(istream& I, vect& v);

};


#endif
